/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	26-Jul-2005
 *
 * Added support for uncertainties, 09-Mar-2023, CT
 */
#ifndef __DATA_USRBIN_H
#define __DATA_USRBIN_H

#include <assert.h>

#include <vector>
#include <string>
#include <fstream>

#include "math/matrix4.h"
#include "data/meshdata.h"

class FortranFile;

// ================================== Usrbin ===================================
class Usrbin : public MeshData {
public:
static const int RunTit_Length = 80;
static const int RunTim_Length = 32;
static const int TitUsb_Length = 10;
static const int USRBINDataHeaderLen = 86;	// header length of a data record
static constexpr char USRBINErrHeader[] = { "STATISTICS" };
// header length of a statistics record
static const int USRBINErrHeaderLen   = sizeof(USRBINErrHeader) - 1 + sizeof(int);
// EOR in 64bit FORTRAN binary data
static const int USRBINEndOfRecordLen = sizeof( unsigned long );

public:
	enum class Type : int {
		undefined      = -1,
		XYZ_point      =  0,
		XYZ            = 10,

		RPhiZ_point    =  1,
		RPhiZ          = 11,

		region_point   =  2,
		region         = 12,

		Xsym_point     =  3,
		Xsym           = 13,

		Ysym_point     =  4,
		Ysym           = 14,

		Zsym_point     =  5,
		Zsym           = 15,

		XYZsym_point   =  6,
		XYZsym         = 16,

		RPhiZsym_point =  7,
		RPhiZsym       = 17,

		special_point  =  8,
		special        = 18
	}; // Type
public:
	char	runtitle[RunTit_Length+1],
		runtime[RunTim_Length+1];

	int	lntzer;			// not zero flag
	float	bkusbn;
	float	b2usbn;
	float	tcusbn;

	float	weipri;
	int	ncase;
	int	mcase;
	int	nbatch;

private:
	Type	_type;			// type of usrbin
	int	_score;			// quantity scored

static	const char*	_typeStr[];

public:
	Usrbin() : MeshData(), _type(Type::undefined), _score(0) {}
	//Usrbin(const char* fn, int det)	: MeshData(fn,det) {}
virtual	~Usrbin()	{ cleanup(); }

virtual	void	cleanup() override;
virtual	bool	load(const char* fn, int det) override;
virtual	bool	save(const char* fn=NULL) override;
static	bool	save(const char* fn, std::vector<Usrbin>&);

	// for saving multiple usrbin in a single one
	bool	writeHeader(FortranFile& f);
	bool	writeData(FortranFile& f, int det=0);
static	bool	writeErrorHeader(FortranFile& f);
	bool	writeError(FortranFile& f);

	/* get/set */
	Type	type()		const	{ return _type;  }
	void	type(Type t)		{ _type = t;     }
	int	score()		const	{ return _score; }

	void	create( int det, Type atype, int ascore,
			const double axlow, const double axhigh, const int anx,
			const double aylow, const double ayhigh, const int any,
			const double azlow, const double azhigh, const int anz );

static	std::vector<std::string> getDetectorNames(const char* fn);
static	std::size_t getNrOfDetectors(const char* fn) { return getDetectorNames(fn).size(); }
}; // Usrbin
#endif
