/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	26-Jul-2005
 */

#include "vtkdata.h"

#include <fstream>
#include <iostream>

#include "math/eps.h"
#include "tool/token.h"

const char* VtkData::_typeStr[] = {
		"RECTILINEAR GRID",
		""
	};

static const std::string vtk_header = "# vtk DataFile Version ";
static const std::string vtk_error  = "ERROR parsing vtk file: ";

/** mustbe */
static bool mustbe(std::ifstream& stream, const char* target, const char* err=nullptr)
{
	std::string token;
	stream >> token;
	if (!token.compare(target)) return true;
	std::cerr << vtk_error;
	if (err)
		std::cerr << *err;
	else
		std::cerr << target << " expected";
	std::cerr << std::endl;
	return false;
} // mustbe

/** skipnewline */
static void skipnewline(std::ifstream& stream)
{
	char ch;
	do {
		stream.read(&ch, 1);
	} while (ch != '\n');
} // skipnewline

// ================================== VtkData ===================================
/** cleanup */
void VtkData::cleanup()
{
	MeshData::cleanup();
} // cleanup

/** load
 * @return	true on success
 *		false on failure
 */
bool VtkData::load(const char* fn, int det)
{
	bool binary;
	std::string title;

	switch (preLoad(fn,det)) {
		case  1: return true;
		case -1: return false;
		default: ;
	}

	std::ifstream stream(_filename, std::ios::binary);

	// Starting comment / header
	std::string line;
	std::getline(stream, line);
	if (line.compare(0, vtk_header.length(), vtk_header)) {
		std::cerr << "ERROR parsing vtk file: Unknown header found" << std::endl;
		goto ERROR;
	}

	// Title
	std::getline(stream, title);

	// Ascii or Binary
	std::getline(stream, line);
	binary = !line.compare("BINARY");
	if (not binary and not !line.compare("ASCII")) {
		std::cerr << vtk_error << "ASCII or BINARY expected" << std::endl;
		goto ERROR;
	}

	// Dataset
	if (not mustbe(stream, "DATASET")) goto ERROR;

	stream >> line;
	if (!line.compare("RECTILINEAR_GRID")) {
		rphizMesh = false;
	} else {
		std::cerr << vtk_error << "Unsupported format" << std::endl;
		goto ERROR;
	}

	// Dimensions
	if (not mustbe(stream, "DIMENSIONS")) goto ERROR;

	stream >> nx >> ny >> nz;
	nx--;
	ny--;
	nz--;

	if (not mustbe(stream, "X_COORDINATES")) goto ERROR;
	int   temp;
	float value;		// to read floating point value
	stream >> temp;
	if (not mustbe(stream, "float")) goto ERROR;
	assert(nx == temp-1);
	if (binary) {
		skipnewline(stream);
		stream.read((char*)&value, sizeof(float));
		xlow = ByteSwap(value);
		stream.seekg((nx-1)*sizeof(float), stream.cur);
		stream.read((char*)&value, sizeof(float));
		xhigh = ByteSwap(value);
		skipnewline(stream);
	} else {
		stream >> xlow;
		for (int i=0; i<nx; i++) stream >> xhigh;
	}

	if (not mustbe(stream, "Y_COORDINATES")) goto ERROR;
	stream >> temp;
	if (not mustbe(stream, "float")) goto ERROR;
	assert(ny == temp-1);
	if (binary) {
		skipnewline(stream);
		stream.read((char*)&value, sizeof(float));
		ylow = ByteSwap(value);
		stream.seekg((ny-1)*sizeof(float), stream.cur);
		stream.read((char*)&value, sizeof(float));
		yhigh = ByteSwap(value);
		skipnewline(stream);
	} else {
		stream >> ylow;
		for (int i=0; i<ny; i++) stream >> yhigh;
	}

	if (not mustbe(stream, "Z_COORDINATES")) goto ERROR;
	stream >> temp;
	if (not mustbe(stream, "float")) goto ERROR;
	assert(nz == temp-1);
	if (binary) {
		skipnewline(stream);
		stream.read((char*)&value, sizeof(float));
		zlow = ByteSwap(value);
		stream.seekg((nz-1)*sizeof(float), stream.cur);
		stream.read((char*)&value, sizeof(float));
		zhigh = ByteSwap(value);
		skipnewline(stream);
	} else {
		stream >> zlow;
		for (int i=0; i<nz; i++) stream >> zhigh;
	}

	dx  = (xhigh-xlow) / (double)nx;
	dy  = (yhigh-ylow) / (double)ny;
	dz  = (zhigh-zlow) / (double)nz;

	// CELL DATA
	if (not mustbe(stream, "CELL_DATA")) goto ERROR;
	stream >> temp;
	assert(temp == nx*ny*nz);

	// FIELD data
	if (not mustbe(stream, "FIELD")) goto ERROR;
	stream >> line;
	int ndet;
	stream >> ndet;
	if (det >= ndet) goto ERROR;

	// skip previous detectors
	for (int i=0; i<det-1; i++) {
		stream >> _name;
		stream >> temp;
		int count;
		stream >> count;
		assert(count == nx*ny*nz);
		// line must be "int","float"
		stream >> line;
		if (binary) {
			skipnewline(stream);
			stream.seekg(count*sizeof(float), stream.cur);
			skipnewline(stream);
		} else
			for (int j=0; j<count; j++) {
				float skip;
				stream >> skip;
			}
	}
	if (det <= ndet) {
		// load current one
		stream >> _name;
		stream >> temp;
		int count;
		stream >> count;
		assert(count == nx*ny*nz);
		stream >> line;
		// line must be "int","float"
		bool asfloat = !line.compare("float");
		allocateData();
		if (binary) {
			skipnewline(stream);
			if (asfloat)
				for (int j=0; j<count; j++) {
					stream.read((char*)&value, sizeof(float));
					data[j] = ByteSwap(value);
				}
			else // as int
				for (int j=0; j<count; j++) {
					stream.read((char*)&temp, sizeof(int));
					data[j] = (float)ByteSwap(temp);
				}
			skipnewline(stream);
		} else
			for (int j=0; j<count; j++)
				stream >> data[j];
		postLoad();
	}

	stream.close();
	return true;

ERROR:
	stream.close();
	return false;
} // load

/** load error from a separate vtk file */
bool VtkData::loadError(const char* fn, int det)
{
	VtkData vtkerror;
	if (!vtkerror.load(fn, det)) return false;

	// move error to 'this'
	assert(_error == nullptr);
	if (totalSize() != vtkerror.totalSize()) return false;

	_error = vtkerror.data;
	_errorsPresent = true;
	vtkerror.data  = nullptr;
	return true;
} // loadError

/** save usrbin data to file
 * @param	filename
 * @return	true on success
 *		false on failure
 */
bool VtkData::save(const char* /*fn*/)
{
	//char buffer[512];
	//if (fn) filename = fn;

	//FortranFile f(filename.c_str(), "wb");
	//if (!f) return false;

	//FortranParser parser(buffer, sizeof(buffer));

	//// Write title
	//parser.write(runtitle, USRBIN_RUNTIT);
	//parser.write(runtime,  USRBIN_RUNTIM);
	//parser.write(weipri);
	//parser.write(ncase);
	//parser.write(mcase);
	//parser.write(nbatch);
	//f.write(parser);

	//// Write detector header
	//parser.reset();
	//parser.write(_detector);		// detector number
	//parser.write(name, USRBIN_TITUSB);	// usrbin detector title
	//parser.write((int)_type);		// type
	//parser.write(_score);			// score

	//parser.write((float)xlow);
	//parser.write((float)xhigh);
	//parser.write(nx);
	//parser.write((float)dx);

	//parser.write((float)ylow);
	//parser.write((float)yhigh);
	//parser.write(ny);
	//parser.write((float)dy);

	//parser.write((float)zlow);
	//parser.write((float)zhigh);
	//parser.write(nz);
	//parser.write((float)dz);

	//parser.write(lntzer);
	//parser.write(bkusbn);
	//parser.write(b2usbn);
	//parser.write(tcusbn);
	//assert(parser.length() == 86);
	//f.write(parser);

	//// Write data
	//int len = totalSize()*sizeof(float);
	//if (f.write(data, len) != len)
	//	fprintf(stderr,"ERROR writing usrbin file \"%s\"\n",filename.c_str());

	return true;
} // save
