////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M - Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//          Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "bxdfdispersivetransmission.h"
#include "bxdffresnel.h"
#include "spectrahelpers.h"

#include <algorithm>

CBxDFDispersiveSpecTransmission::CBxDFDispersiveSpecTransmission( double DispersionPower, unsigned int TotalNrOfSamples, const Color& TransmissionColor /* = Color( 1.0, 1.0, 1.0 ) ) */ ) : CBxDFSpecTransmission( 0.0, TransmissionColor ), m_DispersionPower( DispersionPower ),
m_CauchyA(0), m_CauchyB(0), m_CurSampleIndex( 0 ), m_TotalNrOfSamples( TotalNrOfSamples)
{
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////
Color CBxDFDispersiveSpecTransmission::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0 */ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Shade ); UNUSED( Power ); UNUSED( LightColor );

	if( Type == CBxDFBase::BSDF_TRANSMISSION )
		return Sample_f_Transmission( Out, In, Pdf, Normal, RenderMat, pRandNr, LightColor);

	return Color::Black;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CBxDFDispersiveSpecTransmission::Sample_f_Transmission( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, const Color& LightColor )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	//////////////////////
	// handle dispersion
	// TODO: implement dispersion jittering by using a random offset
	double EtaT = 0.0;
	double JitterOffset = 0.5; // pRandNr[0];
	Color  DispCol;
	m_CurSampleIndex++;

	/* A table of Cauchy coefficients for common optical materials is shown below: (http://en.wikipedia.org/wiki/Cauchy's_equation)
		Material A B (micro-m2)
		Fused silica 1.4580 0.00354
		Borosilicate glass BK7 1.5046 0.00420  --> B = 4200
		Hard crown glass K5 1.5220 0.00459     --> B = 4590
		Barium crown glass BaK4 1.5690         --> B = 5310
		Barium flint glass BaF10 1.6700        --> B = 7430
		Dense flint glass SF10 1.7280          --> B = 1342
	*/

	m_CauchyA = 1.5220;
	m_CauchyB = 4590;
	EtaT = CSpectraHelpers::GetIORColor( (m_CurSampleIndex + JitterOffset) / m_TotalNrOfSamples, m_CauchyA, m_CauchyB, DispCol);

	// Or we can calculate it like yafray. even though this is computationally expensive and not necessarily worth it
	//
	//CSpectraHelpers::CauchyCoefficients( RenderMat.ior(), m_DispersionPower, m_CauchyA, m_CauchyB);
	//EtaT = CSpectraHelpers::GetIORColor( (m_CurSampleIndex + JitterOffset) / m_TotalNrOfSamples, m_CauchyA, m_CauchyB, DispCol);
	//
	///////////////////////

	UNUSED( pRandNr );

	bool   bEntering = Out.type == Ray::RayType::Transmission ? false : true;
	double EtaI = 1.0;			// to be physically precise this IOR should be carried by the outgoing ray --> TODO:

	Pdf = 1.0;

	double Cosi = Clamp( Normal.dot( Out.dir() ), -1.0, 1.0 );
	if ( bEntering ) {
		Cosi = -Cosi;
	}
	else {
		std::swap( EtaI, EtaT );
	}

	double k = 1 - (EtaI / EtaT) * (EtaI / EtaT) * (1.0 - Cosi * Cosi);
	if ( k < 0 ) {
		return Sample_f_Reflection( Out, In, Pdf, Normal, RenderMat, pRandNr, LightColor );

	}

	Vector Temp = (EtaI / EtaT) * Out.dir() + ((EtaI / EtaT) * Cosi - sqrtf( k )) * Normal;
	Temp.normalize();

	// set up ray
	In = Out;
	In.init();
	In.depth = Out.depth + 1;

	RaySegment NewSegment( Out.hit( SMALL3D4 * 100 ), Temp, Out.hitZone() );
	NewSegment.tmin = SMALL3D3;
	NewSegment.tmax = INFINITE;

	In.push( NewSegment );
	In.type = bEntering ? Ray::RayType::Transmission : Ray::RayType::Standard;	// if the out ray was entering then this is a transmission ray (inside a solid), otherwise it is an exiting ray

	// Fresnel
	double F = m_pFresnel->Evaluate( Cosi, EtaI, EtaT );
	assert( F >= 0.0 && F <= 1.0 );

	double FresnelRef = (EtaI * EtaI) / (EtaT * EtaT) * (1.0 - F ) / fabs(Cosi);

	return LightColor * DispCol * FresnelRef;
}
// vim:ts=2
