#pragma once
#ifndef CONFIG_PARSER_H
#define CONFIG_PARSER_H

#include <iostream>
#include <string>
#include <vector>
#include <fstream>
#include <sstream>
#include <map>

class CConfigParser
{
public:
	CConfigParser() : m_bStatus(false) {};

	// create & read config file
	CConfigParser( const std::string& FileName ) {
		m_bStatus = Read( FileName );
	};

	// check status
	bool			IsOK() const { return m_bStatus; };
	bool operator!() const { return !IsOK(); };

	// load
	bool			Load( const std::string& FileName ) { m_Data.clear();  m_bStatus = Append( FileName ); return m_bStatus; }
	bool			Append( const std::string& FileName ) { m_bStatus = m_bStatus | Read( FileName ); return m_bStatus; }

	// check if we have config flags
	bool			Empty() { return m_Data.empty(); };

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// verbose

	void			Show()

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		std::cerr << "         FARM - configuration file" << std::endl;
		std::cout << "--------------------------------------------" << std::endl;

		for( auto Iter = m_Data.begin(); Iter != m_Data.end(); ++Iter ) {
			std::cout << Iter->first << " = ";
			for( auto ValIter = Iter->second.begin(); ValIter != Iter->second.end(); ++ValIter )
				std::cout << *ValIter << " ";
			std::cout << std::endl;
		}
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	template<typename T>
	T		GetValue( const std::string& Key, const T& DefaultVal )

	// get value
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		T			CurVal;
		bool  bSuccess = false;
		std::string		SearchKey( Key );

		if( !m_bStatus )
			return DefaultVal;

		ToUpper( SearchKey );

		// check if there is a value otherwise we return the default
		auto Iter = m_Data.find( SearchKey );
		if( Iter == m_Data.end() ) {
			return DefaultVal;
		}

		// sanity checks
		if( Iter->second.empty() ) {
			return DefaultVal;
		}

		if( Iter->second.size() > 1 ) {
			std::cerr << "Warning: Config parser: key = " << Iter->first << " has multiple values, but only the first is being considered!" << std::endl;
		}

		// check type conversion
		CurVal = FromStringEx<T>( Iter->second.at(0), bSuccess );
		if( !bSuccess ) {
			std::cerr << "Warning: Config parser: value for " << Iter->first << " does not match the requested data type. Using default value = " << DefaultVal << std::endl;
			return DefaultVal;
		}

		return CurVal;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	void		SetValue( const std::string& Key, std::string& Value )

	// set value
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		std::string		SearchKey( Key );

		ToUpper( SearchKey );
		m_Data[SearchKey] = std::vector<std::string>({Value});
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	template<typename T>
	std::vector<T>		GetValueArray( const std::string& Key, const std::vector<T>& DefaultVal )

	// get array of values
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		T								CurVal;
		bool						bSuccess = false;
		std::vector<T>	ValArray;
		std::string			SearchKey( Key );

		if( !m_bStatus )
			return DefaultVal;

		ToUpper( SearchKey );

		// check if there is a value otherwise we return the default
		auto Iter = m_Data.find( SearchKey );
		if( Iter == m_Data.end() ) {
			return DefaultVal;
		}

		// check type conversion
		for( size_t i = 0; i < Iter->second.size(); ++i ) {
			CurVal = FromStringEx<T>( Iter->second.at(i), bSuccess );
			if( !bSuccess ) {
				std::cerr << "Warning: Config parser: value for " << Iter->first << " does not match the requested data type. Using default values" << std::endl;
				return DefaultVal;
			}
			else
				ValArray.push_back( CurVal );
		}

		return ValArray;
	}

protected:

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	bool	Read( const std::string& FileName )

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		std::string								Line;
		std::vector<std::string>	Tokens;
		unsigned int							LineCounter = 0;

		std::ifstream In( FileName );
		if( !In )
			return false;

		while( getline( In, Line ) ) {
			LineCounter++;
			Line = StripWhitespaces( Line );		// eliminate leading & trailing white-spaces

			if( Line.empty() || Line[0] == '#' || Line[0] == '*' || Line[0] == '/' )		// ignore comments and empty lines
				continue;

			Tokens = TokenizeString( Line, "= ,;\t\r\n" );
			if( Tokens.empty() || Tokens.size() == 1)		// ignore empty fields
				continue;

			if( IsString( Tokens[0] ) ) {
				ToUpper( Tokens[0] );

				if ( m_Data.find( Tokens[0] ) != m_Data.end() )
					std::cerr << "FARM config: replacing existing setting for " << Tokens[0] << " by a subsequent definition" << std::endl;

				m_Data[Tokens[0]] = std::vector<std::string>( Tokens.begin() + 1, Tokens.end() );
			}
			else {
				std::cerr << "Warning: Config parser line (" << LineCounter << "): key must be a string. Line ignored!" << std::endl;
			}
		}

		In.close();
		return true;
	};

	//////////////////////////////////////////////////////////////////////////////

	inline std::vector<std::string> TokenizeString( const std::string& Text, const std::string& Delimiters )

	// Tokenize a passed string with respect to the provided delimiters
	//
	//////////////////////////////////////////////////////////////////////////////
	{
		std::vector<std::string> WordList;
		std::string::size_type Begin, End;
		std::string Word;

		Begin = Text.find_first_not_of( Delimiters ); // skip blanks or whatever one finds at the beginning
		while( Begin != std::string::npos ) {

			End = Text.find_first_of( Delimiters, Begin );
			if( End == std::string::npos ) {     // we'v reached the end without finding another delimiter
				End = Text.length();
			}

			Word.assign( Text.begin() + Begin, Text.begin() + End );
			WordList.push_back( Word );

			Begin = Text.find_first_not_of( Delimiters, End );
		}

		return WordList;
	};

	//////////////////////////////////////////////////////////////////////////////

	inline std::string StripWhitespaces( const std::string& Input )

	//////////////////////////////////////////////////////////////////////////////
	{
		std::string::size_type Pos = 0;
		std::string NewStr( Input );

		// strip prefixing whitespaces
		Pos = NewStr.find_first_not_of( " \t" );
		if( Pos == std::string::npos )            // handle the case if only blanks & spaces were found
			return Input;

		NewStr = NewStr.substr( Pos, NewStr.length() - Pos );

		// reverse
		reverse( NewStr.begin(), NewStr.end() );

		// strip trailing (prefixing) whitespaces
		Pos = NewStr.find_first_not_of( " \t" );
		NewStr = NewStr.substr( Pos, NewStr.length() - Pos );

		// reverse
		reverse( NewStr.begin(), NewStr.end() );
		return NewStr;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	bool IsNumber( std::string x )

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		std::stringstream Stream;
		double						Number;

		Stream << x;
		Stream >> Number;

		return !Stream.fail() && Stream.eof();
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	bool IsString( std::string x)

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		return !IsNumber( x );
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	inline void ToUpper( std::string& Text )
	//
	// convert to upper case
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
#if !defined(__GNUC__) || (__GNUC__ > 4)
		std::transform( Text.begin(), Text.end(), Text.begin(), [] ( unsigned char c ) -> unsigned char { return (unsigned char)std::toupper( c ); } );
#else
		std::transform( Text.begin(), Text.end(), Text.begin(), toupper );
#endif
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	inline std::string ToLower( const std::string& Text )
	//
	// convert to lower case
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		std::string Temp( Text );

#if !defined(__GNUC__) || (__GNUC__ > 4)
		std::transform( Text.begin(), Text.end(), Temp.begin(), [] ( unsigned char c ) -> unsigned char { return (unsigned char)std::tolower( c ); } );
#else
		std::transform( Text.begin(), Text.end(), Temp.begin(), tolower );
#endif

		return Temp;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	template<typename T>
	T FromString( const std::string& s )

	// string conversion from string to typename T
	// e.g: int d = FromString<int>( s );
	// e.g: string s = ToString( d );
	//
	// if you want to check whether the conversion was successful then you can
	// use the bSuccess flag!
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		std::istringstream is( s );
		T t;
		is >> t;
		return t;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	template<typename T>
	T FromStringEx( const std::string& s, bool& bSuccess )

	// string conversion from string to typename T with a flag
	// the represents success or failure.
	//
	// e.g: int d = FromString<int>( s );
	// e.g: string s = ToString( d );
	//
	// if you want to check whether the conversion was successful then you can
	// use the bSuccess flag!
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		bSuccess = true;
		std::istringstream is( s );
		T t;
		is >> t;

		if( !is || !is.eof() )
			bSuccess = false;

		return t;
	}


protected:
	std::map<std::string, std::vector<std::string> > m_Data;		// configuration data
	bool																						 m_bStatus;
};

#endif
// vim:ts=2:sw=2
