////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#pragma once

#ifndef FARM_H
#define FARM_H

#include <string>

#include "farmsurfaceintegrator.h"
#include "voxelintegrator.h"
#include "architectureintegrator.h"
#include "graph/color.h"
#include "configparser.h"

// class D3Layer;
// class VVoxel;

class CFARM
{
public:
	CFARM();
	~CFARM();

	// basic initialization
	void Init( D3Layer* pVisLayer, VVoxel* pVoxelLayer, Geometry* pGeometry);

	// pre-draw initialization
	void PreDrawInit();

	// read the optional configuration file
	bool		ReadConfigFile( const std::string& Path );

	// check if config file has been read already
	bool ConfigRead() const { return m_bConfigRead; };

	// access surface integrator
	CSurfaceIntegrator* GetSurfaceInt() { return &m_SurfaceIntegrator; }

	// get/set render mode quality (speed vs. quality). default = speed
	int			GetRenderMode() const { return m_SurfaceIntegrator.GetRenderMode(); };
	void		SetRenderMode( int RenderMode, int QualityLevel = 0 ) { m_SurfaceIntegrator.SetRenderMode( RenderMode, QualityLevel );
																																	m_VoxelIntegrator.SetRenderMode( RenderMode, QualityLevel ); };

	// enable/disable max. intensity projection rendering mode
	void EnableMIPRender( bool bFlag = true, double MinThreshold = -1e99 ) { m_VoxelIntegrator.EnableMIPRender( bFlag, MinThreshold ); }

	// set transfer function for voxel rendering
	void SetTransferFunction( int Index ) { m_VoxelIntegrator.SetTransferFunction( Index ); }
	int  GetTransferFunction() const { return m_VoxelIntegrator.GetTransferFunction(); };


	// treat passed shading ray
	Color		Shade( GeometryEngine* pGeoEngine, Ray* pShadingRay, const int RenderPass, int Light0 );

	// treat passed shading ray
	Color		ShadeVoxel( GeometryEngine* pGeoEngine, Ray* pShadingRay, const int RenderPass, int Light0 );

	// check if the voxel hit by the passed ray is rendered (semi) transparent
	bool		IsVoxelDrawnTransparent( Ray* pShadingRay );

	// return the transparency [0,1] of a voxel associated with the shading ray
	Color		GetVoxelTransparency( Ray* pShadingRay );

	// return shading factor for data [0 = black, 1 = white]
	double	ShadeData( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 );

	// return the configuration objection
	CConfigParser& GetConfig()  { return m_ConfigParser; };

	bool		UseBumpMapForData() { return m_ConfigParser.GetValue( "usrbin-bumpmap", 0 ); };
	double	GetBumpScaling() { return m_ConfigParser.GetValue( "usrbin-bumpmap-scaling", 1.0 ); };
	int			GetBumpDir() { return m_ConfigParser.GetValue( "userbin-bumpmap-dir", 1 ); };

	int			GetResidualAASamples() const { return m_ResidualAASamples; };
	void		SetResidualAASamples( int Val ) { m_ResidualAASamples = std::min( m_MaxResidualAASamples, std::max( 0, Val )); };
	int			GetMaxResidualAASamples() const { return m_MaxResidualAASamples; };
	void		SetMaxResidualAASamples( int Val ) { m_MaxResidualAASamples = Val; };

public:
	bool	m_bNPRRendering;			// flag which indicates that NPR rendering is enabled

protected:
	CSurfaceIntegrator	m_SurfaceIntegrator;
	CVoxelIntegrator		m_VoxelIntegrator;
	CArchiIntegrator		m_ArchitecturIntegrator;
	CColoredArchiIntegrator m_ColoredArchitectureIntegrator;
	CBaseIntegrator*		m_pCurSurfaceRenderer;

	CConfigParser				m_ConfigParser;
	bool								m_bConfigRead;

	// configuration flags
	bool								m_bShadeData;

	D3Layer*						m_pVisLayer;
	VVoxel*							m_pVoxelLayer;

	int									m_ResidualAASamples;
	int									m_MaxResidualAASamples;
	bool								m_bResidualAASampling;

};

#endif
// vim:ts=2
