////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#include "farmsurfaceintegrator.h"
#include "geometry/engine.h"
#include "viewer/d3layer.h"
#include "viewer/viewer.h"
#include "math/randsampler.h"
#include "math/niederreiter.h"
#include "rendermaterial.h"
#include "os.h"

#include <string>
#include <set>

#ifndef MSWIN
	#include <unistd.h>
#endif

using namespace std;

CSurfaceIntegrator::CSurfaceIntegrator() : CBaseIntegrator(), m_AOCMinIntensity(0.3), m_AOCMaxIntensity(1.0), m_AOCMaxDist(25), m_NrOfAOCSamples(20), m_MaxAOCDepth(0), m_AOCSampleMultiplier(1),
m_bUseLightMap(false), m_pLightMap(nullptr), m_NrOfLightMapSamples(32), m_NrOfGISamples(32), m_bUseGlobalIllumination(false), m_GIScaling(1.0),
m_bBumpmapData(false), m_BumpmapScaling(1.0), m_BumpDirection(1), m_MetalBumpScaling( 1.0 ), m_BrushedBumpScaling( 1.0 ), m_bTexAvailable(false), m_MaxGIDist(20)
{
	// TEST: under development
	m_bUseLightMap = false;

	//1st attempt that does not rely on /fonts
	SetupBxDFLUT();
}

CSurfaceIntegrator::~CSurfaceIntegrator( void )
{
   for( size_t i = 0; i < m_PhysMatLUT.size(); ++i )
     delete m_PhysMatLUT[i];

   m_PhysMatLUT.clear();
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void  CSurfaceIntegrator::SetupBxDFLUT()

// Create the look-up table of all availabe BxDF's.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	string TexDir;

	for ( size_t i = 0; i < m_PhysMatLUT.size(); ++i )
		delete m_PhysMatLUT[i];
	m_PhysMatLUT.clear();

	if ( m_pRenderer ) {
		TexDir = m_pRenderer->GetConfig().GetValue<string>( "install-dir", "" );
		TexDir += "/textures/";
	}

	// phong
	m_PhysMatLUT.push_back( new CMatPhong( this ) );
	// polished plastic
	m_PhysMatLUT.push_back( new CMatPlastic( this, 0.1 ) );                   // standard roughness = 1.0 - very broad highlight

	// rough plastic 1
	m_PhysMatLUT.push_back( new CMatPlasticEx( this, new CTrowbridgeReitzDist( CTrowbridgeReitzDist::RoughnessToAlpha( 0.25 ), CTrowbridgeReitzDist::RoughnessToAlpha( 0.25 ) ), true, true, m_PlasticBumpScaling ) );             // 5 = higher roughness -> sharper highlight

	// rough plastic 2
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatPlasticEx( this, new CTrowbridgeReitzDist( CTrowbridgeReitzDist::RoughnessToAlpha( 0.9 ), CTrowbridgeReitzDist::RoughnessToAlpha( 0.9 ), false ), true, false, m_PlasticBumpScaling ) );
	else
		m_PhysMatLUT.push_back( new CMatRoughPlasticTex( this, 1.2, TexDir + "plastic1", 30, 30 ) );


	// shiny metal
	m_PhysMatLUT.push_back( new CMatShinyMetal( this, 0.1 ) );                // standard roughness = 1.0 - very broad highlight
	// shiny metal rough 1
	m_PhysMatLUT.push_back( new CMatShinyMetalEx( this, 10.f, 10.0f, 0.2f, true, false, m_MetalBumpScaling ) );	// highly rough shiny metal with displacement mapping
	// shiny metal rough 2
	m_PhysMatLUT.push_back( new CMatShinyMetalEx( this, 35.0f, 55.0f, 90.2f, true, true, m_MetalBumpScaling ) );	// highly rough shiny metal with displacement mapping

	// brushed metal
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatBrushedMetal( this, true, m_BrushedBumpScaling ) );	// brushed metal with some perlin noise
	else
		m_PhysMatLUT.push_back( new CMatBrushedMetalTex( this, TexDir + "metal1", 100, 100, m_BrushedBumpScaling ) );	// brushed metal with some perlin noise

	// brushed steel
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatShinyMetal( this, 1.5 ) );                // standard roughness = 1.0 - very broad highlight
	else
		m_PhysMatLUT.push_back( new CMatBrushedSteelTex( this, TexDir + "metal4", 200, 200, m_BrushedBumpScaling ) );

	// rusty metal
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatBrushedMetal( this, true, m_BrushedBumpScaling ) );	// brushed metal with some perlin noise
	else
		m_PhysMatLUT.push_back( new CMatRustyMetalTex( this, TexDir + "metal-rusty1", 100, 100, m_MetalBumpScaling ) );	// brushed metal with some perlin noise


	// pitted metal
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatBrushedMetal( this, true, m_MetalBumpScaling ) );	// brushed metal with some perlin noise
	else
		m_PhysMatLUT.push_back( new CMatPittedMetalTex( this, TexDir + "metal2", 100, 100, m_MetalBumpScaling ) );	// scratched metal

	// scratched metal
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatBrushedMetal( this, true, m_MetalBumpScaling ) );	// brushed metal with some perlin noise
	else
		m_PhysMatLUT.push_back( new CMatScratchedMetalTex( this, TexDir + "metal3", 100, 100, m_MetalBumpScaling ) );	// scratched metal


	// matte metal
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatBrushedMetal( this, false ) );
	else
		m_PhysMatLUT.push_back( new CMatMatteMetalTex( this, TexDir + "metal5", 150, 150, m_MetalBumpScaling ) );	// scratched metal

	// galvanized metal
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatShinyMetal( this, 1.6 ) );                // standard roughness = 1.0 - very broad highlight
	else
		m_PhysMatLUT.push_back( new CMatGalvanizedMetalTex( this, TexDir + "metal6", 100, 100, m_MetalBumpScaling ) );

	// painted metal
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatShinyMetal( this, 3.0 ) );                // standard roughness = 1.0 - very broad highlight
	else
		m_PhysMatLUT.push_back( new CMatPaintedMetal( this, TexDir + "metal7", 50, 50, m_MetalBumpScaling  ) );


	// substrate rough
	if( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatFresnelBlend( this, new CTrowbridgeReitzDist( CTrowbridgeReitzDist::RoughnessToAlpha( 0.2 ), CTrowbridgeReitzDist::RoughnessToAlpha( 0.8 ) ), false ) );              // Fresnel blended substrate with highly anisotropic microfacet distributions
	else
		m_PhysMatLUT.push_back( new CMatFresnelBlendTex( this, new CTrowbridgeReitzDist( CTrowbridgeReitzDist::RoughnessToAlpha( 0.2 ), CTrowbridgeReitzDist::RoughnessToAlpha( 0.8 ), false), false, TexDir + "plastic2", 30, 30));              // Fresnel blended substrate with highly anisotropic microfacet distributions

	// substrate polished
	m_PhysMatLUT.push_back( new CMatFresnelBlend( this, 4, 60, true ) );				// Fresnel blended substrate with perfect specular reflection

	// metallic paint
	m_PhysMatLUT.push_back( new CMatBluePaint( this ) );

	// concrete rough
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatOrenNayar( this, 17, true, true, m_ConcreteBumpScaling ) );			// rough concrete
	else
		m_PhysMatLUT.push_back( new CMatOrenNayarTex( this, TexDir + "concrete2", 100, 100, 15, m_ConcreteBumpScaling ) );			// rough concrete
		//m_BxDFLUT.push_back( new CMatOrenNayarTex( this, TexDir + "debug", 1, 1, 15, m_ConcreteBumpScaling ) );			// rough concrete

	// concrete polished
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatOrenNayar( this, 2, true, false, m_ConcreteBumpScaling ) );
	else
		m_PhysMatLUT.push_back( new CMatOrenNayarTex( this, TexDir + "concrete", 100, 100, 2, m_ConcreteBumpScaling ) );

	// concrete - with steel
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatOrenNayar( this, 2, true, false, m_ConcreteBumpScaling ) );
	else
		m_PhysMatLUT.push_back( new CMatOrenNayarTex( this, TexDir + "concrete3", 100, 100, 7, m_ConcreteBumpScaling ) );

	// concrete dirty
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatOrenNayar( this, 2, true, false, m_ConcreteBumpScaling ) );
	else
		m_PhysMatLUT.push_back( new CMatOrenNayarTex( this, TexDir + "concrete4", 100, 100, 7, m_ConcreteBumpScaling ) );

	// concrete poured
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatOrenNayar( this, 2, true, false, m_ConcreteBumpScaling ) );
	else
		m_PhysMatLUT.push_back( new CMatOrenNayarTex( this, TexDir + "concrete5", 250, 250, 3, m_ConcreteBumpScaling ) );


	// minnaert
	m_PhysMatLUT.push_back( new CMatMinnaert( this ) );

	// marble
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatPlasticEx( this, new CTrowbridgeReitzDist( CTrowbridgeReitzDist::RoughnessToAlpha( 0.1 ), CTrowbridgeReitzDist::RoughnessToAlpha( 0.05 ), false ), true, true, m_PlasticBumpScaling ) );             // higher roughness -> sharper highlight
	else
		m_PhysMatLUT.push_back( new CMatMarble( this, TexDir + "marble1", 100, 100 ) );

	// marble 2
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatPlasticEx( this, new CTrowbridgeReitzDist( CTrowbridgeReitzDist::RoughnessToAlpha( 6 ), CTrowbridgeReitzDist::RoughnessToAlpha( 3.7 ) ), true, true, m_PlasticBumpScaling ) );
	else
		m_PhysMatLUT.push_back( new CMatMarble( this, TexDir + "marble2", 100, 100 ) );

	// soil
	if ( !m_bTexAvailable )
		m_PhysMatLUT.push_back( new CMatOrenNayar( this, 27, true, true, m_ConcreteBumpScaling ) );
	else
		m_PhysMatLUT.push_back( new CMatSoilTex( this, TexDir + "soil1", 300, 300) );


	// toon shader
	m_PhysMatLUT.push_back( new CMatToon( this ) );
	// translucent
	m_PhysMatLUT.push_back( new CMatTranslucent( this ) );
	// refractive glass
	m_PhysMatLUT.push_back( new CMatGlass( this, 1.40f ) );
	// clear glass / plexi - glass
	m_PhysMatLUT.push_back( new CMatClearGlass( this ) );
	// dispersive glass
	m_PhysMatLUT.push_back( new CMatDispersiveClearGlass( this, 0.005) );
	// highly dispersive glass
	m_PhysMatLUT.push_back( new CMatDispersiveClearGlass( this, 0.01 ) );


	//m_BxDFLUT.push_back( new CMatFresnelBlend( this, 4, 50) );              // Fresnel blended substrate with highly anisotropic microfacet distributions

	/*
	m_BxDFLUT.push_back( new CMatBlinn( this ) );
	m_BxDFLUT.push_back( new CMatOrenNayar( this ) );
	// m_BxDFLUT.push_back( new CMatSkin( this ) );
	m_BxDFLUT.push_back( new CMatOrenNayar( this, 0.3 ) );			// sigma of 3 degrees --> concrete
  m_BxDFLUT.push_back( new CMatClay( this ) );
  m_BxDFLUT.push_back( new CMatFelt( this ) );
  m_BxDFLUT.push_back( new CMatPrimer( this ) );
  */
  m_pPhysMat = m_PhysMatLUT[0];
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CSurfaceIntegrator::Li( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0, int CallerId /*= cStandard | cFirstCall for first call*/)

// main integrator routine returning the luminance for the passed ray & rendering pass
// [4/27/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Point	 Hit;
	const	 Material* pMaterial = nullptr;
	double ambiOcc = 1.0;
	Color	 ShadingColor;
	Ray		 lightRay;
	int		 BxDFIndex = -1;

	// check next intersection on secondary ray generations
	if ( pRay->depth > 0 && !m_pVisLayer->nextIntersection( pEngine, pRay ) ) {
		if ( m_bUseLightMap && m_RenderMode == CSurfaceIntegrator::cQuality ) {							// take care of env. reflection
			return m_pLightMap->LightEmission( pRay->dir() ) * 1.0;
		}
		else {
			return ShadingColor;	// on end of ray, error or blackbody
		}
	}

	// bail out of GI final gathering for distances that are too far to contribute significantly
	if ( CallerId == cGlobalIllumination && pRay->T() >= m_MaxGIDist ) {
		return ShadingColor;
	}

	// check if we have to deal with a projection of a data mesh instead of real geometry
	if ( pRay->project_insight ) {
		while ( true ) {

			// prevent reflection of data as it looks a bit strange and can lead to confusion even though it would physically be correct
			// TODO: prevent reflections by checking the ray generation??
			if ( pRay->project_insight && pEngine->applyProject( pRay, pRay->T() ) ) {
				if ( CallerId == cReflection || CallerId == cGlobalIllumination )
					return Color::Black;

				if ( !m_bBumpmapData )
					pRay->project_alpha = m_pVisLayer->viewer.usrbin.shade( pRay->hit(), ShadingColor );
				else
					pRay->project_alpha = m_pVisLayer->viewer.usrbin.bumpShade( *pRay, m_pVisLayer->light[Light0], m_BumpmapScaling, m_BumpDirection, ShadingColor );	// bump map test

				if ( pRay->project_alpha != 0xFF && m_pVisLayer->viewer.usrbin.show ) {
					return ShadingColor;
				}
				else { // we hit a full transparent part of the USRBIN data. so let's continue tracing
					pRay->moveby( SMALL3D3 );
					pRay->project_hit = false;

					if ( !m_pVisLayer->nextIntersection( pEngine, pRay ) ) {
						if ( m_bUseLightMap && pRay->depth > 0 ) {	// take care of env. reflection
							return m_pLightMap->LightEmission( pRay->dir() ) * 1.0;
						}
						else {
							return ShadingColor;	// on end of ray, error or blackbody
						}
					}

					continue;
				}
			}

			break;
		} // end of while
	}

	// Get region/voxel color
		// for transmission rays within an object we need to get the material info from the previous region and not the next one to properly re-calculate 
		// the exit direction of the transmisison ray
	pMaterial = pRay->hitMaterial();

	if( pRay->voxelreg > 0 ) {
		ShadingColor.set( m_pVisLayer->viewer.voxel().roiColor( pRay->voxelreg ) );

		if ( m_RenderMode == CBaseIntegrator::cSpeed ) {
			Color totalIntensity = ShadeHitPoint( pEngine, pRay, pMaterial, RenderPass, Light0 );
			float ambient = (float)m_pVisLayer->ambient / 255.0f;
			ShadingColor *= totalIntensity + Color( ambient, ambient, ambient);
		}

		return ShadingColor;		// return to FARM which, based on the selected rendering quality, it will trigger the voxel renderer to obtain full volume rendering effects or not
	}
	else {
		if ( pMaterial ) {
			BxDFIndex = m_RenderMode == CSurfaceIntegrator::cQuality ? pMaterial->bxdf() : 0;
			ShadingColor = m_PhysMatLUT[BxDFIndex]->GetHitColor( pRay );	// get base hit color
		}
	}

		// If we are in the start of the plane do not apply any shadow
	if( pRay->start() ) return ShadingColor;

	//
	// calculate light intensity
	//
	Color totalIntensity = ShadeHitPoint( pEngine, pRay, pMaterial, RenderPass, Light0 );
	float brightening = m_RenderMode == CBaseIntegrator::cSpeed ? 1.0f / 255.0f : 2.0f / 255.0f;
	brightening *= (float)m_pVisLayer->ambient;
	ShadingColor *= totalIntensity + Color( brightening, brightening, brightening);
	if(CallerId == cGlobalIllumination)
		return ShadingColor;

#if 0
	int log2depth;
	frexp( hit.z, &log2depth );
#endif

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Environmental infinite light maps
	//
	if( m_bUseLightMap && pMaterial && m_RenderMode == CSurfaceIntegrator::cQuality ) {
		ShadingColor += Li_LightMap( pEngine, pRay);
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// global illumination (only for primary rays)
	//
	if(m_bUseGlobalIllumination && pMaterial && pRay->depth == 0 /* && pRay->depth < pRay->max_depth*/ &&
		 m_RenderMode == CSurfaceIntegrator::cQuality) {	// to avoid noise and decrease CPU time we only do final gathering for primary rays
		ShadingColor += Li_GlobalIllumination( pEngine, pRay, RenderPass, Light0 ) * m_GIScaling;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Reflection & refraction
	//

	//if(pRay->depth < pRay->max_depth &&
	//	pMaterial != nullptr) {
	if(pRay->depth <= pRay->max_depth &&		// We now allow going to 1 sub-generation even in speed mode in order to
		pMaterial != nullptr) {								// be able to treat data projections in transparent objects.

		// reflection
		if(pMaterial->specular() > 0.0 && m_PhysMatLUT[BxDFIndex]->MatchesType( CBxDFBase::BSDF_REFLECTION )) {
			int oldBodyId = pEngine->bodyCheckId();		// Transmission requires spawning secondary rays and thus, we need to advance the ray cache
			ShadingColor += Li_Reflection( pEngine, pRay, RenderPass, Light0 );
			pEngine->bodyCheckId( oldBodyId );	// restore the ray cache to ensure that there is no conflict with subsequent calls to obtain screen door transparency via d3layer::ShadeXRay()
		}

		// refraction & transmission
		if(!pRay->hitRegion()->notTransparent()) {

			// check if the BxDF supports transparency. in case a projection is in sight we also support (non-physical)
			// transparency for non transmissive BxDFs to avoid confusion because the user might have enabled
			// transparency per region and would expect to see through in any case.
			if(m_PhysMatLUT[BxDFIndex]->MatchesType( CBxDFBase::BSDF_TRANSMISSION ) || pRay->project_insight ) {

				int oldBodyId = pEngine->bodyCheckId();		// Transmission requires spawning secondary rays and thus, we need to advance the ray cache
				Color TmpColor = Li_Transmission(pEngine, pRay, RenderPass, Light0);
				if(pRay->project_hit) {

					// for data we use blending instead of addition as we would do in normal ray tracing. this way we remain closer to the original color shades of the data
					ShadingColor.blend( TmpColor, 1.0 - pRay->hitRegion()->alpha() / 255.0 );
					//ShadingColor = TmpColor;

					pRay->project_hit = false;		// this has been treated by the renderer already and does not require further action like the std. transparency by
																				// blending via the d3layer. So we reset the hit flag for data projections.
				}
				else {
					ShadingColor += TmpColor;
				}

				pEngine->bodyCheckId(oldBodyId);	// restore the ray cache to ensure that there is no conflict with subsequent calls to obtain screen door transparency via d3layer::ShadeXRay()
			}
		}
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Ambient occlusion shading
	// CT 10/7/19
	if(m_pVisLayer->shadows and m_NrOfAOCSamples > 0 &&
		pRay->depth <= m_MaxAOCDepth) {
		// NOTE: To avoid artifacts the AOC distance should be reasonably small
		// (configurable?), false = no AOC sample scrambling
		ambiOcc = CalcAmbientOcclusion( pEngine, pRay,
			m_NrOfAOCSamples * m_AOCSampleMultiplier,
			m_AOCMaxDist,
			m_AOCMinIntensity,
			m_AOCMaxIntensity);

		// ShadingColor.set(ambiOcc, ambiOcc, ambiOcc); // debug to get only AOC pass
		ShadingColor *= ambiOcc;
	}


	ShadingColor.clamp();

	return ShadingColor;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CSurfaceIntegrator::CalcAmbientOcclusion( GeometryEngine* eng, Ray* pixelRay,
																							 int nrOfSamples, const double cMaxDist, const double cMinIntensity,
																							 const double cMaxIntensity )


//	New ambient occlusion integrator
// [12/12/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Point    hit;
	Vector   sampleDir, bodyNormal;
	Ray      ambiOccRay;
	double   aocFactor = 0.0, transparency = 0.0;
	double   randNr1, randNr2;

	// Find body normal
	hit = pixelRay->hit( -SMALL3D4 );
	bodyNormal = pixelRay->normal;

	// check that the normal is opposite to the direction
	if ( bodyNormal * pixelRay->dir() > 0.0 )
		bodyNormal.negate();

	{
		// inherit and propagate general ray a
		ambiOccRay.use_clip = pixelRay->use_clip;
		ambiOccRay.use_project = pixelRay->use_project;

		bool bStep = eng->projectBodyCount() > 0 ? true : false;
		int ScrambleBxDF = eng->random.integer( cPMJLUTSize );

		for ( int i = 0; i < nrOfSamples; i++ ) {
			CBaseSampler::getPMJ02Sample( i, randNr1, randNr2, ScrambleBxDF );
			sampleDir = CBaseSampler::uniformSampleHemisphere( bodyNormal, randNr1, randNr2 );

			ambiOccRay.init();
			RaySegment new_segment( hit, sampleDir, pixelRay->prevZone() );
			new_segment.tmin = SMALL3D3;
			new_segment.tmax = cMaxDist * 2.0;
			ambiOccRay.push( new_segment );

			int oldBodyCheckId = eng->incBodyCheckId();

			if ( !eng->intersectRay( &ambiOccRay, bStep ) ) {		// NOTE: if we pass false we would ignore projection planes which can create artifacts, CT 18/8/22
																													// return false = intersection occurred!
				VZone* pZone = ambiOccRay.hitZone();
				if ( pZone != nullptr &&
						 (pZone->region()->type() == GRegion::Type::NORMAL ||
						 pZone->region()->type() == GRegion::Type::VOXEL) ) {
					eng->bodyCheckId( oldBodyCheckId );
					if ( ambiOccRay.T() <= cMaxDist ) {
						// ATTENTION: the hitRegion->transparent() function only returns true for FULLY transparent regions --> we need to check alpha!!
						transparency = ambiOccRay.hitRegion()->alpha() / 255.0;
						aocFactor += (1.0 - transparency) * sampleDir * bodyNormal;		// take transparency into account when calculating AOC
						continue;
					}
				}
			}

			eng->bodyCheckId( oldBodyCheckId );
			ambiOccRay.pop();
		}
	} // end of sequential

	aocFactor = (1.0 - aocFactor / ((double)nrOfSamples)) /** 1.0 / M_PI*/; // omitting normalization 1/pi is non-physical but avoids that this 
																																						// becomes too dark w.r.t to the speed mode

	return cMinIntensity + (cMaxIntensity - cMinIntensity) * aocFactor;

} // calcAmbientOcclusion


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CSurfaceIntegrator::Li_Reflection( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// [4/28/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color						ShadingColor, TmpColor, Rho;
	const Material* pMaterial = pRay->hitMaterial();
	Vector					R, ReflNormal1, ReflNormal2;
	Ray							ReflRay;
	float						Pdf = 0.0;
	double					Dot = 0.0;


	const int				cMaxScatteringSamples = 256;
	double					pRandNr[2];
	double					pScatteringSamples[cMaxScatteringSamples];

	int BxDFIndex = m_RenderMode == CSurfaceIntegrator::cQuality ? pMaterial->bxdf() : 0;
	int NrOfSamples = (m_PhysMatLUT[BxDFIndex]->RequiresMCSampling() || pMaterial->fuzz() > 0.0) ? std::min( cMaxScatteringSamples, GetNrOfMCSamples() ) : 1;   // restrict the nr. of samples to 256. otherwise we might end up with extreme rendering times

	int ScrambleBxDF = pEngine->random.integer( cPMJLUTSize );

	pEngine->random.realqmcND( min( NrOfSamples * 2, DIM_MAX), pScatteringSamples );

	// disable clipping bodies for reflection rays, to avoid visualization artifacts of background bleeding
	// ReflRay.use_clip = ReflRay.clip_insight = false;
	ReflRay.use_clip = pRay->use_clip; ReflRay.clip_insight = pRay->clip_insight;
	ReflRay.use_project = ReflRay.project_insight = false;

	Vector ShadingNormal = m_PhysMatLUT[BxDFIndex]->GetShadingNormal( pRay );

	for( int i = 0; i < NrOfSamples; ++i ) {

		CBaseSampler::getPMJ02Sample( i, pRandNr[0], pRandNr[1], ScrambleBxDF );

		// sample reflection BxDF to obtain direction of incoming light + the corresponding rho
		Rho = m_PhysMatLUT[BxDFIndex]->Sample_f( *pRay, ReflRay, Pdf, ShadingNormal, *pMaterial, pRandNr, CBxDFBase::BSDF_REFLECTION );
		if( Rho <= Color::Black )
			continue;

		// handle diffuse reflection if requested
		if( pMaterial->fuzz() > 0 && m_PhysMatLUT[BxDFIndex]->MatchesType( CBxDFBase::BSDF_GLOSSY) ) {
			Vector distortedDir = CBaseSampler::uniformSampleCone( ReflRay.dir(), std::min( pMaterial->fuzz() / 10.0, 1.0 ), pScatteringSamples[i*2], pScatteringSamples[i*2+1]);
			ReflRay.segment().dir = distortedDir;
		}

		// calculate incoming reflected luminance
		Dot = fabs( ShadingNormal.dot( ReflRay.dir() ) );
		if( Dot > 0 && Pdf > 0.0 ) {
			int oldId = pEngine->incBodyCheckId();		// signal to cache that we have a new ray direction
			ShadingColor += (Li( pEngine, &ReflRay, RenderPass, Light0, cReflection ) * pMaterial->specular()) * Rho * Dot / Pdf;	// NOTE: for MC sampled rays the PDF should include the factor 1/Pi already!
			pEngine->bodyCheckId( oldId );
		}
	}

	// normalize
	ShadingColor /= (NrOfSamples * 1.0);      // calculate the total reflected contribution

	return ShadingColor;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CSurfaceIntegrator::Li_Transmission( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// [5/4/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color						RefractionColor, TransparencyColor( 1.0, 1.0, 1.0), Rho;
	const Material* pMaterial = pRay->hitMaterial();
	Vector					R, ReflNormal1, ReflNormal2;
	Ray							RefractionRay;
	float						Pdf = 0.0;
	double					Dot = 0.0;

	const int				cMaxScatteringSamples = 256;
	double					pRandNr[2];
	double					pScatteringSamples[cMaxScatteringSamples];

	int BxDFIndex = m_RenderMode == CSurfaceIntegrator::cQuality ? pMaterial->bxdf() : 0;
	int NrOfSamples = (m_PhysMatLUT[BxDFIndex]->RequiresMCSampling() || pMaterial->fuzz() > 0.0) ? std::min( cMaxScatteringSamples, GetNrOfMCSamples() ) : 1;   // restrict the nr. of samples to 32. otherwise we might end up with extreme rendering times

	int ScrambleBxDF = pEngine->random.integer( cPMJLUTSize );

	pEngine->random.realqmcND( NrOfSamples * 2, pScatteringSamples );

	// disable clipping bodies for refraction rays, to avoid visualization artifacts of background bleeding
	RefractionRay.use_clip = RefractionRay.clip_insight = false;

	Vector ShadingNormal = m_PhysMatLUT[BxDFIndex]->GetShadingNormal( pRay );


	for( int i = 0; i < NrOfSamples; ++i ) {

		CBaseSampler::getPMJ02Sample( i, pRandNr[0], pRandNr[1], ScrambleBxDF );

		// sample refraction BxDF to obtain direction of incoming light + the corresponding rho
		Rho = m_PhysMatLUT[BxDFIndex]->Sample_f( *pRay, RefractionRay, Pdf, ShadingNormal, *pMaterial, pRandNr, CBxDFBase::BxDFType::BSDF_TRANSMISSION );
		if ( Rho <= Color::Black )
			continue;

		RefractionRay.use_project = pRay->use_project;


		// handle diffuse transmission if requested
		if( pMaterial->fuzz() && !m_PhysMatLUT[BxDFIndex]->MatchesType( CBxDFBase::BxDFType::BSDF_TRANSLUCENT ) ) {
			Vector distortedDir = CBaseSampler::uniformSampleCone( RefractionRay.dir(), std::min( pMaterial->fuzz() / 10.0, 1.0 ), pScatteringSamples[i * 2], pScatteringSamples[i * 2 + 1] );
			RefractionRay.segment().dir = distortedDir;
		}

		// calculate incoming reflected luminance
		Dot = -ShadingNormal.dot( RefractionRay.dir() );
		if( Dot > 0.0 && Pdf > 0.0 ) {		
			int oldId = pEngine->incBodyCheckId();		// signal to cache that we have a new ray direction
			RefractionColor += (Li( pEngine, &RefractionRay, RenderPass, Light0, cTransmission ) * fabs(Dot) / Pdf) * Rho ;	// NOTE: for MC sampled rays the PDF should include the factor 1/Pi already!

			// handle potential intersection with data mesh
			pRay->project_hit |= RefractionRay.project_hit;
			if(RefractionRay.project_hit) {
				pRay->project_alpha = max( pRay->project_alpha, RefractionRay.project_alpha );	// propagate back alpha channel to camera ray
			}

			pEngine->bodyCheckId( oldId );	// restore cache

		}
	}

	// normalize
	RefractionColor /= (NrOfSamples * 1.0);      // calculate the total reflected contribution

	// data projections are not subject to any other physical effects like Beer's optical absorption etc.
	if ( pRay->project_hit ) {
		return RefractionColor * (1.0 - pMaterial->specular());	// NOTE: we might want to multiply this with the (pRay->project_alpha / 255.0) for correct transparency, otherwise we boost the colors
	}

	//
	// apply Beer's law
	//
	if( pRay->hitRegion()->alpha() != 0 ) {
		double MatDensity = -0.005 * (pRay->T());       // TODO: Get this parameter from the material
		Color Absorbance( pRay->hitRegion()->color() );
		Absorbance *= MatDensity;
		TransparencyColor.set( expf( Absorbance.red() ), expf( Absorbance.green() ), expf( Absorbance.blue() ) );
	}

	return RefractionColor * TransparencyColor * ( pRay->hitRegion()->alpha() / 255.0 ) * ( 1.0 - pMaterial->specular() );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CSurfaceIntegrator::Li_LightMap( GeometryEngine* pEngine, Ray* pRay )

// Return the incoming irradiance from an environmental lighmap
//
// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color		ShadingColor, CurLightCol;
	int			NrOfSamples = std::min( 256, m_NrOfLightMapSamples );   // restrict the nr. of samples to 256. otherwise we might end up with extreme rendering times
	Color		Rho( Color::White );
	float   Transmission = 1.0;
	Vector	LightDir;
	Ray			LightRay;

	const Material* pMaterial = pRay->hitMaterial();
	int							BxDFIndex = m_RenderMode == CSurfaceIntegrator::cQuality ? pMaterial->bxdf() : 0;

	Vector  ShadingNormal = m_PhysMatLUT[BxDFIndex]->GetShadingNormal( pRay );

	int oldBodyCheckId = pEngine->incBodyCheckId();

	// sample light map
	for( int i = 0; i < NrOfSamples; ++i ) {

		// TODO: replace by biased lightmap
		// sample incoming irradiance
		auto oldId = pEngine->incBodyCheckId();
		CurLightCol	= m_pLightMap->SampleLightMap( pEngine, i, *pRay, ShadingNormal, LightRay, Transmission );
		if( CurLightCol != Color::Black ) {
			// calculate color
			Rho = m_PhysMatLUT[BxDFIndex]->f( *pRay, LightRay, ShadingNormal, *pMaterial, CurLightCol, 1.0 );
			double Dot = std::max( 0.0, ShadingNormal.dot( LightRay.dir() ) );
			ShadingColor += CurLightCol * Rho * Dot * Transmission;
		}
		pEngine->bodyCheckId( oldId  );
	}

	pEngine->bodyCheckId(oldBodyCheckId);

	// normalize
	ShadingColor /= (NrOfSamples * 1.0);

	return ShadingColor;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CSurfaceIntegrator::InitFromConfig()

// [8/19/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	assert( m_pRenderer );	// ensure the base-class Init() has been called before. If this triggers, check if this renderer has been registered & initialized via the Init()
													// function in FARM
	string FileName;

	// show config flags
	if( !m_pRenderer->GetConfig().Empty() ) {
#if _DEBUG>1
		m_pRenderer->GetConfig().Show();
#endif
  }

	// set up light maps
	m_bUseLightMap = m_pRenderer->GetConfig().GetValue<int>( "lightmap", 0 );
	if( m_bUseLightMap && !m_pLightMap ) {

		m_pLightMap = new CBiasedLightMap;
		FileName = m_pRenderer->GetConfig().GetValue<string>( "lightmap_file", "" );

		if( !m_pLightMap->LoadMap( FileName, m_pGeometry) ) {
			std::cerr << "FARM ERROR: no image light found" << std::endl;
			m_bUseLightMap = false;
		}

		m_pLightMap->SetDistance( m_pRenderer->GetConfig().GetValue<double>( "lightmap_distance", 1.0 ) );		// NOTE: larger distances will blur out detailed reflections
		std::cout << "FARM: Lightmap initialized: " << FileName << " - " << m_pLightMap->GetDistance() << std::endl;

		// set up projection
		std::string LightMapProjection = m_pRenderer->GetConfig().GetValue<string>( "lightmap_projection", "xy" );
		if( LightMapProjection == "yz" )
			m_pLightMap->SetUVProjection( Vector( 0, 1, 0 ), Vector( 0, 0, 1 ) );
		else if( LightMapProjection == "zy" )
			m_pLightMap->SetUVProjection( Vector( 0, 0, 1 ), Vector( 0, 1, 0 ) );
		else if( LightMapProjection == "zx" )
			m_pLightMap->SetUVProjection( Vector( 0, 0, 1 ), Vector( 1, 0, 0 ) );
		else if( LightMapProjection == "xz" )
			m_pLightMap->SetUVProjection( Vector( 1, 0, 0 ), Vector( 0, 0, 1 ) );
		else if( LightMapProjection == "yx" )
			m_pLightMap->SetUVProjection( Vector( 0, 1, 0 ), Vector( 1, 0, 0 ) );
		else // default = xy
			m_pLightMap->SetUVProjection( Vector( 1, 0, 0 ), Vector( 0, 1, 0 ) );

		// set lightmap intensity
		m_pLightMap->SetIntensity( m_pRenderer->GetConfig().GetValue<double>( "lightmap_intensity", 1.0 ) );

		m_pLightMap->EnableShadows( m_pRenderer->GetConfig().GetValue<int>( "lightmap_shadows", 1 ) );
		std::cout << "Lightmap shadows: " << m_pLightMap->AreShadowsEnabled() << std::endl;
	}

	// set up nr. of samples
	m_NrOfMCSamples				= m_pRenderer->GetConfig().GetValue<int>( "mc_samples", 8 );
	m_NrOfLightMapSamples = m_pRenderer->GetConfig().GetValue<int>( "lightmap_samples", 6 );

	// global illumination
	m_NrOfGISamples				   = m_pRenderer->GetConfig().GetValue<int>( "gi_samples", 32 );
	m_bUseGlobalIllumination = m_pRenderer->GetConfig().GetValue<int>( "global_illumination", 0 );	// off by default as it is very costly
	m_GIScaling							 = m_pRenderer->GetConfig().GetValue<double>("gi_scaling", 1.0);
	m_MaxGIDist							 = m_pRenderer->GetConfig().GetValue<double>( "gi_maxdist", 20.0 );
	if(m_bUseGlobalIllumination) {
		std::cout << "####################################################################################################################" << std::endl;
		std::cout << "#																									F A R M																													 #" << std::endl;
		std::cout << "# !! ATTENTION !! global illumination activated -> increased rendering times. # of samples = " << m_NrOfGISamples << std::endl;
		std::cout << "# GI max dist = " << m_MaxGIDist << std::endl;
		std::cout << "####################################################################################################################" << std::endl;
	}

	// set up ambient occlusion
	m_AOCMinIntensity = m_pRenderer->GetConfig().GetValue<double>( "aoc_min", 0.3 );
	m_AOCMaxIntensity = m_pRenderer->GetConfig().GetValue<double>( "aoc_max", 1.0 );
	m_AOCMaxDist = m_pRenderer->GetConfig().GetValue<double>( "aoc_dist", 25.0 );
	m_MaxAOCDepth = m_pRenderer->GetConfig().GetValue<int>( "aoc_maxdepth", 0 );
	m_AOCSampleMultiplier = m_pRenderer->GetConfig().GetValue<int>( "aoc_sample_multiplier", 1 );

	// bump map rendering of data
	m_bBumpmapData = m_pRenderer->GetConfig().GetValue( "usrbin-bumpmap", 0 );
	m_BumpmapScaling = m_pRenderer->GetConfig().GetValue( "usrbin-bumpmap-scaling", 1.0 );
	m_BumpDirection = m_pRenderer->GetConfig().GetValue( "usrbin-bumpmap-dir", 1 );		// 1 = positive, -1 = negative


	// noise scaling factors
	m_MetalBumpScaling		= m_pRenderer->GetConfig().GetValue( "metal-bump-scaling", 1.0 );
	m_PlasticBumpScaling  = m_pRenderer->GetConfig().GetValue( "plastic-bump-scaling", 1.0 );
	m_BrushedBumpScaling  = m_pRenderer->GetConfig().GetValue( "brushed-bump-scaling", 1.0 );
	m_ConcreteBumpScaling = m_pRenderer->GetConfig().GetValue( "concrete-bump-scaling", 1.0 );

	// enable/disable texture mapping based on the availability of the texture data - WORK IN PROGRESS
	string InstallDir = m_pRenderer->GetConfig().GetValue<string>( "install-dir", "" );
	if ( InstallDir != "" ) {
		InstallDir += string( "/textures" );
		if ( ACCESS( InstallDir.c_str(), 0) != -1 )
		m_bTexAvailable = true;
		else
			m_bTexAvailable = false;
	}
	if ( m_bTexAvailable ) {	// TODO: remove in the future
		std::cout << "FARM: -- macroscopic material visualization enabled" << std::endl;
	}

	// create materials (2nd attempt that considers textures etc. if available)
	SetupBxDFLUT();

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CSurfaceIntegrator::ShadeHitPoint( GeometryEngine* pEngine, Ray* pRay, const Material* pMaterial, const int RenderPass, int Light0 )

// [8/19/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( RenderPass );

	Point  Hit;
	double randNr[2];
	Ray    lightRay;
	Color  Rho( Color::White);
	double pdf = 1.0, GeoTMin = 0.0;
	int		 BxDFIndex = 0;
	Vector ShadingNormal, GeoNormal, GeoHit;

	UNUSED( GeoTMin );

	// All light rays should start from outside the object
	// Move a bit backwards
	// Should be bigger than what is inside distance

	GeoHit = pRay->hit();
	if ( pMaterial ) {		
		BxDFIndex = m_RenderMode == CSurfaceIntegrator::cQuality ? pMaterial->bxdf() : 0;
		GeoNormal = pRay->normal;
		GeoTMin = pRay->segment().tmin;

 		ShadingNormal = m_PhysMatLUT[BxDFIndex]->GetShadingNormal( pRay );
 		m_PhysMatLUT[BxDFIndex]->ModifyShadingGeometry( pRay, pEngine );			// potentially modify the hit point based on the BxDF to account for displacements etc.
	}
	Hit = pRay->hit( -SMALL3D3 );


	// Calculate illumination
	Color totalIntensity(Color::Black), curIntensity(Color::Black);
	Vector dir, curLPos;

	// ignore artificial clipping & projection bodies to avoid vis artifacts
	//lightRay.use_clip = false;
	lightRay.use_clip = pRay->use_clip;
	lightRay.use_project = false;

	int Scramble1 = pEngine->random.integer();
	int Scramble2 = pEngine->random.integer();

	for( int i = 0; i < Min( pRay->nlights, m_pGeometry->nlights ); i++ ) {
		const Light& L = m_pGeometry->light[Light0 + i];	// Geometry Light properties
		const VLight& VL = m_pVisLayer->light[i];				// Viewer light coordinate & direction
		double  dist, power, penumbraScaling = 0.0;

		const int nsamples = L.nsamples;
		//assert(InRange(0,nsamples,100));
		curIntensity = Color::Black;

		// volumetric lights
		for( int lightSample = 0; lightSample < nsamples; ++lightSample ) {
			switch( L.type ) {
				case Light::Type::Sun:
					dir = -VL.dir;
					dist = INFINITE;
					power = L.power;
					if( L.shadow && pRay->shadow ) {
						double dot = m_pVisLayer->view().matrix( 0, 2 ) * dir.x +
							m_pVisLayer->view().matrix( 1, 2 ) * dir.y +
							m_pVisLayer->view().matrix( 2, 2 ) * dir.z;
						if( dot > SMALL3D3 && m_pVisLayer->view().projection != ViewPort::Projection::Perspective ) {
							dist = (m_pVisLayer->view().matrix( 0, 2 ) * (m_pVisLayer->view().matrix( 0, 3 ) - Hit.x) +
											 m_pVisLayer->view().matrix( 1, 2 ) * (m_pVisLayer->view().matrix( 1, 3 ) - Hit.y) +
											 m_pVisLayer->view().matrix( 2, 2 ) * (m_pVisLayer->view().matrix( 2, 3 ) - Hit.z)) /
								dot;
							if( dist < SMALL3D3 ) continue;
						}
					}
					break;

				case Light::Type::Omni:
				case Light::Type::Spot:
					// move the light position for volumetric light sources
					curLPos = VL.pos;
					if( lightSample > 0 ) {
						// we found that the (0,2) series by VDC & Sobol2 aggregation shows highest volumetric convergence
						// at relatively low sample rates. Thus, we stick to this one here instead of Niederreiter or PMJ02.
						randNr[0] = CBaseSampler::VanDerCorput( lightSample, Scramble1 );
						randNr[1] = CBaseSampler::Sobol2( lightSample, Scramble2 );

						dir = (VL.pos - Hit);
						dir.normalize();
						curLPos += CBaseSampler::uniformSampleDisk( dir, L.radius, randNr[0], randNr[1] );
					}
					dir = curLPos - Hit;
					dist = dir.normalize();

					if( dist < SMALL3D3 ) continue;
					if( dist > L.distance ) continue;
					if( L.falloff != Light::FallOff::Constant ) {
						// FIXME could be converted to switch ()
						if( L.falloff == Light::FallOff::Inverse )
							power = L.power * 100.0 / dist;
						else	// must be Light::FallOff::Inverse2
							power = L.power * 10000.0 / Sq( dist );
						if( power < 0.001 ) continue;
					}
					else
						power = L.power;

					if( L.shadow && pRay->shadow ) {
						double dot = m_pVisLayer->view().matrix( 0, 2 ) * dir.x +
							m_pVisLayer->view().matrix( 1, 2 ) * dir.y +
							m_pVisLayer->view().matrix( 2, 2 ) * dir.z;
						if( dot > SMALL3D3 ) {
							double pdist = (m_pVisLayer->view().matrix( 0, 2 ) * (m_pVisLayer->view().matrix( 0, 3 ) - Hit.x) +
															 m_pVisLayer->view().matrix( 1, 2 ) * (m_pVisLayer->view().matrix( 1, 3 ) - Hit.y) +
															 m_pVisLayer->view().matrix( 2, 2 ) * (m_pVisLayer->view().matrix( 2, 3 ) - Hit.z)) /
								dot;
							dist = Min( dist, pdist );
							if( dist < SMALL3D3 ) continue;
						}
					}
					power /= (double)nsamples;   // normalize the light power w.r.t. light samples
					break;

				default:
					continue;
			}

			// set up light ray & find intersection with light
			lightRay.init();

			RaySegment new_segment( Hit, dir, pRay->prevZone() );

			lightRay.push( new_segment );

			//double dot = dir * GeoNormal;
			double dot = dir * ShadingNormal;
			if ( dot < 0.00001 ) continue;

			if( L.shadow && pRay->shadow ) {
				int oldBodyCheckId = pEngine->incBodyCheckId();

				if( !pEngine->intersectRay( &lightRay, false ) ) {
					// Check if we hit a black hole or an error
					VZone* z = lightRay.hitZone();
					if( z != nullptr ) {
						if( (z->region()->type() == GRegion::Type::NORMAL ||
									z->region()->type() == GRegion::Type::VOXEL) ) {
							pEngine->bodyCheckId( oldBodyCheckId );

							// account for "light shadows" due to transparent objects
							double Transparency = z->region()->alpha() / 255.0;
							penumbraScaling += Transparency / (nsamples);		// take transparency into account when calculating AOC
							continue;
						}
						else
							if( z->region()->type() == GRegion::Type::BLACKHOLE )
								penumbraScaling = 1.0;
					}

////////////

				/* WIP: -- multi-step tracking for semi-opaque objects that cast light shadows
				if( !pEngine->intersectRay( &lightRay, true ) ) {
					// Check if we hit a black hole or an error
					VZone* z = lightRay.hitZone();
					if( z != nullptr ) {
						if( (z->region()->type() == GRegion::Type::NORMAL ||
									z->region()->type() == GRegion::Type::VOXEL) ) {

							// account for "light shadows" due to transparent objects
							double Transparency = z->region()->alpha() / 255.0;

							if( Transparency > 0.0 && Transparency < 1.0 ) {
								lightRay.moveby( SMALL3D4 * 10.0 );
								if( !pEngine->intersectRay( &lightRay, false ) ) {
									if( (z->region()->type() == GRegion::Type::NORMAL ||
												z->region()->type() == GRegion::Type::VOXEL) &&
											z->region()->type() != GRegion::Type::BLACKHOLE ) {
										penumbraScaling = 0.0;
									}
								}
							}

							pEngine->bodyCheckId( oldBodyCheckId );
							penumbraScaling += Transparency / (nsamples);		// take transparency into account when calculating AOC
							continue;
						}
						else
							if( z->region()->type() == GRegion::Type::BLACKHOLE )
								penumbraScaling = 1.0;
					}
					*/
///////////////////////////////
			}
				else {
					if( L.type == Light::Type::Sun )
						penumbraScaling = 1.0;	// no penumbra by the sun. we only use 1
																		// sample for the sunlight for reasons of efficiency
					else
						penumbraScaling += 1.0 / (double)nsamples;
				}
				pEngine->bodyCheckId( oldBodyCheckId );
			}
			else // no shadowing requested
				penumbraScaling = 1.0;

			////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
			// Evaluate BxDF for current material
			//
			if( pMaterial ) {
				Rho = m_PhysMatLUT[BxDFIndex]->f( *pRay, lightRay, ShadingNormal, *pMaterial, Color( 1.0, 1.0, 1.0 ), power );
			}

			lightRay.pop();

			// take BxDF for current sample into account
			curIntensity = curIntensity + Rho * (power * dot / pdf);

			// bail out early after 1 sample in case of sunlight
			if( L.type == Light::Type::Sun && lightSample == 0 )
				break;
		} // end of sampling for one light source

		totalIntensity += curIntensity * penumbraScaling;

	} // end of light loop

	return totalIntensity;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CSurfaceIntegrator::Li_GlobalIllumination( GeometryEngine* pEngine, Ray* pRay, const int RenderPass, int Light0 )

// [8/01/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Color						ShadingColor, TmpColor, IndirectColor, Rho;
	const Material* pMaterial = pRay->hitMaterial();
	Vector					R, ReflNormal1, ReflNormal2;
	Ray							GIRay;
	float						Pdf = 0.0;
	double					Dot = 0.0, DistSquared = 0.0;

	const int				cMaxScatteringSamples = 256;
	double					pRandNr[2];
	int							NrOfIndirectSamples = 0;

	if(pMaterial->specular() >= 1.0)
		return Color::Black;

	int BxDFIndex = m_RenderMode == CSurfaceIntegrator::cQuality ? pMaterial->bxdf() : 0;
	int NrOfSamples = std::min( cMaxScatteringSamples, GetNrOfGISamples() );   // restrict the nr. of samples to 256. otherwise we might end up with extreme rendering times

	/*int ScrambleBxDF = pEngine->random.integer( cPMJLUTSize );*/

	int oldBodyId = pEngine->incBodyCheckId();		// signal to cache that we have a new ray direction

	// set general ray attributes
	GIRay.use_clip = GIRay.clip_insight = false;
	GIRay.use_project = GIRay.project_insight = false;

	Vector ShadingNormal = m_PhysMatLUT[BxDFIndex]->GetShadingNormal( pRay );

	// GI sampling
	for(int i = 0; i < NrOfSamples; ++i) {

		// CBaseSampler::getPMJ02Sample( i, pRandNr[0], pRandNr[1], ScrambleBxDF );
		CBaseSampler::getPMJ02BlueNoiseSample( i, pRandNr[0], pRandNr[1]/*, ScrambleBxDF*/ );


		CBxDFBase::SampleRayOnHemisphere(*pRay, GIRay, Pdf, ShadingNormal, pRandNr);
		Rho = m_PhysMatLUT[BxDFIndex]->f(*pRay, GIRay, ShadingNormal, *pMaterial, Color(1.0, 1.0, 1.0), m_GIScaling );

		// disable clipping bodies for reflection rays, to avoid visualization artifacts of background bleeding

		GIRay.depth = GIRay.max_depth + 1;		// avoid branching further into reflection and transmission

		// calculate incoming reflected luminance
		Dot = fabs( ShadingNormal.dot( GIRay.dir() ) );
		if ( Rho > Color::Black && Dot > 0 ) {
			auto oldId = pEngine->incBodyCheckId();		// signal to cache that we have a new ray direction
			TmpColor = Li(pEngine, &GIRay, RenderPass, Light0, cGlobalIllumination) * Dot * Rho / Pdf;

			/*DistSquared = 1.0; */
			DistSquared = max( 5.0, GIRay.T() * GIRay.T() );        // distance attenuation, we enforce an (unphysical) minimum distance factor of 2..3 cm (leading to a factor of ->2.23^2=5) to avoid too bright lights from non-diffuse surfaces
																														
			IndirectColor += TmpColor / DistSquared;
			NrOfIndirectSamples++;

			pEngine->bodyCheckId( oldId );

		}
	}

	// reset the ray cache
	pEngine->bodyCheckId(oldBodyId);

	// normalize
	IndirectColor /= NrOfIndirectSamples;
	ShadingColor += IndirectColor;

	return ShadingColor;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CSurfaceIntegrator::PreDrawInit()

// prior to each draw we collect a list of all physical materials that are actually used and load the textures for those which have not
// been loaded yet. For those which have a texture loaded but are not used, we unload it.
//  [11/12/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	set<int> UsedBxDF;

	// get indices of all used bxdf
	for ( const auto& CurRegion : m_pGeometry->regions ) {
		if ( CurRegion->material() )
			UsedBxDF.insert( CurRegion->material()->bxdf() );
	}

	for ( size_t i = 0; i < m_PhysMatLUT.size(); ++i ) {
		if ( UsedBxDF.find( i ) != UsedBxDF.end() ) {	// BxDF used
			if ( m_PhysMatLUT[i]->RequiresTexture() && !m_PhysMatLUT[i]->IsTextureLoaded() ) // texture not yet loaded
				m_PhysMatLUT[i]->InitTextures();
		}
		else {	// BxDF not used
			if ( m_PhysMatLUT[i]->RequiresTexture() && m_PhysMatLUT[i]->IsTextureLoaded() ) // texture loaded but not needed
				m_PhysMatLUT[i]->ClearTextures();
		}
	}

	return true;
}

// vim:ts=2
