////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "matshinymetal.h"

CMatShinyMetal::CMatShinyMetal( CBaseIntegrator* pInt, const float cRoughness /*= 1.0 */, const float cAbsorption /* = 0.0 */, bool bNoiseTex /*= false*/, bool bDisplacement /*= false*/, double BumpScaling /*= 1*/) : CPhysMatInterface(pInt, static_cast<CBxDFBase::BxDFType>(CBxDFBase::BSDF_DIFFUSE | CBxDFBase::BSDF_SPECULAR | CBxDFBase::BSDF_GLOSSY | CBxDFBase::BSDF_REFLECTION), true /* true = MC sampling required for the microfacet distributions for example */),
                                m_Absorption(cAbsorption), m_bUseNoise( bNoiseTex ), m_bDisplacement( bDisplacement), m_BumpScaling( BumpScaling )
{
  m_pFresnelReflection = new CFresnelPreDefConductor( m_Absorption, 1.0);
  m_pFresnelSpecular   = new CFresnelPreDefConductor( m_Absorption, 1.0);
  m_pBlinnMicroDist    = new CBlinnMicroDist( cRoughness );

  m_pBxDFDiffuse        = new CBxDFPhong;
  m_pDiffuseMicroFacets = new CBxDFMicrofacet( m_pFresnelReflection, m_pBlinnMicroDist );      // microfacet diffuse reflector
  m_pBxDFSpecular       = new CBxDFSpecReflection( m_pFresnelSpecular );

  if ( !bDisplacement ) { // bump mapping only
		m_NoiseTexture.SetMinMax( 0.5, 1.0 );       // restrict noise
		m_NoiseTexture.m_Style = CNoiseTexture::cTurbulent;
  }
  else {
		m_NoiseTexture.SetMinMax( 0.0, 0.7 );       // restrict noise
		m_NoiseTexture.m_Style = CNoiseTexture::cArtistic;
  }
}

CMatShinyMetal::~CMatShinyMetal(void)
{
	delete m_pBxDFDiffuse;      // microfacet diffuse reflector
	delete m_pDiffuseMicroFacets;
	delete m_pBxDFSpecular;     // specular reflection

  delete m_pFresnelReflection;
  delete m_pFresnelSpecular;
  delete m_pBlinnMicroDist;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CMatShinyMetal::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );

  Color Diffuse, Specular, DiffuseMicroFacets;

  m_pBlinnMicroDist->SetExp( std::max( 1.0 / 12.0, RenderMat.shine() / 12.0 ) );
  // approximate eta
  m_pFresnelReflection->SetEta( CFresnelHelpers::FresnelApproxEta( RenderMat.specular() ) );
  m_pFresnelSpecular->SetEta( CFresnelHelpers::FresnelApproxEta( RenderMat.shine() / 100.0 ) );
  // set absorption k --> done in ctor
  //
  if ( m_Absorption == 0.0 ) {
    m_Absorption = CFresnelHelpers::FresnelApproxK( RenderMat.specular() );

	  m_pFresnelReflection->SetAbsorption( m_Absorption );
	  m_pFresnelSpecular->SetAbsorption( m_Absorption );
  }


	double NoiseScaling = Out.hitRegion()->region()->material()->roughness();
	double NoiseFact = m_bUseNoise ? m_NoiseTexture.Value( 0, 0, Out.hit(), NoiseScaling )[0] : 1.0;    // for the moment we only use the 1st channel of our noise, though we could use 3 of them if we want to influence

  Diffuse            = m_pBxDFDiffuse->f( Out, In, Normal, RenderMat, LightColor, Shade );
  DiffuseMicroFacets = m_pDiffuseMicroFacets->f( Out, In, Normal, RenderMat, LightColor, Shade );
  Specular           = m_pBxDFSpecular->f( Out, In, Normal, RenderMat, LightColor, Shade );

  return LightColor * (Diffuse + DiffuseMicroFacets ) * NoiseFact * (1.0 - RenderMat.specular()) + Specular * std::max( RenderMat.specular(), 0.1 );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CMatShinyMetal::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, CBxDFBase::BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0*/ )

// Sample_f samples an incoming ray matching to the outgoing ray + its Pdf. Furthermore it evaluates the
// BxDF according to In & Out and returns the color. This is used for reflection, refraction etc. which will
// continue recursively.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );


  // choose between specular & diffuse reflection
  Color Rho;
  double TotalPDF = 0.0;

  if(pRandNr[0] < 0.5) {
    Rho = m_pBxDFSpecular->Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade );
    if ( Rho <= Color::Black )
      return Rho;


    TotalPDF += Pdf;
    TotalPDF += m_pBxDFDiffuse->GetPDF( -Out.dir(), In.dir(), Normal);
  }
  else {
    Rho = m_pBxDFDiffuse->Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade);
		if ( Rho <= Color::Black )
			return Rho;
    
    TotalPDF += Pdf;
		TotalPDF += m_pBxDFSpecular->GetPDF( -Out.dir(), In.dir(), Normal);
  }

  // renormalize Pdf
  Pdf /= TotalPDF;

  return Rho;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const Vector& CMatShinyMetal::GetShadingNormal( const Ray* pCurRay )

//  [9/12/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  //return pCurRay->normal;

  // check if we need to evaluate and update our cached shading normal based on Perlin noise bump-mapping

  if ( (m_bUseNoise || m_bDisplacement) && (!m_pCachedShadingRay || pCurRay != m_pCachedShadingRay) ) {
    m_CachedShadingNormal = m_NoiseTexture.GetBumpNormal( pCurRay, m_BumpScaling );
    m_pCachedShadingRay = const_cast<Ray*>(pCurRay);
  }
  else if ( !(m_bUseNoise && m_bDisplacement) ) {
		return pCurRay->normal;
	}

	return m_CachedShadingNormal;

}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CMatShinyMetal::ModifyShadingGeometry( Ray* pCurRay, GeometryEngine* pEngine )

//  implement displacement mapping if procedural noise textures are activated
//
//  [9/14/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  if( m_bDisplacement )
	  m_NoiseTexture.DisplaceHitRay( pCurRay, pEngine, m_BumpScaling);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Shiny metal with the possibility to use Trowbridge Reitz microfacet distributions as they are more advanced
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
CMatShinyMetalEx::CMatShinyMetalEx( CBaseIntegrator* pInt, const float cRoughnessX, const float cRoughnessY, const float cAbsorption /* = 0.0 */, bool bNoiseTex /*= false*/, bool bDisplacement /* = false */, double BumpScaling /*= 1*/) :
                  CMatShinyMetal( pInt, 0, cAbsorption, bNoiseTex, bDisplacement, BumpScaling)
{
	m_pFresnelReflection = new CFresnelPreDefConductor( m_Absorption, 1.0 );
	m_pFresnelSpecular = new CFresnelPreDefConductor( m_Absorption, 1.0 );
  m_pTrowbridgeReitz = new CTrowbridgeReitzDist( cRoughnessX, cRoughnessY, true );
	m_pBxDFDiffuse = new CBxDFPhong;

	m_pDiffuseMicroFacets = new CBxDFMicrofacet( m_pFresnelReflection, m_pTrowbridgeReitz );      // microfacet diffuse reflector
	m_pBxDFSpecular = new CBxDFSpecReflection( m_pFresnelSpecular );

}

CMatShinyMetalEx::~CMatShinyMetalEx()
{
	delete m_pTrowbridgeReitz;
}

// vim:ts=2
