////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "mattranslucent.h"

CMatTranslucent::CMatTranslucent( CBaseIntegrator* pInt ) : CPhysMatInterface( pInt, static_cast<CBxDFBase::BxDFType>( CBxDFBase::BSDF_DIFFUSE | CBxDFBase::BSDF_SPECULAR | CBxDFBase::BSDF_TRANSMISSION | CBxDFBase::BSDF_TRANSLUCENT | CBxDFBase::BSDF_REFLECTION ), true /* false = no MC sampling required */ )
{
  m_Absorption = 0.2;
  m_pBxDFDiffuse      = new CBxDFPhong;
  m_pBxDFTranslucent  = new CBxDFTranslucent; // translucent transmission & reflection

  m_pFresnelReflection = new CFresnelPreDefConductor( m_Absorption, 1.0 );
	m_pTrowbridgeReitz = new CTrowbridgeReitzDist( CTrowbridgeReitzDist::RoughnessToAlpha( 0.5 ), CTrowbridgeReitzDist::RoughnessToAlpha( 0.5 ), true );
	m_pDiffuseMicroFacets = new CBxDFMicrofacet( m_pFresnelReflection, m_pTrowbridgeReitz );      // microfacet diffuse reflector
}

CMatTranslucent::~CMatTranslucent(void)
{
  delete m_pBxDFDiffuse;
  delete m_pBxDFTranslucent;
  
  delete m_pTrowbridgeReitz;
  delete m_pDiffuseMicroFacets;
  delete m_pFresnelReflection;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CMatTranslucent::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );
  // return phong for diffuse reflection

	auto Diffuse = m_pBxDFDiffuse->f( Out, In, Normal, RenderMat, LightColor, Shade );

  
	m_pFresnelReflection->SetEta( RenderMat.ior() );

  m_pTrowbridgeReitz->SetRoughness( std::min( 1.0, RenderMat.fuzz() / 5.0 ), std::min( 1.0, RenderMat.fuzz() / 5.0 ) );
	auto DiffuseMicroFacets = m_pDiffuseMicroFacets->f( Out, In, Normal, RenderMat, LightColor, Shade );

	return LightColor * (Diffuse + DiffuseMicroFacets);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CMatTranslucent::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, CBxDFBase::BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0*/ )


// Sample_f samples an incoming ray matching to the outgoing ray + its Pdf. Furthermore it evaluates the
// BxDF according to In & Out and returns the color. This is used for reflection, refraction etc. which will
// continue recursively.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );

  if ( Type == CBxDFBase::BSDF_REFLECTION ) {
		return m_pBxDFTranslucent->Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade );
  }
  else if( Type == CBxDFBase::BSDF_TRANSMISSION )
    return m_pBxDFTranslucent->Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade);

  return CPhysMatInterface::Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade );
}
// vim:ts=2
