////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Standard Perlin noise implementation as proposed by P. Shirley et al.
//

#include <cmath>

#include "perlinnoise.h"
#include "math/random.h"

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CPerlinNoise::CPerlinNoise()

//  [9/5/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{

	for ( size_t i = 0; i < m_RandomVectors.size(); i++ ) {

		m_RandomVectors[i] = m_RNG.vector();
		m_Permutation[0][i] = static_cast<int>(i);
		m_Permutation[1][i] = static_cast<int>(i);
		m_Permutation[2][i] = static_cast<int>(i);
	}

	Shuffle( m_RNG, m_Permutation[0].data(), static_cast<int>(m_Permutation[0].size()) );
	Shuffle( m_RNG, m_Permutation[1].data(), static_cast<int>(m_Permutation[1].size()) );
	Shuffle( m_RNG, m_Permutation[2].data(), static_cast<int>(m_Permutation[2].size()) );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CPerlinNoise::Shuffle( Random& RNG, int* p, int n )

//  [9/5/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	for ( int i = (n - 1); i > 0; i-- ) {
		int t = RNG.integer( i );
		std::swap( p[i], p[t] );
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CPerlinNoise::PerlinInterpolation( const Vector c[2][2][2], const double u, const double v, const double w ) const

//  [9/5/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
// Cubic Hermite spline smoothing
	const double uu = u * u * (3 - (2 * u));
	const double vv = v * v * (3 - (2 * v));
	const double ww = w * w * (3 - (2 * w));

	double Sum = 0;
	for ( int i = 0; i < 2; i++ ) {
		for ( int j = 0; j < 2; j++ ) {
			for ( int k = 0; k < 2; k++ ) {
				const auto ri = static_cast<double>(i);
				const auto rj = static_cast<double>(j);
				const auto rk = static_cast<double>(k);

				const Vector Weight( u - ri, v - rj, w - rk );

				Sum += ((ri * uu) + ((1 - ri) * (1 - uu)))
							 * ((rj * vv) + ((1 - rj) * (1 - vv)))
							 * ((rk * ww) + ((1 - rk) * (1 - ww)))
							 * c[i][j][k].dot( Weight );
			}
		}
	}

	return Sum;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CPerlinNoise::Noise( const Vector& p ) const

//  [9/5/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	const auto i = static_cast<int>(floor( p.x ));
	const auto j = static_cast<int>(floor( p.y ));
	const auto k = static_cast<int>(floor( p.z ));

	const double u = p.x - static_cast<double>(i);
	const double v = p.y - static_cast<double>(j);
	const double w = p.z - static_cast<double>(k);

	Vector c[2][2][2];

	for ( int di = 0; di < 2; di++ ) {
		for ( int dj = 0; dj < 2; dj++ ) {
			for ( int dk = 0; dk < 2; dk++ ) {

				c[di][dj][dk] = m_RandomVectors[static_cast<size_t>( m_Permutation[0][static_cast<size_t>((i + di) & 0xFF)] ^
																														 m_Permutation[1][static_cast<size_t>((j + dj) & 0xFF)] ^
																														 m_Permutation[2][static_cast<size_t>((k + dk) & 0xFF)]
					)];

			}
		}
	}

	return PerlinInterpolation( c, u, v, w );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CPerlinNoise::Turbulence( const Vector& p, const int NoiseDepth /* = 7 */) const

//  [9/5/2023 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double Sum = 0, Weight = 1;
	Vector Temp( p );

	for ( auto i = 0; i < NoiseDepth; ++i ) {
		Sum += Weight * Noise( Temp );
		Weight *= static_cast<double>(0.5);
		Temp *= 2;
	}

	return std::abs( Sum );
}