/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "arbbody.h"

#include "tool/bstring.h"

#include "mesh.h"

// ================================= GARBBody ==================================
// FIXME position, get, Mesh...
void GARBBody::set(int /*n*/, double* what)
{
	std::string err;
	err[0] = 0;

	GBody::createMesh();	// initialize an empty mesh
	resetBBox();
	round0(what, 30, SMALL);
	_mesh->allocateVertices(8);

	// add vertices
	P.set(what[0], what[1], what[2]);
	for (int i=0; i<8; i++)
		_mesh->vertex(i) = Point(what[i*3], what[i*3+1], what[i*3+2]);

	// copy faces
	bool same = true;
	for (int j=0,i=3*8; i<3*8+6; i++,j++) {
		int face = (int)what[i];
		if (face != faces[j]) {
			same = false;
			faces[j] = face;
		}
		int zero = 0;
		for (int k=0; k<4; k++) {
			int v = face%10;
			if (v>8) {
				err = sformat("Face %d contains an invalid vertex %d",
						(int)what[i],v);
				v = 0;
			}
			if (v==0) zero++;
			face /= 10;
		}
		if (zero>1) {
			err = sformat("Face %d contains too few vertices",
					(int)what[i]);
			continue;
		}
	}
	if (!same) {
		_mesh->freeFaces();
		_mesh->freeEdges();
	}
	_mesh->calcBbox();
	if (not err.empty()) throw err;
} // set

/* --- get --- */
int GARBBody::get(double* what) const
{
	int w = 0;
	for (int i=0; i<8; i++) {
		what[w++] = _mesh->vertex(i).x;
		what[w++] = _mesh->vertex(i).y;
		what[w++] = _mesh->vertex(i).z;
	}
	for (int i=0; i<6; i++)
		what[w++] = (double)faces[i];
	return 30;
} // get

/** return face vertices
 * @param f     face to return         [I]
 * @param vi    index to vertices      [O]
 *              initialized to -1
 * @param V     pointer to vertex points  [O]
 *              initialized to NULL
 * @return the number of vertices used for the face <3 on error, 3 or 4
 */
int GARBBody::faceVertices(int f, int vi[4], Point* V[4])
{
	int face = faces[f];
	memset(vi, 0xff, sizeof(vi[0])*4);	// -1
	memset(V,  0,    sizeof(V[0])*4);	// nullptr
	if (!face) return 0;
	int n = 0;
	for (int i=0; i<4; i++) {
		vi[n] = face%10 - 1;
		if (vi[n]>=8) return 0;
		if (vi[n]>=0) {
			V[n] = &_mesh->vertex(vi[n]);
			// check for duplicate vertices
			for (int k=0; k<n; k++)
				if (vi[k] == vi[n]) return false;
			n++;
		}
		face /= 10;
	}
	return n;	// we need at least 3 non identical vertices
} // faceVertices

/** createQuads */
void GARBBody::createQuads()
{
	GBody::createQuads();
	// create faces and quadratics
	for (int i=0; i<6; i++) {
		int vi[4];
		Point* V[4];
		int n = faceVertices(i, vi, V);
		if (n<3) continue;

		// create quads
		Vector N = (*V[1] - *V[0]) ^ (*V[2] - *V[0]);
		N.normalize();
		addQuad(-N.x, -N.y, -N.z, N*(*V[0]));

		if (n==4) {
			// check for co-planarity!
			if (!Eq0(_Q[_nQ-1](V[3]->x, V[3]->y, V[3]->z), SMALL))
				fprintf(stderr,"Face %d not planar\n",(int)faces[i]);
		}

		// Check remaining vertices, from which side they are
		bool first = true;
		bool positive = false;
		for (int j=0; j<8; j++) {
			if (In(j, vi[0], vi[1], vi[2], vi[3])) continue;
			double q = _Q[_nQ-1](	_mesh->vertex(j).x,
						_mesh->vertex(j).y,
						_mesh->vertex(j).z);
			// FIXME throw an error!!!
			if (Eq0(q, SMALL))
				fprintf(stderr,"Vertex %d is coplanar with face %d\n",j+1,(int)faces[i]);
			if (first) {
				positive = (bool)(q>0.0);
				first = false;
			} else
			if (positive != (bool)(q>0.0))
				fprintf(stderr,"Cannot resolve ARB orientation\n");
		}
		// if remaining vertices are from same side with the normal
		if (positive)
			_Q[_nQ-1].negate();		// swap the sign of the quad
	}
} // createQuads

/* createMesh */
void GARBBody::createMesh()
{
	GBody::createMesh();
	if (_mesh->nedges()) return;

	for (int i=0; i<6; i++) {
		int vi[4];
		Point* V[4];
		int n = faceVertices(i, vi, V);
		if (n<3) continue;
		_mesh->add(vi[0],vi[1],vi[2], true, true, n==3);
		if (n==4)
			_mesh->add(vi[2],vi[3],vi[0], true, true, false);
	}
	_mesh->calcBbox();
	_mesh->process();
	if (!_mesh->isClosed()) {
		fprintf(stderr,"ERROR open ARB a face is missing\n");
		return;
	}
	if (!_mesh->isOrientable())
		_mesh->makeOrientable();

	if (_mesh->volume()<0.) _mesh->flip();

	assert(_mesh->isClosed());
	assert(_mesh->isOrientable());
#if _DEBUG>2
	cout << endl;
	cout << "ARB Mesh " << endl;
	cout << mesh << endl;
	cout << " isClosed=" << _mesh->isClosed();
	cout << " isOrientable=" << _mesh->isOrientable() << endl;
	cout << "ARB volume=" << _mesh->volume() << endl;
#endif
} // createMesh

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GARBBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point dR = r - P;
	for (size_t i=0; i<_mesh->nvertices(); i++)
		_mesh->vertex(i) += dR;

	GBody::move(item, r, w);
} // move

/** rotate body around an axis and current body position
 * @param angle	to rotate
 * @param axis	to rotate
 */
void GARBBody::rotate(const double angle, const Vector& axis)
{
	GBody::rotate(angle, axis);
	Matrix4	rot;			// duplicate rot matrix with the GBody::rotate
	rot.rotate(angle, axis);
	for (int i=1; i<8; i++) {
		Vector v = _mesh->vertex(i) - P;
		_mesh->vertex(i) = rot*v + P;
	}
} // rotate

/** @return bounding box of body */
void GARBBody::_updateBBox()
{
	assert(_mesh!=nullptr);
	if (_mesh->nvertices()) {
		// Find the used vertices and add them to the bbox
		// when ARB is defined with less than 8 vertices
		// it will add dummy (0,0,0)!!!
		for (int f=0; f<6; f++) {
			int vi[4];
			Point* V[4];
			int n = faceVertices(f, vi, V);
			if (n<3) continue;
			for (int i=0; i<n; i++)
				if (V[i] != nullptr)
					_bbox.add(*V[i]);
		}
	} else
		_bbox.infinite();
} // _updateBBox

#ifdef USE_OBBOX
/** @return oriented bounding box of body */
void GARBBody::_updateOBB(bool in)
{
	OBBox& bb = _obbox[(int)in];
	if (!in) {
		if (_mesh->nvertices()) {
			// use the vertices as a guess
			bb.add(position());
			for (size_t i=0; i<_mesh->nvertices(); i++)
				bb.add(_mesh->vertex(i));
		} else
			bb.infinite();
	} else {
		// TODO: Implement inside obb?
		assert(false);
	}
} // _updateOBB
#endif
