/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 *
 * Author:	Christian.Theis@cern.ch
 *
 * Date: 30-Jun-2021
 *
 *
 * Code to construct a hierarchical bounding volume based VZone objects that can be queried efficiently to determine in which zone a point is actually located.
 *
 * This version uses a relatively simple midpoint splitting strategy choosing the axis where the bounding box is largest to minimize overlaps of sub-volumes. This gives reasonable trees but they
 * clearly can be optimized. One could think about calculating Morton codes for each volume in the current bounding volume and order them by the Morton codes. The splitting point could then be determined
 * by looking at the index where the highest bit starts to flip for the first time. This is repeated recursively and we would automatically have a quite well balanced tree. This could also be combined
 * with cost functions based on SAH like Pantaleoni et al. did for Weta's Pantaray. To be investigated.
 *
 */

#include "bvh.h"

#include "geometry/ray.h"

#include <iostream>
#include <fstream>
#include <vector>
#include <cassert>


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy::CBoundingVolHierarchy( std::vector<VZone*> ZoneList, bool bIncludeInvalidZones, bool bVerbose /* = false */, int LogLevel /*= 0*/ ) :
                                              m_pLeftLeaf(nullptr), m_pRightLeaf(nullptr), m_pZone(nullptr), m_NrOfLeaves(0), m_NrOfValidZones(0), m_bIncludeInvalidZones( bIncludeInvalidZones ), m_bVerbose( bVerbose ), m_LogLevel( LogLevel ),
                                              m_pNodeList( ZoneList ), m_bTreeBuilt(false)

// [6/30/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	m_bVerbose&& std::cerr << "*** Standard bounding volume hierarchy tree selected ***" << std::endl;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBoundingVolHierarchy::BuildTree()

// [6/30/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Vector CurBoxMin, CurBoxMax, NextBoxMin, NextBoxMax, TmpBoxMin, TmpBoxMax;
	double Pivot;
	int    Axis = cXAxis;

	assert( !m_pNodeList.empty() );

	m_bVerbose&& std::cerr << "BVH: starting construction for " << m_pNodeList.size() << std::endl;
	if( m_pNodeList.size() < 1 )
		// throw std::runtime_error( "BVH: insufficient geometry regions to build BV tree" );
		return;

	if( m_pNodeList.size() == 1 ) {
		*this = CBoundingVolHierarchy( m_pNodeList[0] );
	}

	if( m_pNodeList.size() >= 2 ) {
		// Obtain the centroid of the first bounding box that will be used as a pivot for the qsplit routine
		BBox CurBBox = m_pNodeList[0]->bbox();
		if( !CurBBox.isValid() ) {
				m_pInvalidBBoxZones.push_back( m_pNodeList[0] );
		}


		for( size_t i = 1; i < m_pNodeList.size(); ++i ) {
			BBox NextBBox = m_pNodeList[i]->bbox();

			if( !NextBBox.isValid() ) {
				m_pInvalidBBoxZones.push_back( m_pNodeList[i] );
			}
			else {
				CurBBox = CurBBox | NextBBox;
			}
		}

		// set up the bounding box of the whole union
		m_BoundingBox = CurBBox;

		// get position and axis along which we should split
		GetSplittingPos( m_pNodeList.data(), m_pNodeList.size(), Pivot, Axis );
		Axis = cXAxis;	// for the first generation we enforce splitting of the x-axis and subsequent calls to GetSplittingPos() will iterate over the axes
		size_t MidPoint = QSplit( m_pNodeList.data(), m_pNodeList.size(), Pivot, Axis);

		// Create a new bounding volume & define the next axis as the new split axis
		m_pLeftLeaf = BuildBranch( m_pNodeList.data(), MidPoint, Axis );
		m_pRightLeaf = BuildBranch( &m_pNodeList.data()[MidPoint], m_pNodeList.size() - MidPoint, Axis );
		m_pZone = nullptr;
	}

	m_bVerbose&& std::cerr << "BVH: Total nr of tree leaves = " << m_NrOfLeaves << std::endl;
	m_bVerbose&& std::cerr << "BVH: BBox (" << m_BoundingBox.low() << ") -> (" << m_BoundingBox.high() << ")" << std::endl;

	m_bTreeBuilt = true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy::CBoundingVolHierarchy( VZone* pZone ) : m_pLeftLeaf( nullptr), m_pRightLeaf( nullptr),
                                              m_pZone( pZone), m_NrOfLeaves( 1 ), m_NrOfValidZones( 1 ), m_bIncludeInvalidZones( false ), m_bVerbose( false ), m_LogLevel( 0 ),
																							m_bTreeBuilt(true)

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  m_BoundingBox = m_pZone->bbox();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy::CBoundingVolHierarchy( CBoundingVolHierarchy* pNode1, CBoundingVolHierarchy* pNode2 ) :
                                              m_pLeftLeaf(pNode1), m_pRightLeaf( pNode2), m_pZone(nullptr),
																							m_NrOfLeaves( 1 ), m_NrOfValidZones( 1 ), m_bIncludeInvalidZones( false ), m_bVerbose( false ), m_LogLevel( 0 ),
																						  m_bTreeBuilt( true )

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector Box1Min, Box2Min, Box1Max, Box2Max;

  if( m_pRightLeaf ) {
    m_BoundingBox = pNode1->bbox() | pNode2->bbox();
		m_NrOfLeaves++;
		m_NrOfValidZones++;
  }
  else
    m_BoundingBox = pNode1->bbox();
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy* CBoundingVolHierarchy::BuildBranch( VZone** ppNodeList, size_t NrOfSubBranchNodes, int Axis /* = cXAxis */ )

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  double Pivot;

  m_NrOfLeaves++;

  if( NrOfSubBranchNodes == 1) {
		m_NrOfValidZones++;
		return new CBoundingVolHierarchy( ppNodeList[0] );
	}

  // determine the splitting axis and position on that axis
  GetSplittingPos( ppNodeList, NrOfSubBranchNodes, Pivot, Axis );

  // let's split according to the respective parameters
  size_t  MidPoint = QSplit( ppNodeList, NrOfSubBranchNodes, Pivot, Axis);

  if( m_bVerbose && m_LogLevel >= 2 ) {
    std::cerr << "BVH: Splitting axis = " << Axis << ", pivot = " << Pivot << " splitting " << MidPoint << "/" << NrOfSubBranchNodes - MidPoint << std::endl;
  }

  // assemble a new bounding volume iterating the splitting axis
  CBoundingVolHierarchy* pLeftLeaf = BuildBranch( ppNodeList, MidPoint, Axis);
  CBoundingVolHierarchy* pRightLeaf = BuildBranch( &ppNodeList[MidPoint], NrOfSubBranchNodes - MidPoint, Axis );


  return new CBoundingVolHierarchy( pLeftLeaf, pRightLeaf);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

size_t  CBoundingVolHierarchy::QSplit( VZone** ppNodeList, size_t NrOfNodes, double PivotVal, int Axis)

// determine the number of nodes from a subset (given by ppStartNode + NrOfNodes) that have a centroid
// coordinate < PivotVal with respect to a specific axis.
// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector   BoxMin, BoxMax;
  double   Centroid = 0.0;
  size_t   RetVal  = 0;

  for( size_t i = 0; i < NrOfNodes; ++i) {

    BBox CurBBox = ppNodeList[i]->bbox();

    Centroid = CurBBox.centroid( Axis );
    if( Centroid < PivotVal ) {
      std::swap( ppNodeList[i], ppNodeList[RetVal] );
      RetVal++;
    }
  }

  if( RetVal == 0 || RetVal == NrOfNodes )
    RetVal = NrOfNodes / 2;

  return RetVal;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::Contains( GeometryEngine* pEngine, const Vector& CurPoint, const Vector& CurDir, double t, double eps )

// [7/2/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  VZone* pLeftHitZone = nullptr;
	VZone* pRightHitZone = nullptr;

	// branch into further BHVs
	// TODO: check if we should not better use a ray intersection for better FP accuracy

	if( !m_BoundingBox.inside( CurPoint + CurDir * t, eps ) )
		return nullptr;

	// check if we have are region primitive, otherwise branch further
  if( m_pZone ) {
		if( m_pZone->inside( pEngine, CurPoint, CurDir, t ) )
			return m_pZone;
		return nullptr;
  }
	else if( m_pLeftLeaf ) {
    pLeftHitZone = static_cast<CBoundingVolHierarchy*>(m_pLeftLeaf)->Contains( pEngine, CurPoint, CurDir, t, eps );
		if( pLeftHitZone )
			return pLeftHitZone;
  }
	else if( m_pRightLeaf  ) {
    pRightHitZone = static_cast<CBoundingVolHierarchy*>(m_pRightLeaf)->Contains( pEngine, CurPoint, CurDir, t, eps );
  }

	return pRightHitZone;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::inside( GeometryEngine* pEngine, const Vector& CurPoint, const Vector& CurDir, double t, double eps /* = SMALL3D4 */ )

// Check if the point CurPoint + t * CurDir is inside the BVH and return the corresponding zone or a nullptr
//
// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// [7/2/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  if( m_NrOfLeaves == 0 )
    return nullptr;

  VZone* pCurZone = Contains( pEngine, CurPoint, CurDir, t, eps );
	if( !pCurZone && m_bIncludeInvalidZones ) {
		pCurZone = insideInvalidZones( pEngine, CurPoint, CurDir, t, eps );
	}

	return pCurZone;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBoundingVolHierarchy::PrintTree( const std::string& FileName /*=""*/)

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	std::streambuf* coutbuf = nullptr;

	if( !FileName.empty() ) {
			// save current buffer of cerr
		coutbuf = std::cerr.rdbuf();

		std::ofstream out( FileName);
		std::cerr.rdbuf( out.rdbuf() );                     //redirect cerr
	}

  std::cerr << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;
  std::cerr << "|                 Bounding volume hierarchies              |" << std::endl;
	std::cerr << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;
  std::cerr << " Nr of zones = " << m_pNodeList.size() << ", nr of leaves = " << m_NrOfLeaves << std::endl;
	std::cerr << " Nr of zones with valid bboxes = " << GetNrOfValidZones() << ", nr of zones with invalid bounding boxes = " << std::distance( m_pInvalidBBoxZones.begin(), m_pInvalidBBoxZones.end() ) << std::endl << std::endl;

	if ( m_pLeftLeaf ) {
		m_pLeftLeaf->PrintLeaf();
	}

	if ( m_pRightLeaf ) {
		m_pRightLeaf->PrintLeaf();
	}

	if( m_pZone ) {
		std::cerr << "--> #" << 0 << " - L-GEO: " << m_pZone->name() << " bmin =" << m_pZone->bbox().low() << ", bmax = " << m_pZone->bbox().high() << ", pointer =" << m_pZone << std::endl;
	}

  std::cerr << "================================================================" << std::endl;
  std::cerr << "Nr. of zones with invalid bounding boxes: " << std::distance( m_pInvalidBBoxZones.begin(), m_pInvalidBBoxZones.end() ) << std::endl;
  if( !m_pInvalidBBoxZones.empty() ) {
    std::cerr << "Zones with invalid bounding boxes:" << std::endl;
		for( auto iter : m_pInvalidBBoxZones ) {
			std::cerr << iter->name() << std::endl;
		}
	}

	std::cerr << "!! Safe mode which checks also regions with invalid bounding boxes (might be slow!): " << std::boolalpha << m_bIncludeInvalidZones << std::endl;

	if( !FileName.empty() ) {
		std::cerr.rdbuf( coutbuf );   //reset to standard output again
	}

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBoundingVolHierarchy::PrintLeaf( int Level /*= 0*/ )

// [11/18/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	std::cerr << "+";

  Indent( Level );
  std::cerr << "--> #" << Level << "  BB " << " bmin =" << m_BoundingBox.low() << ", bmax = " << m_BoundingBox.high() << std::endl;

	if ( m_LogLevel >= 1 ) {
		if ( m_pLeftLeaf ) {
			Indent( Level + 1 );
			std::cerr << "|" << std::endl;
			m_pLeftLeaf->PrintLeaf( Level + 1 );
		}

		if ( m_pRightLeaf ) {
			Indent( Level + 1 );
			std::cerr << "|" << std::endl;
			m_pRightLeaf->PrintLeaf( Level + 1 );
		}
	}

	if( m_pZone ) {
		Indent( Level + 1 );
		std::cerr << "--> #" << Level + 1 << " - L-GEO: " << m_pZone->name() << " bmin =" << m_pZone->bbox().low() << ", bmax = " << m_pZone->bbox().high() << ", pointer = " << m_pZone << std::endl;
	}

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::intersectRay( GeometryEngine* pEngine, Ray* pCurRay )

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// [11/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tmin = pCurRay->T();
  Vector InvDir( pCurRay->dir().x != 0.0 ? 1.0 / pCurRay->dir().x : 0.0,
                 pCurRay->dir().y != 0.0 ? 1.0 / pCurRay->dir().y : 0.0,
                 pCurRay->dir().z != 0.0 ? 1.0 / pCurRay->dir().z : 0.0 );

  // check if the current ray already has a potential intersection candidate

	VZone* pCurZone = pCurRay->segment().zone;
  if( pCurZone /*&& tmin > SMALL3D4*/) {
		if( /*const_cast<BBox&>(pCurZone->bbox()).intersectRay( pCurRay->pos(), InvDir, tmin, pCurRay->segment().tmax ) and*/
				pCurZone->inside( pEngine, pCurRay->pos().x, pCurRay->pos().y, pCurRay->pos().z, pCurRay->dir().x, pCurRay->dir().y, pCurRay->dir().z, tmin ) ) {
			return pCurZone;
    }
  }

  return intersectRay( pEngine, pCurRay, InvDir, tmin, pCurRay->segment().tmax );

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::intersect( GeometryEngine* pEngine, const Vector& Pos, const Vector& Dir, double& tmin, double tmax, VZone* pCurZone /*= nullptr*/ )

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// ATTENTION: We have to pass the corresponding geometry engine as each thread has its own and they all share the current BVH!
// [7/21/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tminTmp = tmin;
	Vector InvDir( Dir.x != 0.0 ? 1.0 / Dir.x : 0.0,
								 Dir.y != 0.0 ? 1.0 / Dir.y : 0.0,
								 Dir.z != 0.0 ? 1.0 / Dir.z : 0.0 );

	// check if the current ray already has a potential intersection candidate
	if( pCurZone ) {
		auto pCurBody = pCurZone->distance( pEngine, Pos, Dir, &tminTmp, tmax );
		if( pCurBody && tminTmp < tmax && tminTmp >= tmin ) {
			tmin = tminTmp;
			return pCurZone;
		}
	}

	return intersect( pEngine, Pos, Dir, InvDir, tmin, tmax );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::intersect( GeometryEngine* pEngine, const Vector& Pos, const Vector& Dir, const Vector& InvDir, double& tmin, double tmax )

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// ATTENTION: We have to pass the corresponding geometry engine as each thread has its own and they all share the current BVH!
// [6/25/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tminLeft = tmin, tmaxLeft = tmax, tminRight = tmin, tmaxRight = tmax, tminTemp = tmin;
	VZone* pLeftZone = nullptr;
	VZone* pRightZone = nullptr;


	// check if we have an actual zone to intersect
	if( m_pZone ) {
		VZone* pZone = nullptr;
		double tminOld = tmax;

		// find the closest hit, if any
		auto pCurBody = m_pZone->distance( pEngine, Pos, Dir, &tminTemp, tmax );
		if( pCurBody && tminTemp < tminOld && tminTemp >= tmin ) {
			pZone = m_pZone;
			tminOld = tminTemp;
		}

		// check if there might still be a potential intersection with a region that has an invalid bbox.
		// in perfectly valid geometries this should not be necessary but due to infinite bodies or half-spaces
		// this situation might be encountered
		if( pZone && tminOld >= tmin && tminOld <= tmax ) {
			tmin = tminOld;
		}
		else {
			if( m_bIncludeInvalidZones )
				pZone = intersectWithInvalidZones( pEngine, Pos, Dir, tmin, tmax );
		}

		return pZone;
	}

	// check bbox and leaves
	if( !m_BoundingBox.intersectRay( Pos, InvDir, tmin, tmax ) )
		return nullptr;

	// branch into tree
	if( m_pLeftLeaf ) {
		pLeftZone = m_pLeftLeaf->intersect( pEngine, Pos, Dir, InvDir, tminLeft, tmaxLeft );
	}

	if( m_pRightLeaf ) {
		pRightZone = m_pRightLeaf->intersect( pEngine, Pos, Dir, InvDir, tminRight, tmaxRight );
	}

	// return the closer hit
	if( pRightZone && tminRight < tminLeft ) {
		tmin = tminRight;
		return pRightZone;
	}

	tmin = tminLeft;
	return pLeftZone;


}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::intersectRay( GeometryEngine* pEngine, Ray* pCurRay, const Vector& InvDir, double& tmin, double tmax )

// ATTENTION: We have to pass the corresponding geometry engine as each thread has its own and they all share the current BVH!
// optimized intersection that avoids pre-calculating constants like the inverse direction
//
// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// [12/2/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tminLeft = tmin, tmaxLeft = tmax, tminRight = tmin, tmaxRight = tmax, tminTemp = tmin;
	VZone* pLeftZone = nullptr;
	VZone* pRightZone = nullptr;
	CBody* pLeftBody = nullptr;
	CBody* pRightBody = nullptr;

	// check if we have an actual region to intersect
	if( m_pZone ) {
		VZone* pZone = nullptr;
		double tminOld = tmax;

		// find the closest hit, if any
		auto pCurBody = m_pZone->distance( pEngine, pCurRay->pos(), pCurRay->dir(), &tminTemp, tmax );
		if( pCurBody && tminTemp < tminOld && tminTemp >= tmin ) {
			pZone = m_pZone;
			tminOld = tminTemp;
		}

		if( pZone ) {
			pCurRay->segment().tmin = tminOld;
			pCurRay->segment().zone = pZone;
			pCurRay->segment().body = pCurBody;
      tmin = tminOld;
		}
		// check if there might still be a potential intersection with a region that has an invalid bbox
		// in perfectly valid geometries this should not be necessary but due to infinite bodies or half-spaces
		// this situation might be encountered
		else {
			if( m_bIncludeInvalidZones )
				pZone = intersectWithInvalidZones( pEngine, pCurRay->pos(), pCurRay->dir(), tmin, tmax );
		}

		return pZone;
	}

	// check bbox and leaves
	if( !m_BoundingBox.intersectRay( pCurRay->pos(), InvDir, tmin, tmax ) )
		return nullptr;

	// branch into tree
	if( m_pLeftLeaf ) {
		pLeftZone = m_pLeftLeaf->intersectRay( pEngine, pCurRay, InvDir, tminLeft, tmaxLeft );
		pLeftBody = pCurRay->segment().body;
	}

	if( m_pRightLeaf ) {
		pRightZone = m_pRightLeaf->intersectRay( pEngine, pCurRay, InvDir, tminRight, tmaxRight );
		pRightBody = pCurRay->segment().body;
	}

	// return the closer hit
	if( pRightZone && tminRight < tminLeft ) {
		pCurRay->segment().zone = pRightZone;
		pCurRay->segment().body = pRightBody;
		pCurRay->segment().tmin = tminRight;
		return pRightZone;
	}

	pCurRay->segment().zone = pLeftZone;
	pCurRay->segment().body = pLeftBody;
	pCurRay->segment().tmin = tminLeft;

	return pLeftZone;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBoundingVolHierarchy::GetSplittingPos( VZone* const* ppNodeList, const size_t NrOfNodes, double& Pivot, int& Axis )

// retrieve splitting axis and point on this axis.
// Default implementation is based on simplistic iteration over the splitting axes. Yet, this could be improved via
// surface area heuristics or 3D sorting with Morton indices etc. TODO: check paper on NVidia/Weta's PantaRay
//
// [6/23/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  BBox   CurBBox, NextBBox;

  // Obtain the centroid of the bounding box that will be used as a pivot for the qsplit routine
	CurBBox = ppNodeList[0]->bbox();

  for( size_t i = 1; i < NrOfNodes; ++i ) {
		NextBBox = ppNodeList[i]->bbox();
		CurBBox = CurBBox | NextBBox;
	}

	// for each tree generation we iterate over the splitting axes; this is an extremely simple approach but Shirley et al. showed that
	// while it does not yield optimum performance, tree balancing is sufficient and tree generation costs are kept at bay.
	Axis = (Axis + 1) % 3;
	Pivot = CurBBox.centroid( Axis );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::intersectRayWithInvalidZones( GeometryEngine* pEngine, Ray* pCurRay )

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// intersect ray with regions that have invalid bounding boxes but are still stored in the tree as they are part of the original geometry
// [6/28/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tmin = pCurRay->T(), tmax = pCurRay->segment().tmax;
	VZone* pZone = pCurRay->segment().zone;
	VZone* pRetZone = nullptr;
	CBody* pBody = nullptr;

	pRetZone = intersectWithInvalidZones( pEngine, pCurRay->pos(), pCurRay->dir(), tmin, tmax, pZone, pBody );

	if( pRetZone ) {
		pCurRay->segment().zone = pRetZone;
		pCurRay->segment().body = pBody;
		pCurRay->segment().tmin = tmin;
		pCurRay->segment().tmax = tmax;
	}

	return pRetZone;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::intersectWithInvalidZones( GeometryEngine* pEngine, const Vector& Pos, const Vector& Dir, double& tmin, double tmax, VZone* pCurZone /*= nullptr */,
																												 CBody* pBody /* = nullptr */)

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// ATTENTION: We have to pass the corresponding geometry engine as each thread has its own and they all share the current BVH!
// [6/28/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tminTmp = tmin;
	std::list<VZone*>::iterator ZoneIter;

  if( pCurZone ) {
		auto pCurBody = pCurZone->distance( pEngine, Pos, Dir, &tminTmp, tmax );
		if( pCurBody && tminTmp < tmax && tminTmp >= tmin ) {
			tmin = tminTmp;
			pBody = pCurBody;
			return pCurZone;
		}
  }

	VZone* pZone = nullptr;
	CBody* pTmpBody = pBody;
	double tminOld = tmax;

	// loop over zones to find the closest hit, if any
	pBody = nullptr;
	for( auto iter = m_pInvalidBBoxZones.begin(); iter != m_pInvalidBBoxZones.end(); ++iter ) {
		auto pCurBody = (*iter)->distance( pEngine, Pos, Dir, &tminTmp, tmax );
		if( pCurBody && tminTmp < tminOld ) {
			pZone = *iter;
			pTmpBody = pCurBody;
			tminOld = tminTmp;
			ZoneIter = iter;
		}
	}

	if( pZone && tminOld >= tmin && tminOld < tmax ) {
		tmin = tminOld;
		pBody = pTmpBody;

		// promote zone for next search
		if( ZoneIter != m_pInvalidBBoxZones.begin() ) {
			m_pInvalidBBoxZones.erase( ZoneIter );
			m_pInvalidBBoxZones.push_front( pZone );
		}

		return pZone;
	}

	return nullptr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::insideInvalidZones( GeometryEngine* pEngine, const Vector& Pos, const Vector& Dir, double t, double eps )

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// ATTENTION: We have to pass the corresponding geometry engine as each thread has its own and they all share the current BVH!
// [7/2/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( eps );
	VZone* pZone = nullptr;

	for( auto iter = m_pInvalidBBoxZones.begin(); iter != m_pInvalidBBoxZones.end(); ++iter ) {
		if( (*iter)->inside( pEngine, Pos.x, Pos.y, Pos.z, Dir.x, Dir.y, Dir.z, t ) ) {

			pZone = *iter;	// use a temp var as we might change the order in the array before returning the object!
			// promote zone
			if( iter != m_pInvalidBBoxZones.begin() ) {
				m_pInvalidBBoxZones.erase( iter );
				m_pInvalidBBoxZones.push_front( pZone );
			}
			return pZone;
		}
	}

	return nullptr;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//                                                Surface area heuristics bounding volume hierarchies
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CSAHBoundingVolHierarchy::CSAHBoundingVolHierarchy( std::vector<VZone*> ZoneList, bool bIncludeInvalidZones, bool bVerbose /*= false*/, int LogLevel /*= 0*/ ) :
  CBoundingVolHierarchy( ZoneList, bIncludeInvalidZones, bVerbose, LogLevel)

// [6/23/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	m_bVerbose&& std::cerr << "*** SAH based bounding volume hierarchy tree selected ***" << std::endl;
  m_CentroidBBox = InitCentroidBBox( ZoneList.data(), ZoneList.size() );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CSAHBoundingVolHierarchy::CSAHBoundingVolHierarchy( VZone* pZone ) :
  CBoundingVolHierarchy( pZone )

// [6/23/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  m_CentroidBBox = m_pZone->bbox();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CSAHBoundingVolHierarchy::CSAHBoundingVolHierarchy( CBoundingVolHierarchy* pNode1, CBoundingVolHierarchy* pNode2 ) :
  CBoundingVolHierarchy( pNode1, pNode2 )

// [6/23/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  m_CentroidBBox = m_BoundingBox;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CSAHBoundingVolHierarchy::GetSplittingPos( VZone* const* ppNodeList, const size_t NrOfNodes, double& Pivot, int& Axis )

// [6/23/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  // use standard mid axis splitting for <= 4 nodes
  if( NrOfNodes <= 4 ) {
    CBoundingVolHierarchy::GetSplittingPos( ppNodeList, NrOfNodes, Pivot, Axis );
    return;
  }

  // determine the centroid bounding box of the current node subset + the dominant axis
  BBox CentroidBBox( InitCentroidBBox( ppNodeList, NrOfNodes ) );
  Axis = CentroidBBox.size().dominant();

  // subdivide the projection of the node subset into buckets and initialize the bounding box,
  // record which centroids belong to which bucket and determine the bounding box of all nodes per bucket
	for( unsigned int i = 0; i < NrOfNodes; ++i ) {
		int b = cNrOfBuckets * CentroidBBox.offset( ppNodeList[i]->bbox().centroid() )[Axis];
		if( b == cNrOfBuckets )
			b = cNrOfBuckets - 1;

    assert( b >= 0 && b < cNrOfBuckets );

    m_Buckets[b].m_Count++;
    m_Buckets[b].m_Box = m_Buckets[b].m_Box | ppNodeList[i]->bbox();
	}

  // compute the cost for splitting after each bucket
	constexpr int cNrOfSplits = cNrOfBuckets - 1;
	int   CountBelow[cNrOfSplits], CountAbove[cNrOfSplits];
	BBox  BoundsBelow[cNrOfSplits], BoundsAbove[cNrOfSplits];

  // Initialize CountBelow and BoundsBelow using a forward scan over the splits
	CountBelow[0] = m_Buckets[0].m_Count;
	BoundsBelow[0] = m_Buckets[0].m_Box;
  for( int i = 1; i < cNrOfSplits; ++i ) {
    CountBelow[i] = CountBelow[i - 1] + m_Buckets[i].m_Count;
    BoundsBelow[i] = BoundsBelow[i - 1] | m_Buckets[i].m_Box;
	}

	// Initialize CountAbove and BoundsAbove using a backwards scan
	CountAbove[cNrOfSplits - 1] = m_Buckets[cNrOfBuckets - 1].m_Count;
	BoundsAbove[cNrOfSplits - 1] = m_Buckets[cNrOfBuckets - 1].m_Box;

	for( int i = cNrOfSplits - 2; i >= 0; --i ) {
		CountAbove[i] = CountAbove[i + 1] + m_Buckets[i + 1].m_Count;
		BoundsAbove[i] = BoundsAbove[i + 1] | m_Buckets[i + 1].m_Box;
	}

	// Find bucket to split at that minimizes our SAH metric
	int MinCostSplitIndex = -1;
	double MinCost = INFINITY;

	for( int i = 0; i < cNrOfSplits; ++i ) {
	  // Compute cost for candidate split and update minimum
		if( CountBelow[i] == 0 || CountAbove[i] == 0 )
			continue;
		double CurCost = (CountBelow[i] * BoundsBelow[i].surface() + CountAbove[i] * BoundsAbove[i].surface());

		if( CurCost < MinCost ) {
			MinCost = CurCost;
			MinCostSplitIndex = i;
		}
	}

  // determine pivot coordinate based on MinCostSplitIndex that identifies the splitting bucket
  Pivot = CentroidBBox.low()[Axis] + CentroidBBox.size()[Axis] / (cNrOfBuckets * 1.0) * (MinCostSplitIndex + 1);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

BBox CSAHBoundingVolHierarchy::InitCentroidBBox( VZone* const* ppNodeList, const size_t NrOfNodes )

// calculate BBox of all centroids for the given nodes
// [6/23/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	BBox CurBBox;

	CurBBox.null();

	for( size_t i = 0; i < NrOfNodes; ++i ) {
    CurBBox.Union( ppNodeList[i]->bbox().centroid() );
	}

  return CurBBox;
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//                                              Linear bounding volume hierarchies - based on 3D z-order sorting
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CLinearBoundingVolHierarchy::CLinearBoundingVolHierarchy( std::vector<VZone*> ZoneList, bool bVerbose /*= false*/, int LogLevel /*= 0 */) :
																													CBoundingVolHierarchy( ZoneList, false, bVerbose, LogLevel )

// [7/14/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	// we separate zones without valid bounding boxes from those with valid bounding boxes.
	// treelets are only built for those with valid bounding boxes, while those with invalid bounding boxes
	// are processed separately.

	m_CentroidBBox.null();

	assert( !ZoneList.empty() );
	if( !ZoneList.empty() ) {

		for( size_t i = 0; i < ZoneList.size(); ++i ) {
			if( ZoneList[i]->bbox().isValid() ) {
				m_CentroidBBox.Union( ZoneList[i]->bbox().centroid() );
				m_pValidZones.push_back( ZoneList[i] );
				m_NrOfValidZones++;
			}
			else
				m_pInvalidBBoxZones.push_back( ZoneList[i] );
		}
	}

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CLinearBoundingVolHierarchy::~CLinearBoundingVolHierarchy( void )

// [7/14/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	// as the sub-trees of the treelets can be shared by different copies of the LBVHs used in different threads,
	// they will be released via reference counted smart pointers!

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CLinearBoundingVolHierarchy::CLinearBoundingVolHierarchy( const CLinearBoundingVolHierarchy& rhs ) : CBoundingVolHierarchy( rhs )

// [7/23/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	*this = rhs;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CLinearBoundingVolHierarchy& CLinearBoundingVolHierarchy::operator=( const CLinearBoundingVolHierarchy& rhs )

// [7/23/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	if( this != &rhs ) {
		m_pValidZones = rhs.m_pValidZones;
		m_pTreelets = rhs.getTreelets();
	}

	return *this;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CLinearBoundingVolHierarchy::BuildTree()

// [7/14/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	// calculate 3D morton index for each zone
	// as we want to create clusters it is sufficient to use 10 bits and 32 bit floats for the coordinates. Thus, 3x10 bits nicely fit into a 32 bit data-type
	std::vector<CMortonPrimitive> MortonPrimitives( m_pValidZones.size() );

	for( size_t i = 0; i < m_pValidZones.size(); ++i ) {

		constexpr int MortonBits = 10;
		constexpr int MortonScale = 1 << MortonBits;
		MortonPrimitives[i].m_PrimitiveIndex = i;
		MortonPrimitives[i].m_pZone = m_pValidZones[i];

		Vector CentroidOffset = m_CentroidBBox.offset( m_pValidZones[i]->bbox().centroid() );
		Vector CurOffset = CentroidOffset * MortonScale;

		MortonPrimitives[i].m_MortonCode = Morton3::encode3D<unsigned, float>( (float)CurOffset.x, (float)CurOffset.y, (float)CurOffset.z );
	}

	// sort our valid zones by morton index
	// NOTE: the std sort might not be the fastest (depending on the implementation). a bucket radix sort
	//			 might yield better performance, but for the moment we leave it like this.
	std::sort( MortonPrimitives.begin(), MortonPrimitives.end(),
						 [] ( const CMortonPrimitive& lhs, const CMortonPrimitive& rhs ) -> bool {
								return lhs.m_MortonCode < rhs.m_MortonCode;
							} );

	// determine clusters based on morton index
	// We find clusters by checking the 12 high bits. Zones which are close have similar high bits
	// and we split between clusters when there is a change in those.
	// This practically means that we have 2^4 values per dimension (total 2^12) which is similar to
	// having 16 x 16 x 16 = 4096 strata into which we sort.
	//
	// NOTE: This loop could be parallelized in case of performance needs
	//
	for( unsigned int StartIndex = 0, EndIndex = 1; EndIndex <= MortonPrimitives.size(); ++EndIndex ) {
		// unsigned Mask = 0b00111111111111000000000000000000;
		unsigned Mask = 0x3FFC0000; // this is the equivalent to 0b00111111111111000000000000000000, but binary literals are only supported by >= C++14 or as a compiler extension
		if( EndIndex == MortonPrimitives.size() || ((MortonPrimitives[StartIndex].m_MortonCode & Mask) != (MortonPrimitives[EndIndex].m_MortonCode & Mask)) ) {

			// build a treelet for each cluster
			if( m_LogLevel > 0 )
				m_bVerbose && std::cerr << "*** Building cluster [" << StartIndex << " -> " << EndIndex << "]" << std::endl;

			std::shared_ptr<CLBVHTreelet> pCurTreelet( new CLBVHTreelet );

			for( unsigned int i = StartIndex; i < EndIndex; ++i ) {
				pCurTreelet->m_ZoneList.push_back( MortonPrimitives[i].m_pZone );
			}
			// NOTE: this can be any kind of support BVH type (std., SAH, etc.)
			//pCurTreelet->m_pNodes = new CBoundingVolHierarchy( pCurTreelet->m_ZoneList, false, m_bVerbose, m_LogLevel );
			pCurTreelet->m_pNodes = new CSAHBoundingVolHierarchy( pCurTreelet->m_ZoneList, false, m_bVerbose, m_LogLevel );
			pCurTreelet->m_pNodes->BuildTree();

			// store treelet with built sub-tree.
			// NOTE: for the time being we leave this is a linear array that is resorted depending on the use. But one could build
			//			an SAH BVH out of these, which in overall would result in a well balanced tree
			m_pTreelets.emplace_front( pCurTreelet );

			StartIndex = EndIndex;
		}
	}

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CLinearBoundingVolHierarchy::inside( GeometryEngine* pEngine, const Vector& CurPoint, const Vector& CurDir, double t, double eps /*= SMALL3D4*/ )

// Check if the point CurPoint + t * CurDir is inside the BVH and return the corresponding zone or a nullptr
//
// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// [7/22/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	VZone* pCurZone = nullptr;

	for( auto iter = m_pTreelets.begin(); iter != m_pTreelets.end(); ++iter ) {
		pCurZone = (*iter)->m_pNodes->inside( pEngine, CurPoint, CurDir, t, eps );
		if( pCurZone ) {
			if( iter != m_pTreelets.begin() ) {
				auto pCurTreelet = *iter;
				// promote treelet
				iter = m_pTreelets.erase( iter );
				m_pTreelets.push_front( pCurTreelet );
			}

			return pCurZone;
		}
	}
	// last resort
	// check the zones with invalid bboxes
	return insideInvalidZones( pEngine, CurPoint, CurDir, t, eps );
	//return pCurZone;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CLinearBoundingVolHierarchy::intersectRay( GeometryEngine* pEngine, Ray* pCurRay )

//
// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// intersect a ray with the BVH and return nullptr or the intersected zone
// [7/14/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tmin = pCurRay->T();
	Vector InvDir( pCurRay->dir().x != 0.0 ? 1.0 / pCurRay->dir().x : 0.0,
								 pCurRay->dir().y != 0.0 ? 1.0 / pCurRay->dir().y : 0.0,
								 pCurRay->dir().z != 0.0 ? 1.0 / pCurRay->dir().z : 0.0 );

	// check if the current ray already has a potential intersection candidate

	VZone* pCurZone = pCurRay->segment().zone;
	if( pCurZone ) {
		if( pCurZone->inside( pEngine, pCurRay->pos().x, pCurRay->pos().y, pCurRay->pos().z, pCurRay->dir().x, pCurRay->dir().y, pCurRay->dir().z, tmin ) ) {
			return pCurZone;
		}
	}

	pCurZone = CBoundingVolHierarchy::intersectRay( pEngine, pCurRay, InvDir, tmin, pCurRay->segment().tmax );
	if( pCurZone )
		return pCurZone;

	// last resort
	return intersectRayWithInvalidZones( pEngine, pCurRay );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CLinearBoundingVolHierarchy::intersect( GeometryEngine* pEngine, const Vector& Pos, const Vector& Dir, double& tmin, double tmax, VZone* pZone /*= nullptr */)

//
// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// [7/14/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	VZone* pCurZone = nullptr;

	for( auto iter : m_pTreelets ) {
		pCurZone = iter->m_pNodes->intersect( pEngine, Pos, Dir, tmin, tmax, pZone );

		if( pCurZone ) {
			return pCurZone;
		}
	}

	// last resort
	// check the zones with invalid bboxes
	return intersectWithInvalidZones( pEngine, Pos, Dir, tmin, tmax, pZone );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CLinearBoundingVolHierarchy::PrintTree( const std::string& FileName /*=""*/ )

// [7/14/2021 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	std::streambuf* coutbuf = nullptr;

	if( !FileName.empty() ) {
		// save current buffer of cerr
		coutbuf = std::cerr.rdbuf();

		std::ofstream out( FileName );
		std::cerr.rdbuf( out.rdbuf() );                     //redirect cerr
	}

	unsigned int index = 0;
	for( auto iter : m_pTreelets ) {
		std::cerr << "+----------------------------------------------------------------------------------------------------+" << std::endl;
		std::cerr << "TREELET index #" << index++ << std::endl << std::endl;
		iter->m_pNodes->PrintTree();
		std::cerr << "+----------------------------------------------------------------------------------------------------+" << std::endl;
	}



	std::cerr << "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;
	std::cerr << "++ Passed nr. of zones:" << m_pNodeList.size() << std::endl;
	std::cerr << "++ Processed nr. of zones in LBVH: " << GetTotalNrOfZones() << ", valid bboxes: " << GetNrOfValidZones() << ", invalid bboxes: " << GetNrOfInvalidZones() << std::endl;
	std::cerr << "++ Number of treelets: " << std::distance( m_pTreelets.begin(), m_pTreelets.end() ) << std::endl;
	std::cerr << "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;

	if( !FileName.empty() ) {
		std::cerr.rdbuf( coutbuf );   //reset to standard output again
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// BVH for bodies
// [WIP] to be cleaned up and extended to SAH once we have proof that it pays off performance wise (CT 5.7.22)

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBodyBVH::CBodyBVH( std::vector<CBody*> BodyList,  bool bVerbose /* = false */, int LogLevel /*= 0*/ ) :
                                              m_pLeftLeaf(nullptr), m_pRightLeaf(nullptr), m_pBody(nullptr), m_NrOfLeaves(0), m_bVerbose( bVerbose ), m_LogLevel( LogLevel ),
                                              m_pNodeList( BodyList ), m_bTreeBuilt(false), m_bIncludeInvalidBodies(true)

// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	m_bVerbose&& std::cerr << "*** Standard bounding volume hierarchy tree selected ***" << std::endl;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBodyBVH::BuildTree()

// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Vector CurBoxMin, CurBoxMax, NextBoxMin, NextBoxMax, TmpBoxMin, TmpBoxMax;
	double Pivot;
	int    Axis = cXAxis;

	assert( !m_pNodeList.empty() );

	m_bVerbose&& std::cerr << "BVH: starting construction for " << m_pNodeList.size() << std::endl;
	if( m_pNodeList.size() < 1 )
		// throw std::runtime_error( "BVH: insufficient geometry regions to build BV tree" );
		return;

	if( m_pNodeList.size() == 1 ) {
		*this = CBodyBVH( m_pNodeList[0] );
	}

	if( m_pNodeList.size() >= 2 ) {
		// Obtain the centroid of the first bounding box that will be used as a pivot for the qsplit routine
		BBox CurBBox = m_pNodeList[0]->body()->bbox();
		if( !CurBBox.isValid() ) {
				m_pInvalidBBoxBodies.push_back( m_pNodeList[0] );
		}


		for( size_t i = 1; i < m_pNodeList.size(); ++i ) {
			BBox NextBBox = m_pNodeList[i]->body()->bbox();

			if( !NextBBox.isValid() ) {
				m_pInvalidBBoxBodies.push_back( m_pNodeList[i] );
			}
			else {
				CurBBox = CurBBox | NextBBox;
			}
		}

		// set up the bounding box of the whole union
		m_BoundingBox = CurBBox;

		// get position and axis along which we should split
		GetSplittingPos( m_pNodeList.data(), m_pNodeList.size(), Pivot, Axis );
		Axis = cXAxis;	// for the first generation we enforce splitting of the x-axis and subsequent calls to GetSplittingPos() will iterate over the axes
		size_t MidPoint = QSplit( m_pNodeList.data(), m_pNodeList.size(), Pivot, Axis);

		// Create a new bounding volume & define the next axis as the new split axis
		m_pLeftLeaf = BuildBranch( m_pNodeList.data(), MidPoint, Axis );
		m_pRightLeaf = BuildBranch( &m_pNodeList.data()[MidPoint], m_pNodeList.size() - MidPoint, Axis );
		m_pBody = nullptr;
	}

	m_bVerbose&& std::cerr << "BVH - bodies: Total nr of tree leaves = " << m_NrOfLeaves << std::endl;
	m_bVerbose&& std::cerr << "BVH - bodies: BBox (" << m_BoundingBox.low() << ") -> (" << m_BoundingBox.high() << ")" << std::endl;

	m_bTreeBuilt = true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBodyBVH::CBodyBVH( CBody* pBody ) : m_pLeftLeaf( nullptr), m_pRightLeaf( nullptr),
									m_pBody( pBody ), m_NrOfLeaves( 1 ), m_NrOfValidBodies( 1 ), m_bVerbose( false ), m_LogLevel( 0 ),
									m_bTreeBuilt(true),  m_bIncludeInvalidBodies(true)

// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  m_BoundingBox = m_pBody->body()->bbox();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBodyBVH::CBodyBVH( CBodyBVH* pNode1, CBodyBVH* pNode2 ) : 	m_pLeftLeaf(pNode1), m_pRightLeaf( pNode2), m_pBody(nullptr),
															m_NrOfLeaves( 1 ), m_NrOfValidBodies( 1 ), m_bVerbose( false ), m_LogLevel( 0 ), m_bTreeBuilt( true ),
															 m_bIncludeInvalidBodies(true)

// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector Box1Min, Box2Min, Box1Max, Box2Max;

  if( m_pRightLeaf ) {
    m_BoundingBox = pNode1->bbox() | pNode2->bbox();
		m_NrOfLeaves++;
		m_NrOfValidBodies++;
  }
  else
    m_BoundingBox = pNode1->bbox();
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBodyBVH* CBodyBVH::BuildBranch( CBody** ppNodeList, size_t NrOfSubBranchNodes, int Axis /* = cXAxis */ )

// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  double Pivot;

  m_NrOfLeaves++;

  if( NrOfSubBranchNodes == 1) {
		m_NrOfValidBodies++;
		return new CBodyBVH( ppNodeList[0] );
	}

  // determine the splitting axis and position on that axis
  GetSplittingPos( ppNodeList, NrOfSubBranchNodes, Pivot, Axis );

  // let's split according to the respective parameters
  size_t  MidPoint = QSplit( ppNodeList, NrOfSubBranchNodes, Pivot, Axis);

  if( m_bVerbose && m_LogLevel >= 1 ) {
    std::cerr << "BVH - bodies: Splitting axis = " << Axis << ", pivot = " << Pivot << " splitting " << MidPoint << "/" << NrOfSubBranchNodes - MidPoint << std::endl;
  }

  // assemble a new bounding volume iterating the splitting axis
  CBodyBVH* pLeftLeaf = BuildBranch( ppNodeList, MidPoint, Axis);
  CBodyBVH* pRightLeaf = BuildBranch( &ppNodeList[MidPoint], NrOfSubBranchNodes - MidPoint, Axis );


  return new CBodyBVH( pLeftLeaf, pRightLeaf);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

size_t  CBodyBVH::QSplit( CBody** ppNodeList, size_t NrOfNodes, double PivotVal, int Axis)

// determine the number of nodes from a subset (given by ppStartNode + NrOfNodes) that have a centroid
// coordinate < PivotVal with respect to a specific axis.
// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector   BoxMin, BoxMax;
  double   Centroid = 0.0;
  size_t   RetVal  = 0;

  for( size_t i = 0; i < NrOfNodes; ++i) {

    BBox CurBBox = ppNodeList[i]->body()->bbox();

    Centroid = CurBBox.centroid( Axis );
    if( Centroid < PivotVal ) {
      std::swap( ppNodeList[i], ppNodeList[RetVal] );
      RetVal++;
    }
  }

  if( RetVal == 0 || RetVal == NrOfNodes )
    RetVal = NrOfNodes / 2;

  return RetVal;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBody* CBodyBVH::Contains( const Vector& CurPoint, const Vector& CurDir, double t, double eps )

// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
    CBody* pLeftHit = nullptr;
	CBody* pRightHit = nullptr;

	// branch into further BHVs
	// TODO: check if we should not better use a ray intersection for better FP accuracy

	if( !m_BoundingBox.inside( CurPoint + CurDir * t, eps ) )
		return nullptr;

	// check if we have are region primitive, otherwise branch further
  if( m_pBody ) {
		if( m_pBody->inside( CurPoint, CurDir, t ) )
			return m_pBody;
		return nullptr;
  }
	else if( m_pLeftLeaf ) {
    pLeftHit = static_cast<CBodyBVH*>(m_pLeftLeaf)->Contains( CurPoint, CurDir, t, eps );
		if( pLeftHit )
			return pLeftHit;
  }
	else if( m_pRightLeaf  ) {
    pRightHit = static_cast<CBodyBVH*>(m_pRightLeaf)->Contains( CurPoint, CurDir, t, eps );
  }

	return pRightHit;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBody* CBodyBVH::inside( const Vector& CurPoint, const Vector& CurDir, double t, double eps /* = SMALL3D4 */ )

// Check if the point CurPoint + t * CurDir is inside the BVH and return the corresponding zone or a nullptr
//
// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  if( m_NrOfLeaves == 0 )
    return nullptr;

  CBody* pCurZone = Contains( CurPoint, CurDir, t, eps );
	if( !pCurZone && m_bIncludeInvalidBodies ) {					// check bodies without valid BBOX as last resort
		pCurZone = insideInvalidBodies( CurPoint, CurDir, t, eps );
	}

	return pCurZone;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBodyBVH::PrintTree( const std::string& FileName /*=""*/)

// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	std::streambuf* coutbuf = nullptr;

	if( !FileName.empty() ) {
			// save current buffer of cerr
		coutbuf = std::cerr.rdbuf();

		std::ofstream out( FileName);
		std::cerr.rdbuf( out.rdbuf() );                     //redirect cerr
	}

  std::cerr << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;
  std::cerr << "|             Bounding volume hierarchies of BODIES        |" << std::endl;
	std::cerr << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;
  std::cerr << " Nr of bodies = " << m_pNodeList.size() << ", nr of leaves = " << m_NrOfLeaves << std::endl;
	std::cerr << " Nr of bodies with valid bboxes = " << GetNrOfValidZones() << ", nr of bodies with invalid bounding boxes = " << std::distance( m_pInvalidBBoxBodies.begin(), m_pInvalidBBoxBodies.end() ) << std::endl << std::endl;

	if( m_pLeftLeaf ) {
		m_pLeftLeaf->PrintLeaf();
	}

	if( m_pRightLeaf ) {
      m_pRightLeaf->PrintLeaf();
	}

	if( m_pBody ) {
		std::cerr << "--> #" << 0 << " - L-GEO: " << m_pBody->body()->name() << " bmin =" << m_pBody->body()->bbox().low() << ", bmax = " << m_pBody->body()->bbox().high() << ", pointer =" << m_pBody << std::endl;
	}

  std::cerr << "================================================================" << std::endl;
  std::cerr << "Nr. of bodies with invalid bounding boxes: " << std::distance( m_pInvalidBBoxBodies.begin(), m_pInvalidBBoxBodies.end() ) << std::endl;
  if( !m_pInvalidBBoxBodies.empty() ) {
    std::cerr << "Zones with invalid bounding boxes:" << std::endl;
		for( auto iter : m_pInvalidBBoxBodies ) {
			std::cerr << iter->body()->name() << std::endl;
		}
	}

	std::cerr << "!! Safe mode which checks also bodies with invalid bounding boxes (might be slow!): " << std::boolalpha << m_bIncludeInvalidBodies << std::endl;

	if( !FileName.empty() ) {
		std::cerr.rdbuf( coutbuf );   //reset to standard output again
	}

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBodyBVH::PrintLeaf( int Level /*= 0*/ )

// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	std::cerr << "+";

  Indent( Level );
  std::cerr << "--> #" << Level << "  BB " << " bmin =" << m_BoundingBox.low() << ", bmax = " << m_BoundingBox.high() << std::endl;

  if( m_pLeftLeaf ) {
		Indent( Level + 1 );
    std::cerr << "|" << std::endl;
    m_pLeftLeaf->PrintLeaf( Level + 1 );
  }

  if( m_pRightLeaf ) {
    Indent( Level + 1 );
		std::cerr << "|" << std::endl;
    m_pRightLeaf->PrintLeaf( Level + 1 );
  }

	if( m_pBody ) {
		Indent( Level + 1 );
		std::cerr << "--> #" << Level + 1 << " - L-GEO: " << m_pBody->body()->name() << " bmin =" << m_pBody->body()->bbox().low() << ", bmax = " << m_pBody->body()->bbox().high() << ", pointer = " << m_pBody << std::endl;
	}

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBody* CBodyBVH::distance( Ray* pCurRay )

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  double tmin = pCurRay->T();
  Vector InvDir( pCurRay->dir().x != 0.0 ? 1.0 / pCurRay->dir().x : 0.0,
                 pCurRay->dir().y != 0.0 ? 1.0 / pCurRay->dir().y : 0.0,
                 pCurRay->dir().z != 0.0 ? 1.0 / pCurRay->dir().z : 0.0 );

  return distance( pCurRay, InvDir, tmin, pCurRay->segment().tmax );

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBody* CBodyBVH::distance( const Vector& Pos, const Vector& Dir, double& tmin, double& tmax )

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// ATTENTION: We have to pass the corresponding geometry engine as each thread has its own and they all share the current BVH!
// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Vector InvDir( Dir.x != 0.0 ? 1.0 / Dir.x : 0.0,
								 Dir.y != 0.0 ? 1.0 / Dir.y : 0.0,
								 Dir.z != 0.0 ? 1.0 / Dir.z : 0.0 );

	return distance( Pos, Dir, InvDir, tmin, tmax );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBody* CBodyBVH::distance( const Vector& Pos, const Vector& Dir, const Vector& InvDir, double& tmin, double& tmax )

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// ATTENTION: We have to pass the corresponding geometry engine as each thread has its own and they all share the current BVH!
// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tminLeft = tmin, tmaxLeft = tmax, tminRight = tmin, tmaxRight = tmax;
	CBody* pLeftBody = nullptr;
	CBody* pRightBody = nullptr;

	// check if we have an actual body to intersect
	if( m_pBody ) {
		// find the closest hit, if any
		bool bHit = m_pBody->distance(Pos, Dir );
		if( !bHit && m_bIncludeInvalidBodies ) {
			return distanceInvalidBodies( Pos, Dir, tmin, tmax );
		}
		else {
			tmin = m_pBody->tmin;
			tmax = m_pBody->tmax;
			return m_pBody;
		}
	}

	// check bbox and leaves
	if( !m_BoundingBox.intersectRay( Pos, InvDir, tmin, tmax ) )
		return nullptr;

	// branch into tree
	if( m_pLeftLeaf ) {
		pLeftBody = m_pLeftLeaf->distance(  Pos, Dir, InvDir, tminLeft, tmaxLeft );
	}

	if( m_pRightLeaf ) {
		pRightBody = m_pRightLeaf->distance( Pos, Dir, InvDir, tminRight, tmaxRight );
	}

	// return the closer hit
	if( pRightBody && tminRight < tminLeft ) {
		tmin = pRightBody->tmin;
		tmax = pRightBody->tmax;
		return pRightBody;
	}

	tmin = pLeftBody->tmin;
	tmax = pLeftBody->tmax;
	return pLeftBody;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBody* CBodyBVH::distance( Ray* pCurRay, const Vector& InvDir, double& tmin, double& tmax )

// ATTENTION: We have to pass the corresponding geometry engine as each thread has its own and they all share the current BVH!
// optimized intersection that avoids pre-calculating constants like the inverse direction
//
// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	auto pBody = distance( pCurRay->pos(), pCurRay->dir(), InvDir, tmin, tmax );

	if( !pBody )
		return nullptr;

	if( !pBody->zones.empty() )
		pCurRay->segment().zone = pBody->zones[0];
	else
		pCurRay->segment().zone = nullptr;

	pCurRay->segment().body = pBody;
	pCurRay->segment().tmin = pBody->tmin;
	pCurRay->segment().tmax = pBody->tmax;

	return pBody;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBodyBVH::GetSplittingPos( CBody * const* ppNodeList, const size_t NrOfNodes, double& Pivot, int& Axis )

// retrieve splitting axis and point on this axis.
// Default implementation is based on simplistic iteration over the splitting axes. Yet, this could be improved via
// surface area heuristics or 3D sorting with Morton indices etc. TODO: check paper on NVidia/Weta's PantaRay
//
// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  BBox   CurBBox, NextBBox;

  // Obtain the centroid of the bounding box that will be used as a pivot for the qsplit routine
	CurBBox = ppNodeList[0]->body()->bbox();

  for( size_t i = 1; i < NrOfNodes; ++i ) {
		NextBBox = ppNodeList[i]->body()->bbox();
		CurBBox = CurBBox | NextBBox;
	}

	// for each tree generation we iterate over the splitting axes; this is an extremely simple approach but Shirley et al. showed that
	// while it does not yield optimum performance, tree balancing is sufficient and tree generation costs are kept at bay.
	Axis = (Axis + 1) % 3;
	Pivot = CurBBox.centroid( Axis );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBody* CBodyBVH::distanceInvalidBodies( Ray* pCurRay )

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// intersect ray with regions that have invalid bounding boxes but are still stored in the tree as they are part of the original geometry
// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tmin = pCurRay->T(), tmax = pCurRay->segment().tmax;
	CBody* pBody = nullptr;

	pBody = distanceInvalidBodies( pCurRay->pos(), pCurRay->dir(), tmin, tmax );

	if( pBody ) {
		if( !pBody->zones.empty() )
			pCurRay->segment().zone = pBody->zones[0];
		else
			pCurRay->segment().zone = nullptr;

		pCurRay->segment().body = pBody;
		pCurRay->segment().tmin = pBody->tmin;
		pCurRay->segment().tmax = pBody->tmax;
	}

	return pBody;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBody* CBodyBVH::distanceInvalidBodies( const Vector& Pos, const Vector& Dir, double& tmin, double& tmax )

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// ATTENTION: We have to pass the corresponding geometry engine as each thread has its own and they all share the current BVH!
// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	double tminOld  = tmin;
	double tmaxOld  = tmax;
	CBody* pCurBody = nullptr;

	// loop over bodies to find the closest hit, if any
	for( auto iter = m_pInvalidBBoxBodies.begin(); iter != m_pInvalidBBoxBodies.end(); ++iter ) {
		bool bHit = (*iter)->distance( Pos, Dir);
		if( bHit && (*iter)->tmin < tminOld ) {
			pCurBody = *iter;
			tminOld = (*iter)->tmin;
			tmaxOld = (*iter)->tmax;
		}
	}

	if( pCurBody && pCurBody->tmin >= tmin && tminOld < tmax ) {
		tmin = tminOld;
		tmax = tmaxOld;
		return pCurBody;
	}

	return nullptr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBody* CBodyBVH::insideInvalidBodies( const Vector& Pos, const Vector& Dir, double t, double eps /*= SMALL3D4 */ )

// NOTE: This function is not thread-safe as it modifies the order of an internal array!! If used in threads then each thread must keep its own
//			 copy of this object.
//
// ATTENTION: We have to pass the corresponding geometry engine as each thread has its own and they all share the current BVH!
// [7/05/2022 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( eps );

	for( auto iter = m_pInvalidBBoxBodies.begin(); iter != m_pInvalidBBoxBodies.end(); ++iter ) {
		if( (*iter)->inside(  Pos.x, Pos.y, Pos.z, Dir.x, Dir.y, Dir.z, t ) ) {
			return *iter;
		}
	}

	return nullptr;
}


// vim: ts=2
