/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	24-Feb-2021
 */

#include "conebody.h"

//#include "mesh.h"

// ================================= GTRCBody ==================================
/** set */
void GTRCBody::set(const double x,  const double y,  const double z,
		   const double hx, const double hy, const double hz,
		   const double rb, const double ra)
{
	P.set( x,  y,  z);
	Z.set(hx, hy, hz);

	if (rb<0.0) throw std::string("Negative base radius");
	if (ra<0.0) throw std::string("Negative apex radius");
	if (ra<TOOSMALL && rb<TOOSMALL) throw std::string("Both apex and base radius are too small");
	Rb(Max(rb,0.0));	// Rb base radius
	Ra(Max(ra,0.0));	// Ra apex radius

	Po   = P + Z;
	h2   = Z.length2();	// to keep precision
	zlen = Z.normalize();
	if (zlen<TOOSMALL) throw std::string("Zero height");

	findXYZ();
} // set

/** set */
void GTRCBody::set(int /*n*/, double* what)
{
	resetBBox();
	round0(what, 8, SMALL);
	set( what[0], what[1], what[2],
	     what[3], what[4], what[5],
	     what[6], what[7]);
} // set

/* --- get --- */
int GTRCBody::get(double* what) const
{
	what[0] = P.x;
	what[1] = P.y;
	what[2] = P.z;
	what[3] = zlen*Z.x;
	what[4] = zlen*Z.y;
	what[5] = zlen*Z.z;
	what[6] = Rb();
	what[7] = Ra();
	return 8;
} // get

/**  @return closest point, edge, face */
int GTRCBody::closest(const Point& r, const double d, const Vector& w) const
{
	int close = GBody::closest(r,d,w);
	if (close == 1) {
		// Check which is closer, base or apex?
		if (Abs(Q(1)(r)) > Abs(Q(2)(r)))
			return 4;
	}
	return close;
} // closest

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GTRCBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point T = _hasMatrix? _invMatrix*r : r;
	T -= SP;
	switch (item) {
		case 1:	// Find distance from Z-axis
			Rb(Sqrt(T.length2() - Sq(T*Z)));	// base radius
			break;

		case 2: {	// Move base
				double zp = T*SZ;
				P = SP + zp*SZ;
				zlen = szlen - zp;
				if (zlen<0.0) {
					Z = -SZ;
					zlen = -zlen;
				} else
					Z = SZ;
				h2 = Sq(zlen);
			}
			break;

		case 3:		// Move apex
			zlen = T*SZ;
			if (zlen<0.0) {
				Z = -SZ;
				zlen = -zlen;
			} else
				Z = SZ;
			h2 = Sq(zlen);
			Po = P + zlen*Z;
			break;

		case 4:		// Change apex radius
			Ra(Sqrt(T.length2() - Sq(T*Z)));
			break;

		default:
			GBody::move(item, r, w);
			h2 = Sq(zlen);
	}
} // move

/** createQuads */
void GTRCBody::createQuads()
{
	GBody::createQuads();

	if (ISZERO(zlen)) return;

	// Create the cone around Z then transform
	//  Cxx Cyy Czz  Cxy  Cxz  Cyz  Cx   Cy   Cz    C
	//double s = Ra()-Rb();	// slope=(Ra-Rb)/zlen but avoid division that generates roundup errors
	//addQuad(1.0, 1.0, -s*s/h2, 0.0, 0.0, 0.0, 0.0, 0.0,-2.0*s*Rb()/zlen, -Rb()*Rb());
	//makeConeQuads();
	addQuad(Quad(Quad::Type::CONE, P, zlen*Z, Rb(), Ra()));
	addQuad(-Z.x,-Z.y,-Z.z, Z*P);	// base
	addQuad( Z.x, Z.y, Z.z,-Z*Po);	// apex
} // createQuads

/** @return bounding box */
void GTRCBody::_updateBBox()
{
	Point p  = position();
	Point po = p + zlen*vectorZ();

	_bbox.add(p);
	_bbox.add(po);
	bbox_addRotatedEllipse(p,  Rb(), Rb());
	bbox_addRotatedEllipse(po, Ra(), Ra());
} // bbox

#ifdef USE_OBBOX
/** update oriented bounding box of body */
void GTRCBody::_updateOBB(bool in)
{
	OBBox& bb = _obbox[(int)in];

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();

	if (!in) {
		double R = Max(Ra(), Rb());
		bb.low()  = Point(-R, -R, 0.0);
		bb.high() = Point( R,  R, zlen);
	} else {
		double sc, z;
		if (Rb() > Ra()) {
			sc = Rb() / 2.0;
			if (Ra() >= sc) {
				sc = Ra();
				z  = zlen;
			} else
				z  = sc*zlen/(Rb()-Ra());
			sc /= Sqrt2;
			bb.low()  = Point(-sc, -sc, 0.0);
			bb.high() = Point( sc,  sc, z);
		} else {
			sc = Ra() / 2.0;
			if (Rb() >= sc) {
				sc = Rb();
				z  = 0.0;
			} else
				z = zlen - sc*zlen/(Ra()-Rb());
			sc /= Sqrt2;
			bb.low()  = Point(-sc, -sc, z);
			bb.high() = Point( sc,  sc, zlen);
		}
	}
} // _updateOBB
#endif

// ================================= GConeBody =================================
/** set */
void GConeBody::set(int /*n*/, double* what)
{
	resetBBox();
	if (In(type(), Type::KX, Type::KY, Type::KZ)) {
		round0(what, 3, SMALL);
		switch (type()) {
			case Type::KX:
				P.set(what[0], 0, 0);
				break;
			case Type::KY:
				P.set(0, what[0], 0);
				break;
			case Type::KZ:
				P.set(0, 0, what[0]);
				break;
			default:
				assert(false);
		}
		t2(what[1]);
		k = Cmp0(what[2],SMALL);
	} else {
		round0(what, 5, SMALL);
		P.set(what[0], what[1], what[2]);
		t2(what[3]);
		k = Cmp0(what[4],SMALL);
	}
	switch (type()) {
		case Type::KX:
		case Type::K_X:
			Z = Vector::Xo;
			X = Vector::Yo;
			Y = Vector::Zo;
			break;
		case Type::KY:
		case Type::K_Y:
			Z = Vector::Yo;
			X = Vector::Zo;
			Y = Vector::Xo;
			break;
		case Type::KZ:
		case Type::K_Z:
			X = Vector::Xo;
			Y = Vector::Yo;
			Z = Vector::Zo;
			break;
		default:
			assert(false);
	}
	if (t2() <= 0.0) throw std::string("Zero aperture");
} // set

/* --- get --- */
int GConeBody::get(double* what) const
{
	switch (type()) {
		case Type::KX:
			what[0] = P.x;
			what[1] = t2();
			what[2] = k;
			return 3;
		case Type::KY:
			what[0] = P.y;
			what[1] = t2();
			what[2] = k;
			return 3;
		case Type::KZ:
			what[0] = P.z;
			what[1] = t2();
			what[2] = k;
			return 3;

		case Type::K_X:
		case Type::K_Y:
		case Type::K_Z:
			what[0] = P.x;
			what[1] = P.y;
			what[2] = P.z;
			what[3] = t2();
			what[4] = k;
			return 5;
		default:
			return -1;
	}
} // get

/**  @return closest point, edge, face */
int GConeBody::closest(const Point& r, const double d, const Vector& w) const
{
	int close = GBody::closest(r,d,w);
	return close;
#if 0
	switch (type()) {
		case Type::XEC:
			if (Abs(r.y-P.y) < Abs(r.z-P.z))
				return 2;
			return close;
		case Type::YEC:
			if (Abs(r.z-P.z) < Abs(r.x-P.x))
				return 2;
			return close;
		case Type::ZEC:
			if (Abs(r.x-P.x) < Abs(r.y-P.y))
				return 2;
			return close;
		default:
			return close;
	}
#endif
} // closest

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GConeBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point T = _hasMatrix? _invMatrix*r : r;
	T -= SP;

	if (item == 1) {
		double a, b;

		switch (type()) {
			case Type::K_X:
			case Type::KX:
				a = Sq(T.x);
				b = Sq(T.y) + Sq(T.z);
				break;

			case Type::K_Y:
			case Type::KY:
				a = Sq(T.y);
				b = Sq(T.x) + Sq(T.z);
				break;

			case Type::K_Z:
			case Type::KZ:
				a = Sq(T.z);
				b = Sq(T.x) + Sq(T.y);
				break;

			default:
				a = b = 0.0;	// Never comes here but keeps compiler happy
				assert(false);
		}
		if (a!=0.0)
			xlen = Abs(b/a);
	} else {
		GBody::move(item, r, w);

		switch (type()) {
			case Type::KX:
				if (P.y!=0.0 or P.z!=0.0)
					_type = Type::K_X;
				break;
			case Type::KY:
				if (P.x!=0.0 or P.z!=0.0)
					_type = Type::K_Y;
				break;
			case Type::KZ:
				if (P.x!=0.0 or P.y!=0.0)
					_type = Type::K_Z;
				break;
			case Type::K_X:
				if (P.y==0.0 and P.z==0.0)
					_type = Type::KX;
				break;
			case Type::K_Y:
				if (P.x==0.0 and P.z==0.0)
					_type = Type::KY;
				break;
			case Type::K_Z:
				if (P.x==0.0 and P.y==0.0)
					_type = Type::KZ;
				break;
			default: NOP;
		}
	}
} // move

/** rotate body around an axis */
void GConeBody::rotate(const double angle, const Vector& axis)
{
	GBody::rotate(angle, axis);

	/* find closest axis to align to */
	double ax = Abs(Z.x);
	double ay = Abs(Z.y);
	double az = Abs(Z.z);

	if (az>ax && az>ay) {
		if (Z.z<0.0 and k<0)
			k = 1;
		else
		if (Z.z>0.0 and k>0)
			k = -1;
		Z = Vector::Zo;
		_type = Type::K_Z;
	} else
	if (ay>ax && ay>az) {
		if (Z.y<0.0 and k<0)
			k = 1;
		else
		if (Z.y>0.0 and k>0)
			k = -1;
		Z = Vector::Yo;
		_type = Type::K_Y;
	} else {
		if (Z.x<0.0 and k<0)
			k = 1;
		else
		if (Z.x>0.0 and k>0)
			k = -1;
		Z = Vector::Xo;
		_type = Type::K_X;
	}

	findXYZ();
} // rotate

/** createQuads */
void GConeBody::createQuads()
{
	GBody::createQuads();

	switch (type()) {
		case Type::K_X:
		case Type::KX:
			addQuad(Quad(Quad::Type::CONE_X, P, sqrt(t2())));
			break;
		case Type::K_Y:
		case Type::KY:
			addQuad(Quad(Quad::Type::CONE_Y, P, sqrt(t2())));
			break;
		case Type::K_Z:
		case Type::KZ:
			addQuad(Quad(Quad::Type::CONE_Z, P, sqrt(t2())));
			break;
		default: assert(0);
	}
	if (k>0)
		addQuad(-Z.x,-Z.y,-Z.z,  Z*P);	// positive side only
	else
	if (k<0)
		addQuad( Z.x, Z.y, Z.z, -Z*P);	// negative side only
} // createQuads

/**
 * @param body	other body to check location against
 * @return location of this body with respect to another body
 */
GBody::Location GConeBody::_locationWrt(const GBody* body) const
{
#if 0
	// FIXME TODO
	//if (body->type() == Type::RCC) {
	//} else
	if (isInfCylinder() and body->isInfCylinder()) {
		Vector thisZ = vectorZ();
		Vector bodyZ = body->vectorZ();

		// Minimum distance between the two axes
		double dist = lineLineDistance(position(), thisZ, body->position(), bodyZ);

		// If they are parallel.
		// FIXME could be that after a rotation thisZ = +/- bodyZ
		// FIXME looking for troubles if there is a matrix transformation!!!!
		if (type() == body->type()) {
			if (dist + ((const GConeBody*)body)->R() <  R())
				return Location::BinA;
			else
			if (dist - ((const GConeBody*)body)->R() < -R())
				return Location::AinB;
			else
			if (dist - ((const GConeBody*)body)->R() >  R())
				return Location::OUTSIDE;
			else
				return Location::OVERLAP;
		} else
		// any other skew orientation (after rotation)
		if (dist - ((const GConeBody*)body)->R() >  R())
			return Location::OUTSIDE;
		else
			return Location::OVERLAP;
	}
#endif
	return GBody::_locationWrt(body);
} // _locationWrt

/** @return bounding box of body */
void GConeBody::_updateBBox()
{
	_bbox.infinite();
	if (k!=0)
		switch (type()) {
			case Type::K_X:
			case Type::KX:
				if (k>0)
					_bbox.lowx(P.x);
				else
					_bbox.highx(P.x);
				break;
			case Type::K_Y:
			case Type::KY:
				if (k>0)
					_bbox.lowy(P.y);
				else
					_bbox.highy(P.y);
				break;
			case Type::K_Z:
			case Type::KZ:
				if (k>0)
					_bbox.lowz(P.z);
				else
					_bbox.highz(P.z);
				break;
			default:
				assert(0);
		}
	if (_hasMatrix) _bbox.transform(_matrix);
} // bbox

/** createMesh */
void GConeBody::createMesh()
{
	GBody::createMesh();
#if 0
	// FIXME TODO
	bool first = _mesh->isEmpty();
	const double step = PI2 / (double)N_CYLINDER;
	double ang = 0.0;

	Point A, B, C;
	if (isinfinite) {
		B = P - infinite*Z;
		A = P + infinite*Z;
		C = P;		// additional ring!
	} else {
		B = P;
		A = Po;
	}

	if (first) _mesh->allocateVertices(2+2*N_CYLINDER);

	_mesh->vertex(0) = B;	// base
	_mesh->vertex(1) = A;	// apex

	// Build base
	//if (!Eq0(Rbx,SMALL) and !Eq0(Rby,SMALL))	// ignore Cones with tip
	//if (!Eq0(Rax,SMALL) and !Eq0(Ray,SMALL))	// ignore Cones with tip

	const int base = 2;
	const int apex = 2 + N_CYLINDER;

	for (int i=0; i<N_CYLINDER; i++, ang+=step) {
		double c, s;
		bsincos(ang, &s, &c);
		_mesh->vertex(i+base) = B + (Rbx*c)*X + (Rby*s)*Y;
		_mesh->vertex(i+apex) = A + (Rax*c)*X + (Ray*s)*Y;
	}
	_mesh->calcBbox();

	if (first) {
		// XXX I could do in a single loop
		// but this way I could potentially use the fan opengl commands

		// create base faces
		for (int i=0; i<N_CYLINDER; i++) {
			int in = (i==N_CYLINDER-1? 0 : i+1); // next i
			_mesh->add(0, base+in, base+i, false, true, false);
		//}

		// create apex faces
		//for (int i=0; i<N_CYLINDER; i++) {
		//	int in = (i==N_CYLINDER-1? 0 : i+1); // next i
			_mesh->add(1, apex+i, apex+in, false, true, false);
		//}

		// create side faces
		//for (int i=0; i<N_CYLINDER; i++) {
		//	int in = (i==N_CYLINDER-1? 0 : i+1); // next i
			_mesh->add(base+i, apex+in, apex+i,  false, true, true);
			_mesh->add(base+i, base+in, apex+in, true,  true, false);
		}

		_mesh->process();
		assert(_mesh->isClosed());
		assert(_mesh->isOrientable());
#if _DEBUG>2
		cout << endl;
		cout << "CYL Mesh ";
		cout << " isClosed=" << _mesh->isClosed();
		cout << " isOrientable=" << _mesh->isOrientable() << endl;
		cout << "CYL volume=" << _mesh->volume() << endl;
#endif
	}
#endif
} // createMesh

#ifdef USE_OBBOX
/** update oriented bounding box of body */
void GConeBody::_updateOBB(bool /*in*/)
{
#if 0
	OBBox& bb = _obbox[(int)in];
	bb.infinite();

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();

	if (!in) {
		bb.low().x  = -Rx();
		bb.low().y  = -Ry();
		bb.high().x =  Rx();
		bb.high().y =  Ry();
	} else {
		double sc = 1.0/Sqrt2;
		bb.low().x  = -Rx()*sc;
		bb.low().y  = -Ry()*sc;
		bb.high().x =  Rx()*sc;
		bb.high().y =  Ry()*sc;
	}
#endif
} // _updateOBB
#endif
