/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "engine.h"

#include <iomanip>

#include "kernel.h"
#include "math/point.h"
#include "graph/viewport.h"
#include "geometry/bvh.h"

using std::cout;
using std::endl;

// During debugging to assert if bbox.inside() do not match with body.inside()
//#define ASSERTFALSE	assert(false)
#define ASSERTFALSE

/* ========================== GeometryEngine ========================== */
/** GeometryEngine */
GeometryEngine::GeometryEngine(const Geometry* geo, GeometryKernel* k) :
	_geometry(geo),
	_kernel(k),
	irurAccel(32),
	bvhCache(32),
	_boundingVolTree(nullptr)
{
	gBodyCheckId    = 0;
	gBodyMaxCheckId = 0;
	nClipBodies     = 0;
	// Accelerators
	irurAccel.compare(CBodyOrderAccel::compare);
	lastIRURCheckID = -1;
} // GeometryEngine

/** ~GeometryEngine */
GeometryEngine::~GeometryEngine()
{
	bodies.clear();
	zonesSorted.clear();
	bvhCache.clear();
} /* ~GeometryEngine */

/** view */
ViewPort& GeometryEngine::view() const
{
	return _kernel->view;
} // view

/*
 * Tighten BBoxes as much as possible in each dimension into maxBox.
 */
void GeometryEngine::tightenAllZonesBBoxes(const Vector& minPos, const Vector& maxPos,
					   const BBox& maxBox)
{
	// Loop on all zones.
	for (const auto& zone : zonesSorted) {
		auto curLow = zone->bbox().low();
		auto curHigh = zone->bbox().high();

		// Is the zone's BBox outside of the maxDim extrema?
		if (curLow.x  <= minPos.x || curLow.y  <= minPos.y || curLow.z  <= minPos.z  ||
		    curHigh.x >= maxPos.x || curHigh.y >= maxPos.y || curHigh.z >= maxPos.z) {
			// For each direction, tighten the zone's BBox when relevant.
			if (zone->bbox().lowx()  <= -maxBox.lowx()) zone->bboxLowX( maxBox.lowx());
			if (zone->bbox().lowy()  <= -maxBox.lowy()) zone->bboxLowY( maxBox.lowy());
			if (zone->bbox().lowz()  <= -maxBox.lowz()) zone->bboxLowZ( maxBox.lowz());
			if (zone->bbox().highx() >= maxBox.highx()) zone->bboxHighX(maxBox.highx());
			if (zone->bbox().highy() >= maxBox.highy()) zone->bboxHighY(maxBox.highy());
			if (zone->bbox().highz() >= maxBox.highz()) zone->bboxHighZ(maxBox.highz());
			cout << "-- BBox of " << zone->region()->name() << " tightened" << std::endl;
		}
	}
} // tightenAllZonesBBoxes

/** getRegion */
VRegion* GeometryEngine::getRegion(const int id)
{
	return _kernel->getRegion(id);
} // getRegion

/** cleanup
 * Do not delete the data, simply destroy the links V[Body/Region] -> G[Body/Region]
 */
void GeometryEngine::cleanup()
{
	for (auto body : bodies)
		body.init(NULL);
} // cleanup

// Convert the Bodies,Zones,Regions to Array(VBody,VZone,VRegion)
// do not delete anything but reassign
//
/** derive - derive the VBodies, VZones and VRegions from Geometry
 * It initializes the list of bodies, zones and regions in the Viewer,
 * note that it deletes all previous objects and creates new ones
 */
void GeometryEngine::derive()
{
	// WARNING: bodies and regions vectors can contain more or equal
	// number of VBodies vs _geometry->bodies/regions
	// Synchronize the bodies
	bodies.clear();
	bodies.resize(_geometry->nBodies());
	for (size_t i=0; i<_geometry->nBodies(); i++) {
		GBody* gbody = _geometry->bodies[i];
		VBody* vbody = _kernel->bodies[i];
		assert(gbody == vbody->body());
		bodies[i].init(gbody, vbody);
		bodies[i].checkId(&gBodyCheckId);
	}
	//cache.resize(_geometry->nBodies());

	// Rebuild zones caching array
	zonesSorted.clear();
	bvhCache.clear();
	int maxbodies = 0;
	for (size_t i=0; i<_geometry->nRegions(); i++) {
		VRegion* region = _kernel->getRegion(i);
		for (auto zone : region->zones()) {
			zonesSorted.push_back(zone);
			int bodycount = 0;
			// Fill Body/Zone references
			for (int k=0; k<zone->size(); k++) {
				const GBody* gbody = zone->gexpr(k);
				if (gbody->isBody()) {
					getBody(gbody)->addZone(zone);
					bodycount++;
				}
			}
			maxbodies = Max(maxbodies, bodycount);
		}
	}
	// FIXME assuming we are only dealing with quadratic surfaces
	zoneAccel.reserve(maxbodies*2);
} // derive

/** find zone where point is
 * WARNING this routine is optimized for points close to the 2D viewport ONLY!
 * @param x,y,z   location to search
 * @param u,v,w   direction in case we are on the boundary
 * @param zone    candidate zone to be checked first
 * @return first zone that belongs to x,y,z with direction (u,v,w)
 */
VZone* GeometryEngine::where2D( const double x, const double y, const double z,
				const double u, const double v, const double w,
				VZone* zone)
{
	if (zone)
		if (zone->inside2D(this, x,y,z, u,v,w)) return zone;

	for (auto region : _kernel->regions) {
		if (region->universe() != _kernel->universe) continue;
		VZone* vzone = region->inside2D(this, x,y,z, u,v,w);
		if (vzone) return vzone;
	}
	return NULL;
} // where2D

/** find zone where point is
 * WARNING this routine is optimized for points close to the 2D viewport ONLY!
 * @param p		location structure to search
 * @return first zone that belongs to x,y,z with direction (u,v,w)
 */
int GeometryEngine::where2D(ZoneOfPoint* p)
{
	// FIXME use the "valid/invalid" information
	// if body is not modified but from the previous check
	// then do not check it's zones...
	// WARNING: e.g the second body can be modified, by the check below
	//              then is important to keep it!!!
	p->init();
	const double tol = 1e-8*(Abs(p->x)+Abs(p->y)+Abs(p->z)+1);
	if (p->nOld != 0) {
		// First check zones of all bodies crossed
		for (int i=p->from; i<=p->to; i++) {
			VBody* body = (*p->V)[i].body;
			if (body==NULL) continue;
			CBody* cbody = getBody(body);
			for (int j=0; j<(int)cbody->zones.size() && p->more(); j++) {
				IFSTAT(stat("where_body"));
				VZone* zone = cbody->zones[j];
				if (zone->universe() != _kernel->universe) continue;
				if (not zone->bbox().inside(p->x,p->y,p->z,tol)) {
#if _DEBUG>0
					if (zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
						cout << "*-* where2D.Zone1=" << *(zone->zone()) << endl;
						cout << "+++ pos = " << std::setprecision(22)
								<< p->x << ", " << p->y << ", " << p->z << endl;
						cout << "+++ dir = " << p->dx << ", " << p->dy << ", " << p->dz << endl;
						cout << "+++ tol = " << tol << endl;
						cout << "+++ bbox= " << zone->bbox() << endl;
						cout << ">>> bbox inside=" << zone->bbox().inside(p->x,p->y,p->z,tol) << endl;
						cout << ">>> zone inside=" << zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz) << endl;
						ASSERTFALSE;
					}
#endif
					continue;
				}
				if (zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
					IFSTAT(stat("where_body_found"));
					p->add(zone);
				}
			}
		}

		// loop over current body zones
		for (int i=0; i<(int)getBody(p->body)->zones.size() && p->more(); i++) {
			IFSTAT(stat("where_body"));
			VZone* zone = getBody(p->body)->zones[i];
			if (zone->universe() != _kernel->universe) continue;
			if (not zone->bbox().inside(p->x,p->y,p->z,tol)) {
#if _DEBUG>0
				if (zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
					cout << "*-* where2D.Zone2=" << *(zone->zone()) << endl;
					cout << "+++ pos = " << std::setprecision(22)
							<< p->x << ", " << p->y << ", " << p->z << endl;
					cout << "+++ dir = " << p->dx << ", " << p->dy << ", " << p->dz << endl;
					cout << "+++ tol = " << tol << endl;
					cout << "+++ bbox= " << zone->bbox() << endl;
					cout << ">>> bbox inside=" << zone->bbox().inside(p->x,p->y,p->z,tol) << endl;
					cout << ">>> zone inside=" << zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz) << endl;
						ASSERTFALSE;
				}
#endif
				continue;
			}
			if (zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
				IFSTAT(stat("where_body_found"));
				p->add(zone);
			}
		}

		// Then all zones in zoneOld
		for (int i=0; i<p->nOld && p->more(); i++) {
			IFSTAT(stat("where_zoneold"));
			if (not p->zoneOld[i]->bbox().inside(p->x,p->y,p->z,tol)) {
#if _DEBUG>0
				if (p->zoneOld[i]->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
					cout << "*-* where2D.Zone3=" << *(p->zoneOld[i]->zone()) << endl;
					cout << "+++ pos = " << std::setprecision(22)
							<< p->x << ", " << p->y << ", " << p->z << endl;
					cout << "+++ dir = " << p->dx << ", " << p->dy << ", " << p->dz << endl;
					cout << "+++ tol = " << tol << endl;
					cout << "+++ bbox= " << p->zoneOld[i]->bbox() << endl;
					cout << ">>> bbox inside=" << p->zoneOld[i]->bbox().inside(p->x,p->y,p->z,tol) << endl;
					cout << ">>> zone inside=" << p->zoneOld[i]->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz) << endl;
					ASSERTFALSE;
				}
#endif
				continue;
			}
			if (p->zoneOld[i]->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
				IFSTAT(stat("where_zoneold_found"));
				p->add(p->zoneOld[i]);
			}
		}
	}

	// Nothing found check all the remaining zones
	if (p->n==0) {
		IFSTAT(stat("where_zone_none"));
		// We should check for touching Conics and
		// scan all regions sharing a conic...
		// otherwise it wont find possible errors
		// with touching surfaces
		// Check everything...
		//
		// Check all zones sorted
		for (int i=(int)zonesSorted.size(); i-- && p->more();) {
			VZone* zone = zonesSorted[i];
			if (zone->universe() != _kernel->universe) continue;
			if (not zone->bbox().inside(p->x,p->y,p->z,tol)) {
#if _DEBUG>0
				if (zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
					cout << "*-* where2D.Zone4=" << *(zone->zone()) << endl;
					cout << "+++ pos = " << std::setprecision(22)
							<< p->x << ", " << p->y << ", " << p->z << endl;
					cout << "+++ dir = " << p->dx << ", " << p->dy << ", " << p->dz << endl;
					cout << "+++ tol = " << tol << endl;
					cout << "+++ bbox= " << zone->bbox() << endl;
					cout << ">>> bbox inside=" << zone->bbox().inside(p->x,p->y,p->z,tol) << endl;
					cout << ">>> zone inside=" << zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz) << endl;
					ASSERTFALSE;
				}
#endif
				continue;
			}
			if (zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
				// promote zone
				if (i<(int)zonesSorted.size()-1) {
					zonesSorted.erase(zonesSorted.begin()+i);
					zonesSorted.push_back(zone);
				}
				p->add(zone);
			}
		}
	}

	return p->n;
} // where2D

#if !defined(USE_BBOX_HIERARCHIES)
/** find zone where point is, in the transformed system
 * @param x,y,z    location to search
 * @param u,v,w    direction in case we are on the boundary
 * @param t        distance to check
 * @param zone     current zone to check
 * @param body     check zones linked to body
 * @return first region that belongs to (x+u*t,y+v*t,z+w*t)
 */
VZone* GeometryEngine::whereRay(const double x, const double y, const double z,
				const double u, const double v, const double w,
				const double t, const int universe,
				VZone* zone, CBody* body)
{
	IFSTAT(stat("whereRay.01 call"));

	auto px = x + u*t;
	auto py = y + v*t;
	auto pz = z + w*t;
	const double tol = SMALL4 * (Abs(px) + Abs(py) + Abs(pz) + 1);

	// Check current zone if specified
	if (zone) {
		assert(zone->universe() == universe);
		IFSTAT(stat("whereRay.02 zoneCheck"));
		if (zone->fillAll() or (zone->bbox().inside(px, py, pz, tol) and
				zone->inside(this, x, y, z, u, v, w, t))) {
			IFSTAT(stat("whereRay.03_zoneHit"));
			return zone;
		}
	}

	// Check zones linked to body if any
	if (body) {
		IFSTAT(stat("whereRay.04_bodyCheck"));
		for (int i=body->zones.size(); i--; ) {
			VZone* zzz = body->zones[i];
			if (zzz->universe() != universe) continue;
			IFSTAT(stat("whereRay.05_body-zoneCheck"));
			if (not zzz->fillAll() and not zzz->bbox().inside(px, py, pz, tol)) {
#if _DEBUG>0
				if (zzz->inside(this, x,y,z, u,v,w, t)) {
					cout << "*-* whereRay.Zone1=" << *(zzz->zone()) << endl;
					cout << "+++ pos = " << x << ", " << y << ", " << z << endl;
					cout << "+++ dir = " << u << ", " << v << ", " << w << endl;
					cout << "+++ t   = " << t << endl;
					cout << "+++ end = " << px << ", " << py << ", " << pz << endl;
					cout << "+++ tol = " << tol << endl;
					cout << "+++ bbox= " << zzz->bbox() << endl;
					cout << ">>> bbox inside=" << zzz->bbox().inside(px,py,pz,tol) << endl;
					cout << ">>> zone inside=" << zzz->inside(this, x,y,z, u,v,w, t) << endl;
					ASSERTFALSE;
				}
#endif
				IFSTAT(stat("whereRay.06_body-zone-bbox"));
				continue;
			}
			if (zzz->fillAll() or zzz->inside(this, x,y,z, u,v,w, t)) {
				IFSTAT(stat("whereRay.07_body-Hit"));
				IFSTAT(stat.inc("whereRay.08_body-zoneIndex_sum",
						(body->zones.size() - i)));
				if (i<(int)body->zones.size()-1) {
					IFSTAT(stat("whereRay.09_body-zone-promote"));
					// promote zone
					//body->zones.erase(i);
					body->zones.erase(body->zones.begin()+i);
					body->zones.push_back(zzz);
				}
				return zzz;
			}
		}
	}

	IFSTAT(stat("whereRay.10_check-all"));
	for (int i=(int)zonesSorted.size(); i--;) {
		VZone* zzz = zonesSorted[i];
		if (zzz->universe() != universe) continue;
		IFSTAT(stat("whereRay.11_check-all_zone"));
		if (not zzz->fillAll() and not zzz->bbox().inside(px,py,pz,tol)) {
#if _DEBUG>0
			if (zzz->inside(this, x,y,z, u,v,w, t)) {
				cout << "*-* whereRay.Zone2=" << *(zzz->zone()) << endl;
				cout << "+++ pos = " << x << ", " << y << ", " << z << endl;
				cout << "+++ dir = " << u << ", " << v << ", " << w << endl;
				cout << "+++ t   = " << t << endl;
				cout << "+++ end = " << px << ", " << py << ", " << pz << endl;
				cout << "+++ tol = " << tol << endl;
				cout << "+++ bbox= " << zzz->bbox() << endl;
				cout << ">>> bbox inside=" << zzz->bbox().inside(px,py,pz,tol) << endl;
				cout << ">>> zone inside=" << zzz->inside(this, x,y,z, u,v,w, t) << endl;
				ASSERTFALSE;
			}
#endif
			IFSTAT(stat("whereRay.12_check-all_bbox"));
			continue;
		}
		if (zzz->fillAll() or zzz->inside(this, x,y,z, u,v,w, t)) {
			IFSTAT(stat("whereRay.13_check-all_hit"));
			//int idx = iter.index()+1;
			//IFSTAT(stat.inc("whereRay_zoneIndex_sum",
			//		(zones.size() - idx)));
			if (i<(int)zonesSorted.size()-1) {
				// promote zone
				zonesSorted.erase(zonesSorted.begin()+i);
				zonesSorted.push_back(zzz);
				if (body) body->zones.push_back(zzz);
			}
			return zzz;
		}
		// No need to check the zones
	}
	IFSTAT(stat("whereRay.14_error"));
	return NULL;
} // whereRay

#else  // BVH implementation

/** find zone where point is in the transformed system
 * @param x,y,z    location to search
 * @param u,v,w    direction in case we are on the boundary
 * @param t        distance to check
 * @param zone     current zone to check
 * @param body     check zones linked to body
 * @return first region that belongs to (x+u*t,y+v*t,z+w*t)
 */
VZone* GeometryEngine::whereRay( const double x, const double y, const double z,
				 const double u, const double v, const double w,
				 const double t, const int universe,
				 VZone* zone, CBody* body)
{
	IFSTAT(stat("whereRay.01 call"));

	auto px = x + u * t;
	auto py = y + v * t;
	auto pz = z + w * t;
	const double tol = SMALL4 * (Abs(px) + Abs(py) + Abs(pz) + 1);
	VZone* pZone = nullptr;

#if defined (USE_ONLY_BVH)
	pZone = _boundingVolTree->inside(this, Vector(x, y, z), Vector(u, v, w), t, tol);
	return pZone;
#endif
	// Check current zone if specified
	if (zone) {
		assert(zone->universe() == universe);
		IFSTAT(stat("whereRay.02 zoneCheck"));
		if (zone->fillAll() or (zone->bbox().inside(px, py, pz, tol) and
				 zone->inside(this, x, y, z, u, v, w, t))) {
			IFSTAT(stat("whereRay.03_zoneHit"));
			return zone;
		}
	}

	// Check zones linked to body if any
	if (body) {
		IFSTAT(stat("whereRay.04_bodyCheck"));
		for (int i = body->zones.size(); i--;) {
			VZone* zzz = body->zones[i];
			if (zzz->universe() != universe) continue;
			IFSTAT(stat("whereRay.05_body-zoneCheck"));
			if (not zzz->fillAll() and not zzz->bbox().inside(px, py, pz, tol)) {
#if _DEBUG>0
				if (zzz->inside(this, x, y, z, u, v, w, t)) {
					cout << "*-* whereRay.Zone1=" << *(zzz->zone()) << endl;
					cout << "+++ pos = " << x << ", " << y << ", " << z << endl;
					cout << "+++ dir = " << u << ", " << v << ", " << w << endl;
					cout << "+++ t   = " << t << endl;
					cout << "+++ end = " << px << ", " << py << ", " << pz << endl;
					cout << "+++ tol = " << tol << endl;
					cout << "+++ bbox= " << zzz->bbox() << endl;
					cout << ">>> bbox inside=" << zzz->bbox().inside(px, py, pz, tol) << endl;
					cout << ">>> zone inside=" << zzz->inside(this, x, y, z, u, v, w, t) << endl;
					ASSERTFALSE;
				}
#endif
				IFSTAT(stat("whereRay.06_body-zone-bbox"));
				continue;
			}
			if (zzz->fillAll() or zzz->inside(this, x, y, z, u, v, w, t)) {
				IFSTAT(stat("whereRay.07_body-Hit"));
				IFSTAT(stat.inc("whereRay.08_body-zoneIndex_sum",
								(body->zones.size() - i)));
				if (i < (int)body->zones.size() - 1) {
					IFSTAT(stat("whereRay.09_body-zone-promote"));
					// promote zone
					//body->zones.erase(i);
					body->zones.erase(body->zones.begin() + i);
					body->zones.push_back(zzz);
				}
				return zzz;
			}
		}
	}
/*
	const int cMaxBVHCacheSize = 256*40;
	// check L1 BVH cache
	for (int i = bvhCache.size(); i--;) {
		VZone* zzz = bvhCache[i];
		if (zzz->universe() != universe) continue;
		IFSTAT(stat("whereRay.11_check-all_zone"));
		if (not zzz->fillAll() and not zzz->bbox().inside(px, py, pz, tol)) {
#if _DEBUG>0
			if (zzz->inside(this, x, y, z, u, v, w, t)) {
				cout << "*-* whereRay.Zone2=" << *(zzz->zone()) << endl;
				cout << "+++ pos = " << x << ", " << y << ", " << z << endl;
				cout << "+++ dir = " << u << ", " << v << ", " << w << endl;
				cout << "+++ t   = " << t << endl;
				cout << "+++ end = " << px << ", " << py << ", " << pz << endl;
				cout << "+++ tol = " << tol << endl;
				cout << "+++ bbox= " << zzz->bbox() << endl;
				cout << ">>> bbox inside=" << zzz->bbox().inside(px, py, pz, tol) << endl;
				cout << ">>> zone inside=" << zzz->inside(this, x, y, z, u, v, w, t) << endl;
				ASSERTFALSE;
			}
#endif
			IFSTAT(stat("whereRay.12_check-all_bbox"));
			continue;
		}
		if (zzz->fillAll() or zzz->inside(this, x, y, z, u, v, w, t)) {
			IFSTAT(stat("whereRay.13_check-all_hit"));
			if (i < (int)bvhCache.size() - 1) {
				// promote zone
				bvhCache.erase( bvhCache.begin() + i);
				bvhCache.push_back(zzz);
				if (body) body->zones.push_back(zzz);
			}

			cout << "**** Found in BVH cache: " << zzz->name() << endl;

			return zzz;
		}
		// No need to check the zones
	}
*/
	// Use LBVH acceleration to determine position
	pZone = _boundingVolTree->inside(this, Vector(x, y, z), Vector(u, v, w), t, tol);
	if (pZone) {
//		cout << "**** Found in BVH  " << pZone->name() << endl;

		if ( body ) {
//			cout << "**** BVH: attached zone to body" << endl;
			body->zones.push_back( pZone );
		}
//		if (bvhCache.size() >= cMaxBVHCacheSize) {
//			bvhCache.pop_back();
//		}
//		bvhCache.push_back(pZone);

	}

	return pZone;
} // whereRay
#endif

/** fillIRURegionAccel
 * Fill Accelerator structure for intersectRayUndefinedRegion
 * Initializes a list of bodies ordered by the distance of their intersections
 * with the current ray.
 * Checks if last checkID is still valid and if so skips all computations
 */
void GeometryEngine::fillIRURegionAccel(Ray* ray)
{
	if (gBodyCheckId == lastIRURCheckID) return; // Previous list still valid
	lastIRURCheckID = gBodyCheckId;

	irurAccel.clear();
	RaySegment& segment = ray->segment();
	double tmin = segment.tmin;
	double tmax = segment.tmax;

	for (auto& body : bodies) {
		if (body.zones.empty()) continue;
		if (body.distance(segment.pos, segment.dir)) {
			CBodyOrderAccel accel;
			if (InRangeOpen(tmin, body.tmin, tmax)) { // Add entry point
				accel.t = body.tmin;
				accel.body = &body;
				irurAccel.add(accel);
			}
			if (InRangeOpen(tmin, body.tmax, tmax)) { // Add exit point
				accel.t = body.tmax;
				accel.body = &body;
				irurAccel.add(accel);
			}
		}
	}
} // fillIRURegionAccel

/** intersectRayUndefinedRegion
 * assume an empty region as transparent and check for next intersection
 * @param ray    ray structure filled
 * @return false if ray arrives to end without hitting anything
 *         true  if ray hits any defined region
 *
 * NOTE: Similar logic has the method Zone::distance
 */
bool GeometryEngine::intersectRayUndefinedRegion(Ray* ray)
{
	IFSTAT(stat("intersectRayUndefined_call"));

	if (ray->clip_insight && applyClip(ray)) {
		RaySegment& seg = ray->segment();
		if (ray->ended()) return false;
		if (ray->n==0) ray->prevZone(seg.zone);	// update previous/master zone
		if (seg.zone) return true;
	}

	RaySegment& segment = ray->segment();
	double tmin = segment.tmin;

	fillIRURegionAccel(ray);
	ArrayIterator<CBodyOrderAccel> iter(irurAccel);
#if _DEBUG>1
//	double tmin = segment.tmin;
#endif
	while (iter) {
		CBodyOrderAccel& accel = iter++;
		if (accel.t < segment.tmin) continue;
		segment.tmin = accel.t;
		segment.body = accel.body;
		for (int iz=segment.body->zones.size()-1; iz>=0; iz--) {
			VZone* zone = segment.body->zones[iz];
			if (zone->universe() != segment.universe) continue;
			double t = segment.tmin*(1.0+SMALL3D1);
			IFSTAT(stat("whereRay_zoneIndex_sum"));
			if (zone->inside(this, segment.pos, segment.dir, t)) {
				IFSTAT(stat("whereRay_errorHit"));
				// Lattices are trickier than this
				//if (zone->inside(ray->segment(0).pos, ray->segment(0).dir, ray->getT())) {
				segment.tmin = t;
				segment.zone = zone;
				//DUMP(std::cerr << "Saved ray moving by " << segment.tmin - tmin << std::endl);
				if (!Eq0(t-tmin, SMALL3D1*(ray->pos().max()+1)))
					ray->error = true;
				return true;
			}
		}
	}
#if _DEBUG>1
//	std::cerr << "Cannot save ray" << std::endl;
#endif
	ray->error = true;
	return false;
} // intersectRayUndefinedRegion

/** check4clip
 * First find if there is a clipping body in sight...
 */
void GeometryEngine::check4clip(Ray* ray)
{
	ray->clip_insight     = false;

	const RaySegment& segment = ray->segment(0);
	for (size_t i=0; i<nClipBodies; i++) {
		int bid = clipBodiesId[i];
		assert(bid>=0 and bid<(int)bodies.size());
		CBody* cbody = getBody(bid);

		bool intersect = cbody->distance(segment.pos, segment.dir);
		if (intersect || clipNegative[i]) {
			bool tinv;
			if (clipNegative[i])
				tinv = !cbody->tinverse;
			else
				tinv = cbody->tinverse;

			if (!tinv) {
				double a = Max(cbody->tmin, segment.tmin);
				double b = Min(cbody->tmax, segment.tmax);
				ray->clip_insight |= a<b;
			} else {
				ray->clip_insight |= segment.tmin < Min(cbody->tmin, segment.tmax);
				ray->clip_insight |= Max(cbody->tmax, segment.tmin) < segment.tmax;
			}
		}
	}
} // check4clip

/** @return true if the clip body was applied */
inline bool GeometryEngine::applyClip(Ray* ray)
{
	assert(ray->clip_insight);
	RaySegment& segment = ray->segment();
	double tpos = ray->T();
	ray->clip_hit = false;
	//if (ray->clip_hit)
	//	ray->clip_hit = segment.tmin > SMALL3D;

	// loop until we are out of all clips
	bool more;
	do {
		more = false;
		for (size_t i=0; i<nClipBodies; i++) {
			int bid = clipBodiesId[i];
			assert(bid>=0 and bid<(int)bodies.size());
			CBody* cbody = getBody(bid);
			bool tnormal = clipNegative[i]? cbody->tinverse : !cbody->tinverse;
			if (tnormal) {	// normal clip
				if (InRange(cbody->tmin, tpos, cbody->tmax)) {
					// Hit inside the clip body, move to clip end and continue
					segment.tmin = (cbody->tmax - ray->Tsum())*(1.0+SMALL3D1);
					if (segment.tmax <= segment.tmin) {
						segment.tmin = segment.tmax;
						ray->clip_hit = true;
						break;
					}
					ray->clip_hit = more = true;
					tpos = ray->T();
					if (whereRay(segment))
						segment.body = cbody;
				}
			} else	// clip with concave or outside of body
			if (tpos < cbody->tmin) {
				// Hit inside cbody (first segment),
				// move to boundary and continue
				segment.tmin = (cbody->tmin - ray->Tsum())*(1.0+SMALL3D1);
				if (segment.tmax <= segment.tmin) {
					segment.tmin = segment.tmax;
					ray->clip_hit = true;
					break;
				}
				ray->clip_hit = more = true;
				tpos = ray->T();
				if (whereRay(segment))
					segment.body = cbody;
			} else	// clip with concave or outside of body
			if (tpos >= cbody->tmax) {
				// Hit inside cbody (second segment)
				// therefore it will never leave
				segment.tmin = segment.tmax;
				ray->clip_hit = true;
				break;
			}
		}
	} while (more && !segment.ended());
	check4clip(ray);	// check again if we need to use clip
	return ray->clip_hit;
} // applyClip

/** check4project
 * First find if there is a projecting body in sight...
 */
void GeometryEngine::check4project(Ray* ray)
{
	ray->project_insight = false;
	const RaySegment& segment = ray->segment(0);
	for (auto pid : projectBodiesId) {
		assert(pid>=0 and pid<(int)bodies.size());
		CBody* pbody = getBody(pid);
		if (pbody->distance(segment.pos, segment.dir)) {
			ray->project_insight |=
					InRange(segment.tmin, pbody->tmin, segment.tmax) ||
					InRange(segment.tmin, pbody->tmax, segment.tmax);
		}
	}
} // check4project

/** applyProject */
bool GeometryEngine::applyProject(Ray* ray, const double prevT)
{
	assert(ray->project_insight);
	RaySegment& segment = ray->segment();
	double T = ray->T();
	ray->project_hit = false;
	for (auto pid : projectBodiesId) {
		assert(pid>=0 and pid<(int)bodies.size());
		CBody* pbody = getBody(pid);
		if (InRange(prevT, pbody->tmin, T)) {
			ray->project_hit = true;
			T = pbody->tmin;
			segment.tmin = pbody->tmin - ray->Tsum();
			segment.body = pbody;
		}
		if (InRange(prevT, pbody->tmax, T)) {
			ray->project_hit = true;
			T = pbody->tmax;
			segment.tmin = pbody->tmax - ray->Tsum();
			segment.body = pbody;
		}
	}
	check4project(ray);	// check again if we need to use project
	return ray->project_hit;
} // applyProject

/** intersectVoxel
 * @return   1  hit found or ended
 *          -1  skip this zone
 *          -2  exit from voxel
 */
inline int GeometryEngine::intersectVoxel(Ray* ray)
{
	if (_kernel->voxel.invalid()) return -2;

	// Normal tracking, find next intersection
	RaySegment& segment = ray->segment();
	VZone* zone = segment.zone;

	double tentry = segment.tmin;
VOX:
	double tmin = segment.tmin;
	double T = ray->T();
	segment.tmin -= segment.tmin * SMALL3D;	// move backwards a bit

	if (_kernel->voxel.distance(
			segment.pos,
			segment.dir,
			&segment.tmin)) {

		// if previous hit was clip
		if (ray->clip_hit && segment.tmin - tmin < SMALL3D4) return 1;

		// Apply clip
		if (ray->clip_insight && applyClip(ray)) {
			if (ray->project_insight && applyProject(ray, T)) return 1;
			if (ray->ended()) return 1;
			if (segment.zone != zone) return -1;
			// continue to check if we hit on a filled voxel
			goto VOX;
		}

		if (ray->project_insight && applyProject(ray, T)) return 1;
		if (segment.tmin - tmin > SMALL3D4) segment.body = NULL;
		// either we hit the clipbody or the voxel, update master zone
		if (ray->n==0) ray->prevZone(segment.zone);

		// we are outside the region
		ray->moveby(SMALL3D);	// move a bit to ensure we are in the new voxel
		return 1;
	}

	// restore tmin and exit voxel zone
	segment.tmin = tentry;
	return -2;
} // intersectVoxel

/** intersect a lattice (fluka) like structure or another universe
 * @return true  if lattice was applied
 *         false if an error occurred
 */
inline bool GeometryEngine::intersectLattice(Ray* ray)
{
	RaySegment& segment = ray->segment();
	VZone*      zone    = segment.zone;
	VRegion*    region  = zone->region();
	RaySegment  new_segment(region->region(), zone->fill());

	// Find exit distance location from lattice
	double texit = segment.tmin;

	do {
		// find exit from zone
		CBody* body = zone->distance(this,
					 segment.pos,
					 segment.dir,
					&texit,
					 segment.tmax);

		// end of the ray?
		if (body==NULL) break;

		// check if we exited the region
		zone = region->inside(this,
					segment.pos,
					segment.dir,
					texit);
	} while (zone!=NULL);

	// save previous segment number
	segment.bodyCheckId = bodyCheckId();
	if (region->region()->hasMatrix()) {
		// Transform ray at the location that enters the lattice
		// to the prototype and come back to screen coordinates
		// Transformation: M⁻¹ * R * M
		// M = screen transformation
		// R = rotdefi transformation
		// find new limits in the lattice
		// from 0 to Min(texit,tmax) - tmin
		// with a small shift
		new_segment.tmin = SMALL3D2;
		new_segment.tmax = Min(texit, segment.tmax) - segment.tmin;
		new_segment.pos  = new_segment.matrix() * segment.hit();
		new_segment.dir  = new_segment.matrix() * segment.dir;
		new_segment.dir.renormalize();
		new_segment.acc  = 8.0*SMALL3D * (segment.pos.max()
						+ Abs(new_segment.matrix()(0,3))
						+ Abs(new_segment.matrix()(1,3))
						+ Abs(new_segment.matrix()(2,3)));
		incBodyCheckId();
	} else {
		// keep the same coordinate system, preserve the checkId
		// to save calculations
		new_segment.incremental = false;
		new_segment.tmin = segment.tmin;
		new_segment.tmax = texit;
		new_segment.pos  = segment.pos;
		new_segment.dir  = segment.dir;
		new_segment.acc  = segment.acc;
	}

	// Find the new zone in the prototype
	if (whereRay(new_segment) == nullptr) {
		// error in geometry, move to texit
		segment.zone = NULL;
		return false;
	}

	// start with the original hit body
	//if (ray->clip_hit)
	//	new_segment.body = ray->segment().body;

	if (ray->push(new_segment)) { // segment is no longer the last segment
		_kernel->error("Nesting level exceeded maximum.Possibly LATTICE infinite loop");
		bodyCheckId(segment.bodyCheckId);
		segment.zone = NULL;
		return false;
	}
	return true;
} // intersectLattice

/** intersect a repeated lattice structure
 * @return true  if lattice was applied
 *         false if an error occurred
 */
inline bool GeometryEngine::intersectRepeatedLattice(Ray* ray)
{
	RaySegment& segment = ray->segment();
	VZone* zone   = segment.zone;
	auto   region = zone->gregion();

	assert(region->type() == GRegion::Type::RECTLATTICE);
	assert(region->dlat.x != 0);

	// Do we need to assume infinite periods?
	// get replica indices
	Point  p   = segment.hit();
	Vector dp  = p - region->plat;
	double iil = region->dlat.x!=0.0? (dp * region->ilat) / region->dlat.x : 0;
	double jjl = region->dlat.y!=0.0? (dp * region->jlat) / region->dlat.y : 0;
	double kkl = region->dlat.z!=0.0? (dp * region->klat) / region->dlat.z : 0;

	// transform to user supplied indices
	int il = Int(iil);
	int jl = Int(jjl);
	int kl = Int(kkl);

	// correct boundary problems
	if (iil < (double)il + SMALL3D and segment.dir * region->ilat < 0)
		il--;
	else
	if (iil > (double)(il+1) - SMALL3D and segment.dir * region->ilat > 0)
		il++;

	// FIXME could it be better to check
	if (region->dlat.y != 0.0) {
		if (jjl < (double)jl + SMALL3D and segment.dir * region->jlat < 0)
			jl--;
		else
		if (jjl > (double)(jl+1) - SMALL3D and segment.dir * region->jlat > 0)
			jl++;

		if (region->dlat.z != 0.0) {
			if (kkl < (double)kl + SMALL3D and segment.dir * region->klat < 0)
				kl--;
			else
			if (kkl > (double)(kl+1) - SMALL3D and segment.dir * region->klat > 0)
				kl++;
		}
	}

	// point entering the lattice cell
	Point plat = p - (il*region->dlat.x * region->ilat
			+ jl*region->dlat.y * region->jlat
			+ kl*region->dlat.z * region->klat);

	// find universe
	int universe;
	if (region->fill()==0) {
		// start indices from 0
		il -= region->ilow;
		jl -= region->jlow;
		kl -= region->klow;
		//if (region->latticei()) il = 0;
		if (not region->latticej()) jl = 0;
		if (not region->latticek()) kl = 0;
		// be more flexible on the boundaries to ensure we are in the correct cell
		//assert(InRange(region->ilow, il, region->ihigh));
		//assert(InRange(region->jlow, jl, region->jhigh));
		//assert(InRange(region->klow, kl, region->khigh));
		int idx = (kl * region->latticejSize() + jl) * region->latticeiSize() + il;
		if (idx<0 or idx>=(int)region->latticeSize()) return false;
		assert(idx>=0 and idx<(int)region->latticeSize());
		universe = region->lattice(idx);
		const Matrix4* matrix = region->latticeMatrix(idx);
		if (matrix != nullptr)
			plat *= *matrix;
	} else
		universe = region->fill();

	// loop to find exiting distance. Normally one step will be sufficient
	double tmin = SMALL3D;
	double tmax = segment.tmax - segment.tmin;

	// due to precision we might be slightly outside
	segment.bodyCheckId = incBodyCheckId();

//#if _DEBUG>0
	for (int i=0; i<10; i++) {
		if (zone->inside(this, plat, segment.dir, tmin)) break;
		tmin *= 2.0;
		//assert(false);
	}
//#endif
	if (!zone->inside(this, plat, segment.dir, tmin))
		return false;
	double texit = tmin;
	zone->distance(this, plat, segment.dir, &texit, tmax);

	RaySegment new_segment(region, universe);

	// FIXME it doesn't have any matrix
	new_segment.pos  = plat;
	new_segment.dir  = segment.dir;
	new_segment.tmin = tmin;
	new_segment.tmax = texit;
	new_segment.acc  = 8.0*SMALL3D * segment.pos.max();
	new_segment.zone = zone;

	auto gbody = zone->zone()->closest(plat, new_segment.tmin*16);
	new_segment.body = gbody!=nullptr? getBody(gbody) : nullptr;

	// Find the new zone in the prototype
	if (whereRay(new_segment) == nullptr) {
		// error in geometry, move to texit
		segment.zone = NULL;
		return false;
	}

	if (ray->push(new_segment)) { // segment is no longer the last segment
		_kernel->error("Nesting level exceeded maximum.Possibly LATTICE infinite loop");
		bodyCheckId(segment.bodyCheckId);
		segment.zone = NULL;
		return false;
	}
	return true;
} // intersectRepeatedLattice

/** intersectRay
 * find the next intersection of the ray with any zone and return.
 * It takes care of multiple-lattices as well as voxels
 * @param ray    structure contains the ray/depth
 * @param step   if true, the function returns after every step otherwise only
 *               when hitting a non-transparent object
 * @return true  if we arrived to the end of the ray or error (and
 *               ray->segment().zone will be NULL)
 *         false the ray hit a solid object or next step was computed and there
 *               is still way to go (if step=true)
 */
bool GeometryEngine::intersectRay(Ray* ray, bool step)
{
	IFSTAT(stat("intersectRay_call"));

	if (!step) {
		if (ray->use_clip && nClipBodies) {
			ray->clip_hit = false;
			check4clip(ray);
		}
		if (ray->use_project && projectBodyCount()) {
			ray->project_hit = false;
			check4project(ray);
		}
	}

	// Loop until we exit the region (zone->region())
	while (true) {
		// Check for ending condition
		// XXX it doesn't work for negative rays!!!
		RaySegment& segment = ray->segment();

		// Segment has ended
		if (segment.ended()) {
			// Initial segment and ended, we arrived to the end
			do {
				if (ray->n == 0) return true;
				// keep topmost segment matrix and tmax before pop()
				bool inc  = ray->segment().incremental;
				auto tmax = ray->segment().tmax;
				ray->pop();
				// propagate position to the new topmost segment
				if (inc)
					ray->segment().tmin += tmax;
				else
					ray->segment().tmin  = tmax;
			} while (ray->segment().ended());
			bodyCheckId(ray->segment().bodyCheckId);
			// find zone in the new topmost segment
			if (whereRay(ray->segment(), ray->segment().zone)==nullptr)
				return true;
			if (step) return false;
			continue;	// we cannot modify the 'segment' variable => go back to start
		}

		VZone* zone = segment.zone;

		// Check for error
		if (zone == NULL) {
			if (intersectRayUndefinedRegion(ray)) continue;
			// When hitting an undefined region check
			// if we are at the start of the ray
			if (ray->prevZone() == NULL) return false;
			// the previous one was BLACKHOLE = no problem
			if (ray->prevZone()->region()->type() == GRegion::Type::BLACKHOLE) return false;
			return true;
		}

		// Check if we have to stop the ray and return
		if (!step && ray->type != Ray::RayType::Transmission ) {
			// We've hit a blackhole?
			if (zone->region()->type() == GRegion::Type::BLACKHOLE and
			    (ray->skip_black==Ray::SkipBlack::StopOnBlack or
			    (ray->skip_black==Ray::SkipBlack::Skip1stBlack and ray->prevZone())))
				return false;

			// Inside a solid (non transparent) object
			if (!ray->skip_current && zone->region()->type() == GRegion::Type::NORMAL &&
					(!ray->skip_transparent || !zone->region()->transparent() )) {
				double T = ray->T();
				// are we inside a clipping body
				if (ray->clip_insight && applyClip(ray)) {
					if (ray->project_insight && applyProject(ray, T))
						return ray->ended();
					if (ray->ended()) return true;
					if (ray->n==0) ray->prevZone(zone);	// update master zone
					continue;
				} else
					return false;
			}
			ray->skip_current = false;
		}

		// Only increment stats when considering a new intersection or segment
		IFSTAT(stat("intersectRay_retry"));

		// ------------------- LATTICE -------------------
		// Could be that a region is marked as LATTICE but there is
		// no transformation associated with it
		if (zone->region()->type() == GRegion::Type::LATTICE) {
			if (!intersectLattice(ray)) return true;
			if (step) return false;
			// we just entered in a lattice
			auto region = ray->segment().zone->region();
			if (!ray->skip_current && region->type() == GRegion::Type::NORMAL &&
					(!ray->skip_transparent || !region->transparent()))
				return false;
			continue;
		} else
		// ------------------- VOXEL -------------------
		if (zone->region()->type() == GRegion::Type::VOXEL) {
			int ret = intersectVoxel(ray);
			if (ret ==  1) return segment.ended();	// Intersection in voxel
			if (ret == -1) continue;		// Skip this zone
			// if (ret == -2) let it continue...
		} else
		// ---------------- RECT LATTICE ---------------
		if (zone->region()->type() == GRegion::Type::RECTLATTICE) {
			if (!intersectRepeatedLattice(ray)) return true;
			if (step) return false;
			continue;
		} else
		// ---------------- HEX LATTICE ----------------
		if (zone->region()->type() == GRegion::Type::HEXLATTICE) {
			assert(false);
			continue;
		}

		// ------------------- NORMAL tracking -------------------
		if (ray->n==0) ray->prevZone(segment.zone);	// update master zone

		double T = ray->T();				// Move to next intersection

		segment.body = segment.zone->distance(this,
							 segment.pos,
							 segment.dir,
							&segment.tmin,
							 segment.tmax);
		if (ray->project_insight && applyProject(ray, T)) return ray->ended();
		if (segment.ended()) continue;
		if (segment.body==NULL) return false;

		// Check for clipping body
		if (ray->clip_insight && applyClip(ray)) {
			if (ray->project_insight && applyProject(ray, T)) return ray->ended();
			if (ray->ended()) return true;
			if (ray->n==0) ray->prevZone(segment.zone);	// update master zone
		}

		// Find the zone we are moving to
		whereRay(segment, NULL, segment.body);
		if (step || ray->type == Ray::RayType::Transmission ) return (segment.zone==NULL);
	}
} // intersectRay

/** bodiesMemory */
size_t GeometryEngine::bodiesMemory() const
{
	size_t bmem = bodies.capacity() * sizeof(CBody);
	for (auto body : bodies)
		bmem += body.memory();
	return bmem;
} // bodiesMemory

/** memory */
size_t GeometryEngine::memory() const
{
	return sizeof(GeometryEngine) + bodiesMemory()
			+ sizeof(zonesSorted) + sizeof(VZone*)*zonesSorted.capacity()
			+ irurAccel.memory();
} // memory

/** printMemory */
void GeometryEngine::printMemory() const
{
	cout << endl << "GeometryEngine:" << endl;
	cout << "Memory:"     << endl;
	cout << "\tSelf:\t"   << sizeof(GeometryEngine) << endl;
	cout << "\tBodies:\t" << bodiesMemory() << endl;
	cout << "\tZones:\t"  << sizeof(zonesSorted)
				+ sizeof(VZone*)*zonesSorted.capacity() << endl;
	cout << "\tTotal:\t"  << memory() << endl;
} // printMemory
