/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#ifndef __GEOMETRY_GREGION_H
#define __GEOMETRY_GREGION_H

#include <climits>

#include "gzone.h"
#include "material.h"

#include "math/vector.h"

class GBody;
class Geometry;
class GeometryEngine;

// ================================== GRegion ==================================
class GRegion {
	using ZoneFormat = GZone::Format;
	using ZonesArray = std::vector<GZone*>;
public:
	// Special region types
	enum class Type : int {
		NORMAL      = 0,
		BLACKHOLE   = 1,
		LATTICE     = 2,
		VOXEL       = 3,
		RECTLATTICE = 4,
		HEXLATTICE  = 5
	};
protected:
	std::string	_name;		// name of the region
	int		_id;		// index id in regions array
	Type		_type;		// type of region: normal/blackhole/lattice/voxel
	int		_generation;	// Generation number
	ZonesArray	_zones;		// zones defining the region
mutable	dword		_hash;		// Cached hash value
	Material*	_material;	// Region material
	bool		_hasMatrix;	// if rotation matrix is set
	Matrix4		_matrix;	// rotation matrix
	Matrix4		_invMatrix;	// inverse rotation matrix
	int		_universe;	// universe index of region
	int		_fill;		// filled by universe index
					// <0 for lattice filling everything
	// repeated lattice limits
	int		ilow, ihigh;	// lattice indices limits
	int		jlow, jhigh;
	int		klow, khigh;
	Point		plat;		// lattice origin
	Vector		ilat,		// vector describing lattice direction
			jlat,
			klat;
	Vector		dlat;		// lattice step in each direction
	unsigned short*	_lattice;	// lattice universes
	unsigned	_latMatrixSize;	// size of lattice matrices
	unsigned short*	_latMatrixId;	// matrix id for each lattice universe
	Matrix4*	_latMatrix;	// lattice matrices

public:
	int		show;		// show region flag

public:
	GRegion(const std::string& aname="");
virtual ~GRegion()			{ clear(); };
	void	clear(bool fast=false);

	void	name(const std::string& aname) { _name = aname; }
const	std::string&	name()	const	{ return _name; }
const	char*	cname()		const	{ return _name.c_str(); }

	void	type(const Type t)	{ _type = t; }
	Type type()		const	{ return _type; }

	void	id(const int i)		{ _id = i; }
	int	id()		const	{ return _id; }

	GRegion* clone(const std::string& aname) const;

	// Matrix and transformations
	void	clearMatrix()		{ _hasMatrix = false; }
	bool	hasMatrix()	 const	{ return _hasMatrix; }
	void	matrix(const Matrix4& M);
const	Matrix4& matrix()	const	{ return _matrix; }
const	Matrix4& invMatrix()	const	{ return _invMatrix; }
	void	transform(const Matrix4& M);

	// universe and filling
	void	universe(int u)		{ _universe = u; }
	int	universe()	const	{ return _universe; }
	void	fill(int f)		{ _fill = f; }
	int	fill()		const	{ return _fill; }
	// fill entire space
	bool	fillAll()	const	{ return In(_type, Type::RECTLATTICE, Type::HEXLATTICE); }

	// @return true lattice is defined along direction
	bool	latticei()	const	{ return ihigh != ilow; }
	bool	latticej()	const	{ return jhigh != jlow; }
	bool	latticek()	const	{ return khigh != klow; }
	// @return lattice size
	unsigned latticeiSize() const	{ return (ihigh-ilow+1); }
	unsigned latticejSize() const	{ return (jhigh-jlow+1); }
	unsigned latticekSize() const	{ return (khigh-klow+1); }
	unsigned latticeSize()	const
				{ return latticeiSize() * latticejSize() * latticekSize(); }
	bool	latticeInit(int il, int ih, int jl, int jh, int kl, int kh);
	void	lattice(unsigned pos, unsigned univ) {
				assert(_lattice!=nullptr);
				assert(pos < latticeSize());
				_lattice[pos] = (unsigned short)univ;
			}
	unsigned short	lattice(unsigned pos) const { return _lattice[pos]; }
	bool	latticeMatrixInit(const unsigned size);
	void	latticeMatrix(unsigned pos, unsigned mid, const Matrix4& matrix) {
				assert(pos <  latticeSize());
				assert(mid < _latMatrixSize);
				assert(_latMatrixId != nullptr);
				assert(_latMatrix != nullptr);

				_latMatrixId[pos] = mid;
				_latMatrix[mid]   = matrix.inv();
				//_latMatrix[mid]   = matrix;
				std::cout << "pos=" << pos << " mid=" << mid << "matrix=\n" << _latMatrix[mid] << std::endl;
			}
const	Matrix4* latticeMatrix(unsigned pos) const {
				assert(pos <  latticeSize());
				if (_latMatrixId==nullptr) return nullptr;
				auto mid = _latMatrixId[pos];
				if (mid == USHRT_MAX) return nullptr;
				return &_latMatrix[mid];
			}

	// Material
	void	material(Material* mat)	{ _material = mat; }
const	Material* material()	const	{ return _material; }

	// Zones
	int	size()		const	{ return _zones.size(); }
const	ZonesArray& zones()	const	{ return _zones; }
	GZone*	operator [] (const int i){ return _zones[i]; }
const	GZone*	operator [] (const int i)const	{ return _zones[i]; }
	GZone*	zone(const int i)	{ return _zones[i]; }
const	GZone*	zone(const int i) const	{ return _zones[i]; }

	GZone*	addZone(ZoneFormat aformat=ZoneFormat::PRODUCT);
	GZone*	addZone(GZone* z);
	GZone*	insertZone(int pos, GZone* z);
	void	delZone(int i);
	bool	add2exp(const char* token, GBody* body);
	void	convertFromNormal();
	void	setup();

	// Optimization
	bool	removeDup(int iz);
	bool	merge(int iz);
	bool	contains(const GZone& zone);
	bool	expandRegions(Geometry* geometry);
	bool	optimize();

	// Generation
	int	generation()	const	{ return _generation; }
	void	generation(int newGen) {
			_generation = newGen;
			//clearBB();
			_hash = 0;
		}

	// Hash
	dword	hash()   const;

	// bounding box
	BBox	bbox()   const;
#ifdef USE_OBBOX
	OBBox	obbox()  const;
#endif
	void	updateBBox();

	// Inside (WARNING: non-caching functions)
	GZone*	inside( const double x, const double y, const double z,
			const double u, const double v, const double w) const;
	GZone*	inside( const Point& p, const Vector& d) const
			{ return inside(p.x,p.y,p.z, d.x,d.y,d.z); }

	// Inside cached
	GZone*	inside( const Point& p, const Vector& d,
			const double t, Cache<CBody3D>& cache) const;

	// return boundary crossings and track-length of a ray from position pos
	// with direction dir
	// On call tlength should have the maximum length to scan
	unsigned boundaryCrossings(const Point& pos, const Point& dir,
				double* tlength, Cache<CBody3D>& cache) const;

	size_t	memory() const;

private:
	//void	clearBB();
friend class VRegion;
friend class GeometryEngine;		// FIXME remove, make functions for all lattice elements dlat, plat, ....
}; // GRegion

std::ostream& operator << (std::ostream&, const GRegion&);
#endif
