/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "quabody.h"

// ================================= GQUABody ==================================
// FIXME position, get, Mesh...
void GQUABody::set(int /*n*/, double* what)
{
	resetBBox();
	switch (type()) {
		case Type::QUA:
			round0(what, 10, TOOSMALL);
			q = Quad(what[0], what[1], what[2],	// x² y² z²
				 what[3], what[4], what[5],	// xy xz yz
				 what[6], what[7], what[8],	// x  y  z
				 what[9]);			// c
			break;
		case Type::GQ:
			round0(what, 10, TOOSMALL);
			q = Quad(what[0], what[1], what[2],	// Ax² By² Cz²
				 what[3], what[5], what[4],	// Dxy Eyz Fxz warning inverted E,F!
				 what[6], what[7], what[8],	// Gx  Hy  Jz
				 what[9]);			// K
			break;
		case Type::SQ: {
			round0(what, 10, TOOSMALL);
			double A  = what[0];		// A(x-xo)²
			double B  = what[1];		// B(y-yo)²
			double C  = what[2];		// C(z-zo)²
			double D  = what[3];		// 2D(x-xo)
			double E  = what[4];		// 2E(y-yo)
			double F  = what[5];		// 2F(z-zo)
			double G  = what[6];		// G
			P.set(what[7], what[8], what[9]);
			q = Quad(A, B, C,		// x² y² z²
				 0, 0, 0,		// xy xz yz
				 2*(D-A*P.x),		// x
				 2*(E-B*P.y),		// y
				 2*(F-C*P.z),		// z
				 A*Sq(P.x) + B*Sq(P.y) + C*Sq(P.z) - 2*(D*P.x + E*P.y + F*P.z) + G);
			}
			break;
		default:
			assert(false);
	}
	q.normalize();

	Matrix3 M;
	q.matrix3(&M);
	Point v(-q.Cx/2.0, -q.Cy/2.0, -q.Cz/2.0);
	M.inverse();
	P = M * v;
} // set

/* --- get --- */
int GQUABody::get(double* what) const
{
	switch (type()) {
		case Type::QUA:
			what[0] = q.Cxx;
			what[1] = q.Cyy;
			what[2] = q.Czz;
			what[3] = q.Cxy;
			what[4] = q.Cxz;
			what[5] = q.Cyz;
			what[6] = q.Cx;
			what[7] = q.Cy;
			what[8] = q.Cz;
			what[9] = q.C;
			round0(what, 10, TOOSMALL);
			return 10;
		case Type::GQ:
			what[0] = q.Cxx;		// A
			what[1] = q.Cyy;		// B
			what[2] = q.Czz;		// C
			what[3] = q.Cxy;		// D xy
			what[4] = q.Cyz;		// E yz Warning inverted
			what[5] = q.Cxz;		// F zx      -//-
			what[6] = q.Cx;			// G
			what[7] = q.Cy;			// H
			what[8] = q.Cz;			// J
			what[9] = q.C;			// K
			round0(what, 10, TOOSMALL);
			return 10;
		case Type::SQ:
			what[0] = q.Cxx;		// A
			what[1] = q.Cyy;		// B
			what[2] = q.Czz;		// C
			what[3] = q.Cx/2 + q.Cxx*P.x;	// D
			what[4] = q.Cy/2 + q.Cyy*P.y;	// E
			what[5] = q.Cz/2 + q.Czz*P.z;	// F
			what[6] = q.C - q.Cxx*Sq(P.x) - q.Cyy*Sq(P.y) - q.Czz*Sq(P.z)
				      + 2*(what[3]*P.x + what[4]*P.y + what[5]*P.z);
			what[7] = P.x;
			what[8] = P.y;
			what[9] = P.z;
			round0(what, 10, TOOSMALL);
			return 10;
		default:
			assert(false);
			return 0;
	}
} // get

/** position
 * @param x,y,z	new position to set
 */
void GQUABody::position(const Point& r)
{
	GBody::position(r);

	Point T = _hasMatrix? _invMatrix*r : r;
	T -= SP;
	q = sq;
	q.translate(-T);
} // position

/** save */
void GQUABody::save()
{
	GBody::save();
	sq = q;
} // save

/** restore */
void GQUABody::restore()
{
	q = sq;
	GBody::restore();
} // restore

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GQUABody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	if (item==1) {
		Point T = _hasMatrix? _invMatrix*r : r;
		T -= SP;
		q.translate(SP);		// move to origin
		q.C = 0.0;			// The Cx,Cy,Cz terms should be zero
		q.C = -q(T);			// To scale calculate the inverse of the at position with C=0
		q.translate(-SP);		// translate back to position
	} else
		GBody::move(item, r, w);
} // move

/** rotate body around an axis */
void GQUABody::rotate(const double angle, const Vector& axis)
{
	Matrix4 Rot, RotInv;
	Rot.rotate(-angle, axis);
	RotInv.rotate(angle, axis);
	q = sq;
	q.translate(SP);			// move to origin
	q.transform(Rot, RotInv);		// rotate
	q.translate(-SP);			// translate back to position
	if (type()==Type::SQ) _type = Type::GQ;
	resetBBox();
} // rotate

/** @return bounding box of body */
void GQUABody::_updateBBox()
{
	_bbox = Q(0).bbox();
} // bbox

#ifdef USE_OBBOX
/** @return oriented bounding box of body */
void GQUABody::_updateOBB(bool in)
{
	if (!in)
		_obbox[(int)in] = OBBox(Q(0).obbox());
} // _updateOBB
#endif
