/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#ifndef __GEOMETRY_VBODY_H
#define __GEOMETRY_VBODY_H

#include "gbody.h"
#include "segment.h"

#include "math/conic.h"
#include "tool/array.h"
#include "graph/viewport.h"

/* ============================== VBody =============================== */
/** VBody viewer body class */
class VBody {
private:
	GBody*	_body;		// reference to Gbody
	int	_generation;	// Generation number, to be checked for validity against the _body generation

public:
	bool	notref;		// set when NO visible zone is referring to the body
	bool	visible;	// set if body is visible in clipping area
	int	nC;		// Number of conics for the region plotting
	GBody::Location location;// if window is inside or outside the body
				// condition to be checked ONLY if nC=0
	Conic	C[GBody::BODYCONICS];		// Conics
	double	acc[GBody::BODYQUADS];		// accuracy factor to be multiplied on each quad
	int	c2q[GBody::BODYCONICS];		// Conic to quad index conversion
	std::vector<Segment>	V[GBody::BODYCONICS]; // Array of intersecting vertices

protected:
#ifdef THREAD
mutable	pthread_mutex_t	mutex;	// locking mutex when running through threadpool
				// XXX rwlocks are less efficient due to the overload
				// involved. For big operations might pay back
				// but not for simple operations like accessing
				// information
#endif

public:
	VBody(GBody* abody=NULL);
	~VBody();
	void	init(GBody* abody);

const	std::string&	name()	const { return _body->name(); }
const	char*	cname()		const { return _body->cname(); }
	GBody::Type type()	const { return _body->type(); }
const	char*	typeStr()	const { return _body->typeStr(); }
	GBody*	body()		const { return _body; }
	int	id()		const { return _body->id(); }
	dword	color()		const { return _body->color(); }
	dword	lineWidth()	const { return _body->lineWidth(); }
	dword	show()		const { return _body->show; }

	// if body is valid
	bool	valid()		const { return _generation == _body->generation(); }
	bool	invalid()	const { return _generation != _body->generation(); }
	void	setInvalid()	      { _generation = -1; }
	void	setValid()	      { _generation = _body->generation(); }

	// Conics
	void	makeConics(const ViewPort& view);
	void	intersectSelf(const ViewPort& view);
	void	intersectViewport(const ViewPort& view);
	void	calculateAccuracy(const ViewPort& view);
	void	updateLocation(const ViewPort& view);
	bool	intersectBody(VBody* abody, const ViewPort& view);
	void	intersectConic(const Conic& conic, const ViewPort& view);

#ifdef THREAD
	// mutex
	int	lock()		const	{ return pthread_mutex_lock(&mutex); }
	int	unlock()	const	{ return pthread_mutex_unlock(&mutex); }
#else
	void	lock()		const	{}
	void	unlock()	const	{}
#endif

#if 0
	// zone bookkeeping operations, protected by mutex locks
	int	nzones() const {
			lock();
			int nz = zones.size();
			unlock();
			return nz;
		}
	VZone*	zone(const int i) const	{
			lock();
			VZone* z = zones[i];
			unlock();
			return z;
		}
	void	appendZone(VZone* zone) {
			lock();
			zones.add(zone);
			unlock();
		}
	void	insertZone(VZone* zone) {
			lock();
			zone->hitCounter++;
			for (int j=0; j<zones.size(); j++)
				if (zones[j]->hitCounter > zone->hitCounter) {
					zones.insert(j, zone);
					unlock();
					return;
				}
			zones.add(zone);
			unlock();
		}
	void	promoteZone(int i) {
			lock();
			VZone* z = zones[i];
			z->hitCounter++;
			if (i<zones.top()) {
				int j=i+1;
				while (z->hitCounter>zones[j]->hitCounter and j<zones.size())
					j++;

				if (j>i+1) {
					if (j==zones.size()) {
						zones.erase(i);
						zones.add(z);
					} else {
						zones.erase(i);
						zones.insert(j, z);
					}
				}
			}
			unlock();
		}
#endif

	// Vertices
	void	addVertex(const int i, const double x, const double y, VBody* b) {
			lock();
			double t = C[i].getT(x,y);
			V[i].push_back(Segment(t,x,y,b));
#if _DEBUG>1
			if (Eq(t,0, 1e-10) and
			    Eq(x,7.74, 1e-10)) {
				printf("break1\n");
				t = C[i].getT(x,y);
			}
#endif
			unlock();
		}
	void	delVertices();
	void	removeInvalidVertices();
	void	markInvalidVertices(const VBody* invalidBody);
	void	removeWrongVertices();
	void	sortVertices();

	// Quadrics
	int	nQ()		const { return _body->nQ(); }
const	Quad& Q(int n)		const { return _body->Q(n); }

	bool	inside( const double  x, const double  y, const double  z,
			const double dx, const double dy, const double dz,
			const int ignore1=-1, const int ignore2=-1)
			{ return _body->inside(x,y,z, dx,dy,dz, acc, ignore1, ignore2); }

//	bool	isOn(const double x, const double y, const double d) const;

	// bounding box
	BBox	bbox2D() const;

	// No visible zone is referring to this body
	void	notReferenced(const bool n) { notref = n; }
	bool	notReferenced() const { return notref or _body->zones.empty(); }
	bool	ignore()	const { return (location != GBody::Location::OVERLAP); }

	size_t	memory() const;

private:
	double	accuracy(const int q, const double x, const double y, const double z) const
			{ return Abs(Q(q)(x,y,z)) / Q(q).acc(x,y,z,SMALL1); }
}; // VBody

std::ostream& operator << (std::ostream&, const VBody&);

#endif
