/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	20-Apr-2020
 */

#ifndef __GEOMETRY_VOLUME_H
#define __GEOMETRY_VOLUME_H

#include "gjob.h"
#include "gbody.h"
#include "math/random.h"

class BBox;
class Random;
class GRegion;
class Geometry;

//namespace flair {

// =============================== VolumeWorker ================================
// Calculate the volume and the area of a region from a MC tracklength and
// boundary crossing estimation using sflood
#ifdef THREAD
class VolumeJob;
class VolumeWorker : public ThreadPoolWorker {
public:
	Cache<CBody3D> cache;		/** cache of each worker	*/
	int	idx;			/** index of event		*/
	int	to;			/** up to which event to calc	*/
	int	bdx;			/** boundary crossings		*/
	double	tracklength;		/** track length tally		*/
	Random	random;			/** random number generator	*/
	long	randSeed;		/** seed for QMC sequence generation **/

public:
	VolumeWorker(): idx(0), to(0), bdx(0) {}
virtual	void	operator ()();
	void	reset(long s, int nbodies) {
				cache.resize(nbodies);
				random.seed(s);
				randSeed = s;
				idx = to = bdx = 0;
				tracklength = 0;
			}
}; // class VolumeWorker

// ================================ VolumeFeeder ================================
class VolumeFeeder : public ThreadPoolFeeder {
public:
const	VolumeJob*	job;		/** volume job pointer		*/
	int		samples;	/** maximum number of samples	*/

private:
	int		nworkers;	/** number of workers allocated	*/
	VolumeWorker*	workers;	/** array of body workers	*/
	int		idx;		/** next start index		*/
	int		step;		/** step per worker		*/

public:
	VolumeFeeder(ThreadPool& p) : ThreadPoolFeeder(p),
		  nworkers(0),
		  workers(NULL)
		  {}
	virtual	~VolumeFeeder()		{ if (workers) delete [] workers; }

	// Initialize workers
	void	allocate();
virtual	void	reset(const VolumeJob* j, int nsamples, int nstep, Random& random);

virtual	ThreadPoolWorker* feed(int threadId);

	// Accumulate bdx
	long	bdx() {
			long h=0;
			for (int i=0; i<nworkers; i++)
				h += (long)workers[i].bdx;
			return h;
		}

	/** @return total tracklength from all workers */
	double	tracklength() {
			double t=0;
			for (int i=0; i<nworkers; i++)
				t += workers[i].tracklength;
			return t;
		}
}; // class VolumeFeeder
#endif

// ================================== VolumeJob ==================================
class VolumeJob : public GeometryJob {
public:
const	GRegion*	region;		/** region to calculate area	*/
	BBox		bbox;		/** bounding box of region	*/
	Point		center;		/** center of bounding sphere	*/
	double		radius;		/** bounding sphere radius	*/
	int		run_samples;	/** number of samples to run	*/
	int		nsamples;	/** number of samples run	*/

	double		area;		/** area of region		*/
	double		volume;		/** volume of region		*/
	double		mass;		/** mass of region		*/
	double		error;		/** relative error of result	*/

public:
	VolumeJob() : GeometryJob() {}
virtual	~VolumeJob() {}

	bool	init(const Geometry* geo, const GRegion* reg, int samples, int threads=0);

	// FIXME setting to virtual causes a coredump
	void	run();

	/** spawn a job in a background thread */
	void	spawn();
private:
static	void*	runInThread(VolumeJob* job);
}; // class VolumeJob

//} // namespace flair

#endif
