/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	05-Apr-2011
 */

#ifndef __GEOMETRYOBJECT_H
#define __GEOMETRYOBJECT_H

#include <structmember.h>
#include <Python.h>

#include <vector>

#include "os.h"
#include "math/point.h"
#include "geometry/volume.h"
#include "geometry/qmcdebug3d.h"

class GBody;
class GZone;
class GRegion;
class GObject;
class GRuler;
class Geometry;
class Material;
class GeometryKernel;

extern PyTypeObject GeometryType;

enum class CursorType {
	CCursor,
	CBody,
	CRegion,
	CZone,
	CObject
};

// ============================== GeometryObject ===============================
/** Geometry python class */
struct GeometryObject {
	PyObject_HEAD
	Geometry*	geometry;		// geometry class

	bool		cursorShow;		// show cursor
	int		cursorId;		// id of cursor object
	int		cursorSize;		// size of rotate and axis locked move
	int		cursorMoveSize;		// size of move handler (central circle)
	CursorType	cursorType;		// type of object that set the cursor
	Point		cursor;			// cursor position

	double		trackballSize;		// percent of window
	double		snapDistance;		// minimum distance in pixel to align
	double		snapAngle;		// minimum angle in radians to align

	std::vector<GObject*>	objects;	// drawable objects on screen

static const int	rulers = 3;		// Rulers to show only during body editing
	GRuler*		ruler[rulers];		// rulers for displaying body dimensions
	std::vector<std::string> errors;	// errors accumulated since last read/clear

	VolumeJob	volumeJob;		// volume job parameters
	Debug3DJob	debug3DJob;		// stochastic 3D debugging
}; // struct GeometryObject

/* function prototypes */
GBody*	  Py_GBody(   GeometryObject* self, PyObject* obj);
GRegion*  Py_GRegion( GeometryObject* self, PyObject* obj);
GObject*  Py_Object(  GeometryObject* self, PyObject* obj);
Material* Py_Material(GeometryObject* self, PyObject* obj);
PyObject* Py_ZoneExpr(GZone* zone, Geometry::Mode mode);
PyObject* Py_RegionExpr(GRegion* region, Geometry::Mode mode);

/* type */
extern PyTypeObject GeometryType;

#endif
