/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	15-Jun-2018
 */

#include "zpainter.h"

// ================================= ZPainter ==================================
/** ~ZPainter */
ZPainter::~ZPainter()
{
	if (_zdepth)		delete[] _zdepth;
	if (_superSamplingBuf)	delete[] _superSamplingBuf;
} /* ~ZPainter */

/** init
 * @param w	width of image
 * @param h	height of image
 */
bool ZPainter::init(const int w, const int h)
{
	if (Painter::init(w,h)) {
		if (_zdepth)		delete[] _zdepth;
		if (_superSamplingBuf)	delete[] _superSamplingBuf;

		_zdepth = new float[_maxSize];
		_superSamplingBuf = new dword[_maxSize];

		std::fill_n(_zdepth,           _maxSize, _zfar);
		std::fill_n(_superSamplingBuf, _maxSize, background());
		return true;
	} else
		return false;
} // init

/** clear image */
void ZPainter::clear()
{
	Painter::clear();
	if (_clip.left==0         and _clip.top==0 and
	    _clip.right==_width-1 and _clip.bottom==_height-1)
		std::fill_n(_zdepth, _dataSize, _zfar);
	else {
		int mx = _clip.right  - _clip.left;
		int my = _clip.bottom - _clip.top;
		float* src = _zdepth + _clip.top*_width+_clip.left;
		for (int i=0; i<my; i++) {
			std::fill_n(src, mx, _zfar);
			src += _width;
		}
	}
} // clear

/** unclippedZline
 * @param x1,y1,z1	starting point of line
 * @param x2,y2,z2	end point of line
 * @param color	color of line
 * @return true if it drawn any pixel
 */
bool ZPainter::unclippedZline(int x1, int y1, float z1, int x2, int y2, float z2,
				const dword color, int alpha)
{
	bool ret = false;
	int dx = Abs(x1-x2);
	int dy = Abs(y1-y2);

	if (dx > dy) {
		if (x1 > x2) {			// force x1<x2
			Swap(x1,x2);
			Swap(y1,y2);
			Swap(z1,z2);
		}
		int yinc = (y2>y1)? 1: -1;	// determine increment for y
		int d = 2*dy - dx;
		int Ainc = 2*(dy-dx);
		int Binc = 2*dy;

		int x = x1;
		int y = y1;
		float z = z1;
		ret |= zpixel(x,y,z,color,alpha);

		float dz = (z2-z1) / (float)dx;	// FIXME very crude
		for (x=x1+1; x<=x2; x++) {
			if (d >= 0) {		// setpixel A
				y += yinc;
				d += Ainc;
			} else
				d += Binc;	// setpixel B
			ret |= zpixel(x,y,z,color,alpha);
			z += dz;
		}
	} else {	/* DO THE SAME FOR Y */
		if (y1 > y2) {			// force y1<y2
			Swap(x1,x2);
			Swap(y1,y2);
			Swap(z1,z2);
		}
		int xinc = (x2>x1)? 1: -1;	// determine increment for x
		int d = 2*dx - dy;
		int Ainc = 2*(dx-dy);
		int Binc = 2*dx;

		int x = x1;
		int y = y1;
		float z = z1;
		ret |= zpixel(x,y,z,color,alpha);

		float dz = (z2-z1) / (float)dy;	// FIXME very crude
		for (y=y1+1; y<=y2; y++) {
			if (d >= 0) {		// setpixel A
				x += xinc;
				d += Ainc;
			} else
				d += Binc;	// setpixel B
			ret |= zpixel(x,y,z,color,alpha);
			z += dz;
		}
	}
	return ret;
} // unclippedZline
