/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 *					Christian.Theis@cern.ch (blue noise distributions and LD samples)
 * Date:	26-Jul-2022
 */

#ifndef __MATH_RANDOM_H
#define __MATH_RANDOM_H

#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <iostream>

#include <vector>
#include <array>

#include "point.h"

#ifdef __ANDROID_API__
	// Apparently android has drand48 but not the re-entrant version
	#define HAS_RAND48
//	#define BASIC_RAND
#elif defined(__MACH__) || defined(CYGWIN) || defined(ANDROID)
	#define HAS_RAND48
#endif

/* =============================== Vector ============================= */
/** Random class, generate random distributions using the drand48
 * C library
 */
class Random {
public:
#if defined(BASIC_RAND)
	typedef unsigned int	Int;
#else
	typedef long int	Int;
#endif
private:
	Int	_seed;			// random number seed
	Int	_qmc_seed;		// random number seed used for qmc sampling
	double	_normal;		// second normal number saved
	bool	_normalSave;
#if !defined(BASIC_RAND) && !defined(HAS_RAND48)
	struct drand48_data	rnd;
#endif

public:
	Random(Int as=0) : _normalSave(false) { seed(as!=0?as:time(NULL)); }

	/** set seed value */
	void	seed(Int as) {
			_seed = _qmc_seed = as;
#if defined(BASIC_RAND)
//			srand(_seed);
#elif defined(HAS_RAND48)
			srand48(_seed);
#else
			srand48_r(_seed, &rnd);
#endif
		}

	/** @return seed value */
	Int	seed()		const	{ return _seed; }

	/** @return seed value */
	Int	qncseed()	const { return _qmc_seed; }

	/** @return an integer random number [0, 2^31) */
	Int	integer() {
#if defined(BASIC_RAND)
			return rand_r(&_seed);
#elif defined(HAS_RAND48)
			return lrand48();
#else
			Int x;
			lrand48_r(&rnd, &x);
			return x;
#endif
		}

	/** @return an integer random number [0, imax) */
	Int	integer(Int imax) {
#if defined(BASIC_RAND)
			Int x = rand_r(&_seed);
			return x%imax;
#elif defined(HAS_RAND48)
			return lrand48()%imax;
#else
			Int x;
			lrand48_r(&rnd, &x);
			return x%imax;
#endif
		}

	/** @return a real random number [0, 1) */
	double	real() {
#if defined(BASIC_RAND)
			return (double)rand_r(&_seed) / (double)RAND_MAX;
#elif defined(HAS_RAND48)
			return drand48();
#else
			double x;
			drand48_r(&rnd, &x);
			return x;
#endif
		}

	/** @return a real random number [0, max) */
	double	real(const double max)				{ return max*real(); }

	/** @return a real random number [min, max) */
	double	real(const double min, const double max)	{ return (max-min)*real() + min; }
	double	uniform(const double min, const double max)	{ return (max-min)*real() + min; }

	/** @return a real random number [0, 1) */
	double operator ()()					{ return real(); }

	/** @return a real random number [0, max) */
	double operator ()(const double max)			{ return max*real(); }

	/** @return a real random number [min, max) */
	double operator ()(const double min, const double max)	{ return (max-min)*real() + min; }

	/** @return a real stratified random number [0,1) */
	double realqmc();

	/** @return an N-dimensional tuple of real stratified random numbers [0,1)
	     based on Niederreiter2 sequences */
	void	realqmcND( int dim, double randVec[] );

	/** @return exponential spectrum exp(-t/a) to sample the
	 *  collision distance or random decay
	 */
	double	exp(const double a)				{ return -log(real())/a; }

	void	sincos(double *s, double *c);
	double	normal();
	void	normal(double *z1, double *z2);

	void	disc(const double r, double *x, double *y);
	void	sphere(const double r, double *x, double *y, double *z);
	Vector	sphere(const double r) {
			double x, y, z;
			sphere(r, &x, &y, &z);
			return Vector(x,y,z);
		}

	Vector	vector();

	void	sflood( Point* pos, Vector* dir ) { sflood( pos, dir, real(), real() ); }
	void	sflood( Point* pos, Vector* dir, double randNr1, double randNr2);

	/** @return a vector of pairs that contain samples with blue noise that can be used for dithering. */
	std::vector<std::array<double, 2>> generateBlueNoise( size_t nrOfSamples, const int cSampleMultiplier /*= 5*/ );

protected:
	/** @return the toroidal distance (or squared distance for efficiency) between two sample points [0,1)
	 */
	double ToroidalDist( double x, double y, double curSampleX, double curSampleY ) {
			double dx = fabs( x - curSampleX );
			double dy = fabs( y - curSampleY );
			if (dx > 0.5) dx = 1.0 - dx;
			if (dy > 0.5) dy = 1.0 - dx;
			return sqrt(dx * dx + dy * dy);
		}

	double ToroidalDistSquared( double x, double y, double curSampleX, double curSampleY ) {
			double dx = fabs( x - curSampleX );
			double dy = fabs( y - curSampleY );
			if (dx > 0.5) dx = 1.0 - dx;
			if (dy > 0.5) dy = 1.0 - dx;
			// return squared distance which is more efficient for comparisons
			return dx * dx + dy * dy;
		}
}; // Random

#endif
