#
# Copyright and User License
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# DISCLAIMER
# ~~~~~~~~~~
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY, OF
# SATISFACTORY QUALITY, AND FITNESS FOR A PARTICULAR PURPOSE
# OR USE ARE DISCLAIMED. THE COPYRIGHT HOLDERS AND THE
# AUTHORS MAKE NO REPRESENTATION THAT THE SOFTWARE AND
# MODIFICATIONS THEREOF, WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADE SECRET OR OTHER PROPRIETARY RIGHT.
#
# LIMITATION OF LIABILITY
# ~~~~~~~~~~~~~~~~~~~~~~~
# THE COPYRIGHT HOLDERS AND THE AUTHORS SHALL HAVE NO
# LIABILITY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
# CONSEQUENTIAL, EXEMPLARY, OR PUNITIVE DAMAGES OF ANY
# CHARACTER INCLUDING, WITHOUT LIMITATION, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES, LOSS OF USE, DATA OR PROFITS,
# OR BUSINESS INTERRUPTION, HOWEVER CAUSED AND ON ANY THEORY
# OF CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE), PRODUCT
# LIABILITY OR OTHERWISE, ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGES.
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   07-Apr-2016
#

LIBRARY = lib$(NAME).a
SOLIB   = $(NAME).$(SO)
OBJECTS = $(SOURCES:.cc=.$(OBJ))
DEPS    = $(SOURCES:.cc=.d)

# **************************
# Version
# **************************
ifeq ($(CXX),clang++)
	CXXVERSION := $(shell $(CXX) --version | grep version | cut -f3 -d\ )
else
	CXXVERSION := $(shell $(CXX) -dumpfullversion 2> /dev/null)
endif
ifeq ($(CXXVERSION),)
	CXXVERSION := $(shell $(CXX) --version | head -1 | cut -f3 -d\ )
endif
CXXRELEASE := $(shell echo "$(CXXVERSION)" | cut -f1,2 -d.)
CXXMAJOR   := $(shell echo "$(CXXVERSION)" | cut -f1 -d.)

# **************************
#  CXX Flags
# **************************
CXXFLAGS  = -g
CXXFLAGS += -D$(SYSTEM)
CXX_STD_FLAG = -std=c++11
ifeq ($(STD),14)
	CXX_STD_FLAG = -std=c++14
endif
ifeq ($(STD),17)
	CXX_STD_FLAG = -std=c++17
endif
ifeq ($(STD),20)
	CXX_STD_FLAG = -std=c++20
endif
CXXFLAGS += ${CXX_STD_FLAG}

# **************************
#  Debug + Optimization
# **************************
ifeq ($(DEBUG),yes)
	# _DEBUG can be 0, 1 or 2
	# NDEBUG is used by assert()
ifdef DEBUGLEVEL
	CXXFLAGS += -D_DEBUG=${DEBUGLEVEL}
else
	CXXFLAGS += -D_DEBUG=1
endif
else
	# Optimised version
ifeq ($(COMP),clang)
	CXXFLAGS += -O
	CXXFLAGS += -ftrapv
else
	CXXFLAGS += -O3
#	CXXFLAGS += -fexpensive-optimizations
	ifeq ($(CXXRELEASE), 7.2)
		# Work around reported bug https://gcc.gnu.org/bugzilla/show_bug.cgi?id=82524 (7.2.x)
		# fixed from 7.3.x
		CXXFLAGS += -fno-expensive-optimizations
	endif
endif
	CXXFLAGS += -funroll-loops
ifneq ($(DEBUG),yes)
	# disable in optimized version the setting of errno on floating operations
	# there is some speed gain with that
	CXXFLAGS += -fno-math-errno
endif
ifeq ($(FASTMATH), yes)
	# Floating point, prefer sse and if not then 387
	CXXFLAGS += -ffast-math
	CXXFLAGS += -ftree-vectorizer-verbose=5
	CXXFLAGS += -ftree-vectorize
	CXXFLAGS += -mfpmath=sse,387
	CXXFLAGS += -mfpmath=387
endif
	CXXFLAGS += -DNDEBUG
	CXXFLAGS += -D_DEBUG=0
	#CXXFLAGS += -mtune=native
	#CXXFLAGS += -march=native
	CXXFLAGS += -ftree-vectorize

	# verbose information on optimization
#	CXXFLAGS += -ftree-vectorizer-verbose=2
#	CXXFLAGS += -fopt-info
#	CXXFLAGS += -fopt-info-vec-missed
endif

# **************************
#  Defines
# **************************
ifeq ($(THREAD),yes)
	CXXFLAGS += -DTHREAD
endif

ifeq ($(MEM),yes)
	CXXFLAGS += -DMEM
endif

# Use the DUMP=yes to enable logging
ifeq ($(DUMP),yes)
	CXXFLAGS += -D_DUMP
endif

ifeq ($(EXP),yes)
	CXXFLAGS += -DEXPERIMENTAL
endif

ifeq ($(FARMEXP),yes)
	CXXFLAGS += -DFARM_EXPERIMENTAL
endif

ifeq ($(SLOWAA),yes)
	CXXFLAGS += -DSINGLE_AA
endif

ifeq ($(OPENGL),yes)
	CXXFLAGS += -DOPENGL
endif

ifeq ($(STAT),yes)
	CXXFLAGS += -DSTAT
endif

ifeq ($(OPENMP),yes)
	CXXFLAGS += -fopenmp
endif

ifeq ($(XFT),yes)
	CXXFLAGS += -DXFT
	CXXFLAGS += $(shell pkg-config --cflags freetype2)
endif

ifeq ($(PROF),yes)
	# abbreviation for PROFILE
	PROFILE=yes
endif

ifeq ($(PROFILE),yes)
	CXXFLAGS += -pg
	# Bug on Ubuntu/Debian
	#CXXFLAGS += -no-pie
endif

CXXFLAGS += ${USERFLAGS}

# **************************
#  Warnings
# **************************
CXXFLAGS += -pedantic
CXXFLAGS += -W
CXXFLAGS += -Wall
CXXFLAGS += -Wcast-qual
ifeq ($(SIGN),yes)
	CXXFLAGS += -Wconversion
	#CXXFLAGS += -Wfloat-conversion	# to be removed and replaced with the above
endif
CXXFLAGS += -Wextra
CXXFLAGS += -Wformat
CXXFLAGS += -Wpointer-arith
CXXFLAGS += -Wshadow
CXXFLAGS += -Wno-write-strings
ifneq ($(CXX), icc)
	CXXFLAGS += -Wcast-align
	CXXFLAGS += -Wredundant-decls
endif
#ifeq ($(CXX),g++)
#	CXXFLAGS += -Wconversion-null
#endif
#CXXFLAGS += -Wfloat-equal
#CXXFLAGS += -Winline
#CXXFLAGS += -Wno-deprecated
ifeq ($(WERROR),yes)
	CXXFLAGS += -Werror
endif
CXXFLAGS += -Wno-long-long
CXXFLAGS += -Wno-non-virtual-dtor
CXXFLAGS += -Wno-variadic-macros
CXXFLAGS += -Woverloaded-virtual
CXXFLAGS += -Wpointer-arith
CXXFLAGS += -Wshadow
CXXFLAGS += -Wunused
CXXFLAGS += -Wuninitialized
CXXFLAGS += -Wwrite-strings
CXXFLAGS += -Wshadow
CXXFLAGS += -pipe
CXXFLAGS += -pthread
#CXXFLAGS += -ftls-model=initial-exec
#CXXFLAGS += -Wl,--no-undefined

ifeq ($(COMP), gcc)
	# FIXME gcc v8 complains for array.h
	ifeq ($(shell test $(CXXMAJOR) -gt 7; echo $$?), 0)
		CXXFLAGS += -Wno-class-memaccess
	endif
endif

# **************************
#  Warnings
# **************************
# XXX important for floating point speed up. Look fedora-14 release notes
#CXXFLAGS += -fexcess-precision=fast	# for gcc 4.5
# Signed arithmetic overflow of +,-,* wraps using 2-complement
CXXFLAGS  += -fwrapv
CXXFLAGS  += -fstack-protector
ALIASING   = -fstrict-aliasing
NOALIASING = -fno-strict-aliasing

ifeq ($(SYSTEM),UNIX)
	CXXFLAGS += -fpic
else ifeq ($(SYSTEM),ANDROID)
	CXXFLAGS += -fpic
	CXXFLAGS += -fexceptions
#	CXXFLAGS += -frtti
endif

ifeq ($(SYSTEM),Darwin)
	CXXFLAGS += -isysroot $(shell xcrun --show-sdk-path)
	SOFLAGS  = -dynamiclib
else
	SOFLAGS  = -shared
	LDFLAGS += -Wl,-O4
endif

# **************************
# Include directories
# **************************
INCDIR = -I. -I..

# **************************
# Link flags
# **************************
LDFLAGS += $(LIBDIR) $(LIBS)

# **************************
# Archive
# **************************
ifeq ($(SYSTEM),Darwin)
	AR  = ar rv
else
	AR  = ar rvU
endif

# ******************************************
# Generic rule for compiling any cc to o
# ******************************************
%.$(OBJ): %.cc
	$(CXX) -c $< $(ARCH) $(CXXFLAGS) $(ALIASING) $(INCDIR)

# **************************
# Dependencies rule
# **************************
%.d: %.cc
	$(CXX) -MM $(CXX_STD_FLAG) $(INCDIR) $< > $@

# **************************
# For debugging makefile
# Variable printing target
# make print-INCDIR
# **************************
print-%:
	@echo "$* = $($*)"
