/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#include "usrbinlayer.h"

#include <string.h>

#include "viewer.h"
#include "data/usrbin.h"
#include "data/vtkdata.h"
#include "data/meshdata.h"

// ================================ UsrbinLayer ================================
UsrbinLayer::UsrbinLayer(const Geometry& g, GeometryKernel& k, GeometryViewer& v)
	: Layer(g, k, v)
{
	ndata = 0;
	bzero(meshdata, sizeof(meshdata));
	bzero(_alpha,   sizeof(_alpha));
	bzero(_palette, sizeof(_palette));
	bzero(_showErr, sizeof(_showErr));
} // UsrbinLayer

/** UsrbinLayer */
UsrbinLayer::~UsrbinLayer()
{
	for (int i=0; i<ndata; i++)
		if (meshdata[i] != nullptr)
			delete meshdata[i];
} // UsrbinLayer

/** load mesh from filename/det at position idx
 * @param idx          mesh position to load it to. if <0 append at the end
 * @param filename     filename to load (usrbin, vtk legacy are supported)
 * @param det          detector to load from file
 * @return true on success, false on failure
 */
bool UsrbinLayer::load(int idx, const char* filename, const int det)
{
	if (idx<0) idx = ndata++;
	if (idx >= MAXMESH) return false;
	ndata = Max(ndata, idx+1);

	const char* ext = strchr(filename,'.');
	if (ext and !strcmp(ext,".vtk")) {
		if (!meshdata[idx])
			meshdata[idx] = new VtkData();
		else
		if (!dynamic_cast<VtkData*>(meshdata[idx])) {
			delete meshdata[idx];
			meshdata[idx] = new VtkData();
		}
	} else {
		// default is Usrbin format
		if (!meshdata[idx])
			meshdata[idx] = new Usrbin();
		else
		if (!dynamic_cast<Usrbin*>(meshdata[idx])) {
			delete meshdata[idx];
			meshdata[idx] = new Usrbin();
			return meshdata[idx]->load( filename, det + 1);
		}
	}

	return meshdata[idx]->load(filename, det);
} // load

/** checker */
bool UsrbinLayer::checker(int idx, const bool rphiz,
		double axlow, double axhigh, const int anx,
		double aylow, double ayhigh, const int any,
		double azlow, double azhigh, const int anz)
{
	if (idx<0) idx = ndata++;
	if (idx >= MAXMESH) return false;
	ndata = Max(ndata, idx+1);

	if (!meshdata[idx])
		meshdata[idx] = new MeshData();
	else
	if (!dynamic_cast<MeshData*>(meshdata[idx])) {
		delete meshdata[idx];
		meshdata[idx] = new MeshData();
	}

	meshdata[idx]->checker( rphiz,
			axlow, axhigh, anx,
			aylow, ayhigh, any,
			azlow, azhigh, anz);

	return true;
} // checker

/** draw
 * FIXME 1. Can be optimized by introducing a bounding box around the image and calculate
 *          first the intersection of the 6 planes
 *       2. Modified to check for non-lines, in order to draw also the regions
 */
void UsrbinLayer::draw(Painter& painter)
{
	if (ndata<=0) return;
	Point r(0.0, 0.0, 0.0);
	int H = painter.height();	// save variables that might change
	int W = painter.width();	// during the scanning...
	Color32* ptr = (Color32*)painter.data();	// data pointer

	for (int j=0; j<H; j++) {
		if (stop()) break;
		r.y = view().jc2v(j);

		for (int i=0; i<W; i++, ptr++) {
			dword flag = *ptr & FLAG_INFOMASK;
			if (flag==0) continue;
			if ((viewer.d3.show && flag!=FLAG_REGION) ||
			    flag&(FLAG_FINAL|FLAG_ALPHA)) continue;
			r.x = view().ic2u(i);

			for (int idx=0; idx<=ndata; idx++) {
				auto usrbin = meshdata[idx];
				if (usrbin == nullptr) continue;
				if (!usrbin->hasData() and !usrbin->checker()) continue;
				Point p = view().matrix() * r;

				double value;
				if (!usrbin->getData(p.x, p.y, p.z, &value, _showErr[idx])) continue;
				// convert errors to %
				if (_showErr[idx]) value *= 100;

				Color32 usr;
				int pal = _palette[idx];
				if (usrbin->checker()) {
					if (value<0.5)
						usr = viewer.palette[pal].first();
					else
						usr = viewer.palette[pal].last();
				} else {
					dword color = viewer.palette[pal].color(value);
					if (color == COLOR_TRANSPARENT)
						break;
					else
						usr = color;
				}
				ptr->alpha() = alpha(idx);
				usr.alpha()  = 0xFF;
				ptr->blend(usr);
			}
		}
	}
} // draw

/** shade
 * @param hit   input 3D hit position
 * @param color output color at hit position
 * @return 0xFF if no usrbin is found at hit position otherwise 0..255 alpha level
 */
byte UsrbinLayer::shade(const Point& hit, Color& color) const
{
	// move a bit the hit position to avoid boundary effects
	for (int i=0; i<ndata; i++) {
		const auto usrbin = meshdata[i];
		if (usrbin == nullptr) continue;
		if (!usrbin->hasData() && !usrbin->checker()) continue;

		double value;
		if (!usrbin->getData(hit.x, hit.y, hit.z, &value, _showErr[i] )) continue;
		if ( _showErr[i] ) value *= 100;

		Color usrcolor;
		//color->rgb.alpha = alpha(i);
		int pal = _palette[i];
		dword c;
		if (usrbin->checker()) {
			if (value<0.5)
				c = viewer.palette[pal].first();
			else
				c = viewer.palette[pal].last();
		} else {
			c = viewer.palette[pal].color(value);
			if (c == COLOR_TRANSPARENT) break;
		}
		usrcolor.set(c);
		color.blend(usrcolor, (float)alpha(i)/255.0f);
		return alpha(i);
	}
	return 0xFF;
} // shade

/** regionColorFromMeshData
 * Set color of all regions using the USRBIN values
 * if not defined set to 0
 */
void UsrbinLayer::regionColorFromMeshData(int idx)
{
	if (meshdata[idx]==nullptr or !meshdata[idx]->hasData()) return;
	int pal = _palette[idx];
	for (auto region : kernel.regions) {
		region->color(viewer.palette[pal].first());
		double value;
		if (meshdata[idx]->getData(region->id(),&value, _showErr[idx] ))
			region->color(viewer.palette[pal].color(value));
	}
} // regionColorFromMeshData

/** bump map shading
 * @param hit   input 3D hit position
 * @param color output color at hit position
 * @return 0xFF if no usrbin is found at hit position otherwise 0..255 alpha level
 */
byte UsrbinLayer::bumpShade(Ray& hitRay, const VLight& VL, double scaling, int direction, Color& color) const
{
	const MeshData* usrbin = nullptr;
	const Point     hit = hitRay.hit();

	byte alphaVal = shade( hitRay.hit(), color );
	if ( alphaVal == 0xFF )
		return alphaVal;

	// let's get the perturbed normal
	int i = -1, j = -1, k = -1;
	int idx;
	for ( idx = 0; idx < ndata; idx++ ) {
		usrbin = meshdata[idx];
		if ( usrbin == nullptr ) continue;
		if ( !usrbin->hasData() && !usrbin->checker() ) continue;
		if ( !usrbin->xyz2ijk( hit.x, hit.y, hit.z, &i, &j, &k) ) continue;
		break;
	}
	assert( i >= 0 && j >= 0 && k >= 0 );

	Vector GradientVec;
	double lhX = usrbin->getData( std::min( i + 1, usrbin->nx - 1 ), j, k, _showErr[idx] );
	double rhX = usrbin->getData( std::max( i - 1, 0 ), j, k, _showErr[idx] );
	double lhY = usrbin->getData( i, std::min( j + 1, usrbin->ny - 1 ), k, _showErr[idx] );
	double rhY = usrbin->getData( i, std::max( j - 1, 0 ), k, _showErr[idx] );
	double lhZ = usrbin->getData( i, j, std::min( k + 1, usrbin->nz - 1 ), _showErr[idx] );
	double rhZ = usrbin->getData( i, j, std::max( k - 1, 0 ), _showErr[idx] );

	double dX = scaling * (lhX - rhX) / (2.0 * usrbin->dx);
	double dY = scaling * (lhY - rhY) / (2.0 * usrbin->dy);
	double dZ = scaling * (lhZ - rhZ) / (2.0 * usrbin->dz);

	Vector lightDir = VL.pos - hit;
	lightDir.normalize();

	Vector bumpNormal = hitRay.normal - Vector( dX, dY, dZ);
	bumpNormal.normalize();

	double dot = direction * bumpNormal * lightDir;
	color *= fabs(dot);

	return alphaVal;

} // bumpShade
